/*--------------------------------------------------------------------

        CHILD.C  [anonymous pipe version]

        Contains the child process for the PROCESS demo program.
        The program in parent.c creates this child process.
        In this version, the two processes communicate through
        an anonymous pipe.

        FUNCTIONS
            WinMain             receive and dispatch messages
            Init                initialize program data
            CreateMainWindow    creates program's overlapping window
            Child_WndProc       sort out messages for main window
            Child_OnCreate      create named pipe
            Child_OnPaint       draw selected figure
            Child_OnCommand     process menu commands
            Child_OnDestroy     end program
            PipeThread          thread procedure for reading from pipe
            DoRead              procedure to read from pipe
            About_DlgProc       sort out messages for About box
            ShowErrorMsg        calls GetLastError and displays the message

        from Mastering Windows NT Programming
        copyright 1993 by Brian Myers & Eric Hamer

  -------------------------------------------------------------------*/

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include "child.h"
#include "global.h"             // values common to both processes

/*--------------------------------------------------------------------
        GLOBAL VARIABLES
  -------------------------------------------------------------------*/

HWND        hwndMain;           // main window
HINSTANCE   hInst;              // current instance
HANDLE      hPipeRead;          // handle to the anonymous pipe
HANDLE      hThread;            // thread that reads from pipe
DWORD       dwThreadID;         // id of pipe thread
int         iShape;             // figure to draw
int         iSize;              // size of figure to draw (SMALL/LARGE)
int         iColor;             // color of figure to draw
BOOL        bTerminate = FALSE; // TRUE when program should end

/*------------------------------------------------------------------
        WIN MAIN
        Calls initializing procedures and runs the message loop
  ------------------------------------------------------------------*/

int WINAPI WinMain( HINSTANCE hinstThis, HINSTANCE hinstPrev,
				    LPSTR lpszCmdLine,   int iCmdShow )
{
    MSG msg;

    hInst = hinstThis;
    iShape = 0;								// don't draw immediately
    if( ! Init() ) return( FALSE );
    ShowWindow( hwndMain, iCmdShow );		// display the main window
    UpdateWindow( hwndMain );
    while( GetMessage( &msg, NULL, 0, 0 ) )	// translate the message
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }
    return( msg.wParam );
    UNREFERENCED_PARAMETER( hinstPrev );
    UNREFERENCED_PARAMETER( lpszCmdLine );
}

/*------------------------------------------------------------------
        INIT
        Register the window class and the create the window
  ------------------------------------------------------------------*/

BOOL Init()
{
    WNDCLASS wc;
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

    if( ! LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) ) )
    {
        ShowErrorMsg();
        return( FALSE );
    }
    if( ! LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle) ) )
    {
        ShowErrorMsg();
        return( FALSE );
    }
		// register the main window class
    wc.lpszClassName    = szAppName;
    wc.hInstance        = hInst;
    wc.lpfnWndProc      = Child_WndProc;
    wc.hCursor          = LoadCursor(NULL, IDC_ARROW);
    wc.hIcon            = LoadIcon( hInst, MAKEINTRESOURCE(ICON_APP) );
    wc.lpszMenuName     = MAKEINTRESOURCE( MENU_MAIN );
    wc.hbrBackground    = (HBRUSH)( COLOR_WINDOW + 1 );
    wc.style            = CS_HREDRAW | CS_VREDRAW;
    wc.cbClsExtra       = 0;
    wc.cbWndExtra       = 0;
    if( ! RegisterClass( &wc ) ) return( FALSE );
		// create the window
    return( CreateMainWindow() );
}

/*------------------------------------------------------------------
        CREATE MAIN WINDOW
        Create the program's overlapping window.
  ------------------------------------------------------------------*/

BOOL CreateMainWindow()
{
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

		// load the relevant strings
    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle) );
		// create the parent window
    hwndMain = CreateWindow( szAppName, szTitle,
							 WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
							 CW_USEDEFAULT, CW_USEDEFAULT, 
							 CW_USEDEFAULT, CW_USEDEFAULT,
							 NULL, NULL, hInst, NULL );
		// return FALSE for an error
    return( hwndMain != NULL );
}

/*------------------------------------------------------------------
        CHILD_WNDPROC
        This is where the messages for the main window are processed.
  ------------------------------------------------------------------*/

LRESULT WINAPI Child_WndProc( HWND hWnd,     UINT uMessage,
							  WPARAM wParam, LPARAM lParam )
{
    switch( uMessage )
    {
        HANDLE_MSG( hWnd, WM_CREATE,  Child_OnCreate );
        HANDLE_MSG( hWnd, WM_PAINT,	  Child_OnPaint );
        HANDLE_MSG( hWnd, WM_COMMAND, Child_OnCommand );
        HANDLE_MSG( hWnd, WM_DESTROY, Child_OnDestroy );
        default:
            return( DefWindowProc(hWnd, uMessage, wParam, lParam) );
    }
    return( 0L );
}

/*------------------------------------------------------------------
        CHILD_ONCREATE
        On startup open the pipe and start the thread that will
        read from it.
  ------------------------------------------------------------------*/

BOOL Child_OnCreate( HWND hWnd, LPCREATESTRUCT lpCreateStruct )
{
		// open the pipe for reading
    hPipeRead = GetStdHandle( STD_INPUT_HANDLE );
    if (hPipeRead == INVALID_HANDLE_VALUE)
    {
        ShowErrorMsg();
        return( FALSE );
    }

		// Create the thread that will read from the pipe.  It is created
		// suspended so its priority can be lowered before it starts.
    hThread = CreateThread( NULL,           // security attributes
                0,                          // initial stack size
                (LPTHREAD_START_ROUTINE)PipeThread,
                NULL,                       // argument
                CREATE_SUSPENDED,           // creation flag
                &dwThreadID );              // new thread's id

    if( ! hThread )
    {
        ShowErrorMsg();
        return( FALSE );
    }
		// lower the thread's priority and let it run
    SetThreadPriority( hThread, THREAD_PRIORITY_BELOW_NORMAL );
    ResumeThread( hThread );

    return( TRUE );
    UNREFERENCED_PARAMETER( hWnd );
    UNREFERENCED_PARAMETER( lpCreateStruct );
}

/*------------------------------------------------------------------
        CHILD_ONPAINT
        Redraw the window, adding the figure indicated by iShape.
        The IDM_ constants used in this procedure are defined in
        global.h.  In the parent process they represent menu
        commands.  The parent uses the constants to tell the
        child what to draw.
  ------------------------------------------------------------------*/

void Child_OnPaint( HWND hWnd )
{
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rcClient;          // client area
    COLORREF cr;            // a color
    HBRUSH hBrush;          // a new brush
    HBRUSH hBrushOld;       // the default brush
    HPEN hPen;              // a new pen
    HPEN hPenOld;           // the default pen
    POINT point[3];         // points defining a triangle
    int cxClient;           // width of client area
    int cyClient;           // height of client area

    if( iShape )			// do not draw anything if iShape is zero
    {		// set the color according to the data read from the pipe
        switch( iColor )
        {
            case IDM_RED:	cr = RGB( 255, 0, 0 );	break;
            case IDM_GREEN:	cr = RGB( 0, 255, 0 );	break;
            default:		cr = RGB( 0, 0, 255 );	break;     // IDM_BLUE
        }
    }
    hDC = BeginPaint( hWnd, &ps );	// draw the figure
    if( hDC )
    {		// get the window's dimensions
        GetClientRect( hWnd, &rcClient );
        cxClient = (int)(rcClient.right - rcClient.left);
        cyClient = (int)(rcClient.bottom - rcClient.top);
			// set the pen and brush to the selected color
        hBrush = CreateSolidBrush( cr );
        if( hBrush )
            hBrushOld = SelectBrush( hDC, hBrush );
        hPen = CreatePen( PS_SOLID, 0, cr );
        if( hPen )
            hPenOld = SelectObject( hDC, hPen );

	        // If iShape is not zero draw either the large or
		    // small version of the figure.
        switch( iShape )
        {
            case IDM_ELLIPSE:
                if( iSize == IDM_LARGE )
                    Ellipse( hDC, rcClient.left, rcClient.top,
							 rcClient.right, rcClient.bottom );
                else
                    Ellipse( hDC, cxClient / 4, cyClient / 4,
                             3 * cxClient / 4, 3 * cyClient / 4 );
                break;

            case IDM_RECTANGLE:
                if( iSize == IDM_LARGE )
                    Rectangle( hDC, rcClient.left, rcClient.top,
							   rcClient.right, rcClient.bottom );
                else
                    Rectangle( hDC, cxClient / 4, cyClient / 4,
                               3 * cxClient / 4, 3 * cyClient / 4 );
                break;

            case IDM_TRIANGLE:	// use polygon function
                if( iSize == IDM_LARGE )
                {
                    point[0].x = cxClient / 2;
                    point[0].y = rcClient.top;
                    point[1].x = rcClient.right;
                    point[1].y = rcClient.bottom;
                    point[2].x = rcClient.left;
                    point[2].y = point[1].y;
                }
                else
                {
                    point[0].x = cxClient / 2;
                    point[0].y = cyClient / 4;
                    point[1].x = (3 * cxClient) / 4;
                    point[1].y = (3 * cyClient) / 4;
                    point[2].x = cxClient / 4;
                    point[2].y = point[1].y;
                }
                Polygon( hDC, point, 3 );
                break;

            default:
                break;
        }
			// free the resources
        if( hBrush )
            DeleteBrush( SelectBrush( hDC, hBrushOld ) );
        if( hPen )
            DeletePen( SelectPen( hDC, hPenOld ) );
    }
    EndPaint( hWnd, &ps );          // release the DC
    return;
}

/*------------------------------------------------------------------
        CHILD_ONCOMMAND
        Process commands from the menu.
  ------------------------------------------------------------------*/

void Child_OnCommand( HWND hWnd, int iCmd, HWND hWndCtl, UINT uCode )
{
    switch( iCmd )
    {
        case IDM_ABOUT:				// show the about dialog box
            DialogBox( hInst, MAKEINTRESOURCE(DLG_ABOUT),
                       hWnd, About_DlgProc );
            break;

        case IDM_EXIT:				// exit this program
            bTerminate = TRUE;
            DestroyWindow( hWnd );
            break;

        default:
            break;
    }
    return;
    UNREFERENCED_PARAMETER( hWndCtl );
    UNREFERENCED_PARAMETER( uCode );
}

/*------------------------------------------------------------------
        CHILD_ONDESTROY
        Post the quit message.
  ------------------------------------------------------------------*/

void Child_OnDestroy( HWND hWnd )
{
    PostQuitMessage(0);
    return;
    UNREFERENCED_PARAMETER( hWnd );
}

/*------------------------------------------------------------------
        PIPE THREAD
        The WM_CREATE handler starts a thread with this procedure
        to manage the pipe connection. This thread waits for messages
        to arrive through the pipe and acts on them.
  ------------------------------------------------------------------*/

LONG PipeThread ( LPVOID lpThreadData )
{
		// read from the named pipe until the terminate flag is set to true
    while( ! bTerminate )
        DoRead();
		// when bTerminate is TRUE, time to end program
    FORWARD_WM_DESTROY( hwndMain, PostMessage );
    return( 0L );                       // implicit ExitThread()
    UNREFERENCED_PARAMETER( lpThreadData );
}

/*------------------------------------------------------------------
        DO READ
        Read from the pipe and set the figure to be drawn.
  ------------------------------------------------------------------*/

void DoRead()
{
    FIGURE	Figure;
    DWORD	dwRead;
    BOOL	bTest;

		// read from the pipe
    bTest = ReadFile( hPipeRead,		// place to read from
					  &Figure,			// buffer to store input
					  sizeof(Figure),	// bytes to read
					  &dwRead,			// bytes read
					  NULL );
    if( bTest )							// the read command succeeded
    {
        if( Figure.iShape == IDM_TERMINATE )
		{
            bTerminate = TRUE;			// if Terminate, set flag to end thread
		}
		else
        {								// No? Thread continues; draw new shape
            iShape = Figure.iShape;		// copy the new shape attributes
            iColor = Figure.iColor;		// to global variables
            iSize  = Figure.iSize;
				// force the parent window to repaint itself
            InvalidateRect( hwndMain, NULL, TRUE );
            UpdateWindow( hwndMain );
        }
    }
    else								// the read command failed
    {
        ShowErrorMsg();					// tell user what happened
        bTerminate = TRUE;				// let the child end
    }
    return;
}

/*------------------------------------------------------------------
        ABOUT BOX DIALOG PROCEDURE
        Process messages for the About box window here.
  ------------------------------------------------------------------*/

BOOL WINAPI About_DlgProc( HWND hDlg,	  UINT uMessage,
						   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        case WM_INITDIALOG:
            return( TRUE );

        case WM_COMMAND:
            EndDialog( hDlg, TRUE );
            break;
    }
    return( FALSE );
    UNREFERENCED_PARAMETER( wParam );
    UNREFERENCED_PARAMETER( lParam );
}

/*--------------------------------------------------------------------
        SHOW ERROR MESSAGE
  --------------------------------------------------------------------*/

void ShowErrorMsg ( void )
{
    LPVOID lpvMessage;                      // temporary message buffer

		// retrieve a message from the system message table
    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				   NULL, GetLastError(),
				   MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
				   (LPTSTR)&lpvMessage, 0, NULL );
		// display the message in a message box
    MessageBox( hwndMain, lpvMessage, "Child Message",
                MB_ICONEXCLAMATION | MB_OK );
		// release the buffer FormatMessage allocated
    LocalFree( lpvMessage );
    return;
}
