//===================================================================
//	CHILD.C  [named pipe version]
//		Contains the child process for the PROCESS demo program.
//		The program in parent.c creates this child process.  In this
//		version, the two processes communicate through a named pipe.
//	FUNCTIONS
//		WinMain				receive and dispatch messages
//		Init				initialize program data
//		CreateMainWindow	creates program's overlapping window
//		Child_WndProc		sort out messages for main window
//		Child_OnCreate		create named pipe
//		Child_OnPaint		draw selected figure
//		Child_OnCommand		process menu commands
//		Child_OnDestroy		end program
//		PipeThread			thread procedure for managing the pipe
//		DoRead				procedure to read from pipe
//		About_DlgProc		sort out messages for About box
//		ShowErrorMsg		calls GetLastError and displays the message

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <string.h>                 // strcpy, strcat
#include "child.h"
#include "global.h"                 // values common to both processes

//===================================================================
//		GLOBAL VARIABLES

HWND        hwndMain;               // main window
HINSTANCE   hInst;                  // current instance
int         iShape;                 // figure to draw
int         iSize;                  // size of figure to draw (SMALL/LARGE)
int         iColor;                 // color of figure to draw
BOOL        bTerminate = FALSE;     // TRUE when program should end
char        szTitle[MAX_BUFFER];    // window caption string

//===================================================================
//		WIN MAIN
//		Calls initializing procedures and runs the message loop

int WINAPI WinMain( HINSTANCE hinstThis, HINSTANCE hinstPrev,
					LPSTR lpszCmdLine,	 int iCmdShow )
{
    MSG msg;

    hInst = hinstThis;
		// iShape = 0 do not draw anything during the first WM_PAINT message
    iShape = 0;
    if( ! Init() )        // if not initialized, exit here
        return( 0 );
		// display the main window 
    ShowWindow( hwndMain, iCmdShow );
    UpdateWindow( hwndMain );
		// translate the message
    while (GetMessage( &msg, NULL, 0, 0 ))
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }
    return( msg.wParam );
    UNREFERENCED_PARAMETER( hinstPrev );
    UNREFERENCED_PARAMETER( lpszCmdLine );
}

//===================================================================
//		INIT
//		Register the window class and the create the window

BOOL Init()
{
    WNDCLASS wc;
    char szAppName[MAX_BUFFER];

    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
		// register the main window class
    wc.lpszClassName    = szAppName;
    wc.hInstance        = hInst;
    wc.lpfnWndProc      = Child_WndProc;
    wc.hCursor          = LoadCursor(NULL, IDC_ARROW);
    wc.hIcon            = LoadIcon( hInst, MAKEINTRESOURCE(ICON_APP) );
    wc.lpszMenuName     = MAKEINTRESOURCE( MENU_MAIN );
    wc.hbrBackground    = (HBRUSH)( COLOR_WINDOW + 1 );
    wc.style            = CS_HREDRAW | CS_VREDRAW;
    wc.cbClsExtra       = 0;
    wc.cbWndExtra       = 0;
    if( ! RegisterClass( &wc ) )
        return( FALSE );
		// create the window
    return( CreateMainWindow() );
}

//===================================================================
//		CREATE MAIN WINDOW
//		Create the program's overlapping window.

BOOL CreateMainWindow()
{
    char szAppName[MAX_BUFFER];
    char szBuffer[MAX_BUFFER];
    char *pToken;

		// load the relevant strings
    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
		// Create window's caption using command line string from parent.
		// The basic caption has the form:
		//      Child 1 %s
		// The identifying number comes from the parent through the
		// command line.  We'll use wsprintf to insert the phrase
		// "waiting" or "connected" into the title as appropriate.
    strcpy( szTitle, "Child " );        // begin with "Child "
    strtok( GetCommandLine(), " " );    // move past first word
    pToken = strtok( NULL, " " );       // get first argument
    if( pToken )                        // is there one?
        strcat( szTitle, pToken );      // append it
    strcat( szTitle, " %s" );           // append a wsprintf format mark
		// The global szTitle now contains the base caption.
		// Insert a current status marker in it.
    wsprintf( szBuffer, szTitle, (LPSTR)" (waiting)" );
		// create the parent window
    hwndMain = CreateWindow( szAppName, szBuffer,
					         WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
							 CW_USEDEFAULT, CW_USEDEFAULT, 
							 CW_USEDEFAULT, CW_USEDEFAULT,
							 NULL, NULL, hInst, NULL );
		// return FALSE for an error
    return( hwndMain != NULL );
}

//===================================================================
//		CHILD_WNDPROC
//		This is where the messages for the main window are processed.

LRESULT WINAPI Child_WndProc( HWND hWnd,	 UINT uMessage,
						      WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        HANDLE_MSG( hWnd, WM_CREATE, Child_OnCreate );
        HANDLE_MSG( hWnd, WM_PAINT, Child_OnPaint );
        HANDLE_MSG( hWnd, WM_COMMAND, Child_OnCommand );
        HANDLE_MSG( hWnd, WM_DESTROY, Child_OnDestroy );
        default:
            return( DefWindowProc(hWnd, uMessage, wParam, lParam) );
    }
    return( 0L );
}

//===================================================================
//		CHILD_ONCREATE
//		On startup open the named pipe and start the thread that will
//		read from it.

BOOL Child_OnCreate( HWND hWnd, LPCREATESTRUCT lpCreateStruct )
{
    HANDLE hThread;                         // handle to new thread
    DWORD dwThreadID;                       // id of new thread

    // Create the secondary thread whose job it is to wait
    // for a pipe connection and then loop reading messages.
    // Make the thread be suspended at first so its priority
    // can be lowered before it starts.

    hThread = CreateThread( NULL,			// security attributes
							0,				// initial stack size
							(LPTHREAD_START_ROUTINE)PipeThread,
							NULL,				// argument
							CREATE_SUSPENDED,	// creation flag
							&dwThreadID );		// new thread's id
    if( ! hThread )
    {
        ShowErrorMsg( );
        return( FALSE );		// thread creation failed
    }
		// lower the thread's priority and let it run
    SetThreadPriority( hThread, THREAD_PRIORITY_BELOW_NORMAL );
    ResumeThread( hThread );
		// let go of the handle, for which we have no further use
    CloseHandle( hThread );
    return( TRUE );
    UNREFERENCED_PARAMETER( hWnd );
    UNREFERENCED_PARAMETER( lpCreateStruct );
}

//===================================================================
//		CHILD_ONPAINT
//		Redraw the window, adding the figure indicated by iShape. The
//		IDM_ constants used in this procedure are defined in global.h.
//		In the parent process they represent menu commands.  The
//		parent uses the constants to tell the child what to draw.

void Child_OnPaint( HWND hWnd )
{
    HDC hDC;                // device context
    PAINTSTRUCT ps;
    RECT rcClient;          // client area
    COLORREF cr;            // a color
    HBRUSH hBrush;          // a new brush
    HBRUSH hBrushOld;       // the default brush
    HPEN hPen;              // a new pen
    HPEN hPenOld;           // the default pen
    POINT point[3];         // points defining a triangle
    int cxClient;           // width of client area
    int cyClient;           // height of client area
    
		// do not draw anything if iShape is zero
    if( iShape )
    {		// set the color according to the data read from the pipe
        switch (iColor)
        {
            case IDM_RED:	cr = RGB(255, 0, 0);	break;
            case IDM_GREEN:	cr = RGB(0, 255, 0);	break;
            default:		cr = RGB(0, 0, 255);	break;	// IDM_BLUE
        }
    }
		// draw the figure
    hDC = BeginPaint( hWnd, &ps );
    if (hDC)
    {		// get the window's dimensions
        GetClientRect( hWnd, &rcClient );
        cxClient = (int)(rcClient.right - rcClient.left);
        cyClient = (int)(rcClient.bottom - rcClient.top);
			// set the pen and brush to the selected color
        hBrush = CreateSolidBrush( cr );
        if (hBrush)
            hBrushOld = SelectBrush( hDC, hBrush );
        hPen = CreatePen( PS_SOLID, 0, cr );
        if (hPen)
            hPenOld = SelectObject(hDC, hPen);
			// If iShape is not zero draw either the large or
			// small version of the figure.
        switch( iShape )
        {
            case IDM_ELLIPSE:
                if( iSize == IDM_LARGE )
                    Ellipse( hDC, rcClient.left, rcClient.top,
							 rcClient.right, rcClient.bottom );
                else
                    Ellipse( hDC, cxClient / 4, cyClient / 4,
                             (3 * cxClient) / 4, (3 * cyClient) / 4 );
                break;

            case IDM_RECTANGLE:
                if( iSize == IDM_LARGE )
                    Rectangle( hDC, rcClient.left, rcClient.top,
							   rcClient.right, rcClient.bottom );
                else
                    Rectangle( hDC, cxClient / 4, cyClient / 4,
                               (3 * cxClient) / 4, (3 * cyClient) / 4 );
                break;

            case IDM_TRIANGLE:
					// use the polygon function to draw the triangle
                if( iSize == IDM_LARGE )
                {
                    point[0].x = cxClient / 2;
                    point[0].y = rcClient.top;
                    point[1].x = rcClient.right;
                    point[1].y = rcClient.bottom;
                    point[2].x = rcClient.left;
                    point[2].y = point[1].y;
                }
                else
                {
                    point[0].x = cxClient / 2;
                    point[0].y = cyClient / 4;
                    point[1].x = (3 * cxClient) / 4;
                    point[1].y = (3 * cyClient) / 4;
                    point[2].x = cxClient / 4;
                    point[2].y = point[1].y;
                }
                Polygon( hDC, point, 3 );
                break;

            default:
                break;
        }
        if( hBrush )				// free the resources
            DeleteBrush( SelectBrush(hDC, hBrushOld) );
        if( hPen )
            DeletePen( SelectPen(hDC, hPenOld) );
    }
    EndPaint( hWnd, &ps );          // release the DC
    return;
}

//==================================================================
//		CHILD_ONCOMMAND
//		Process commands from the menu.

void Child_OnCommand( HWND hWnd, int iCmd, HWND hWndCtl, UINT uCode )
{
    switch (iCmd)
    {
        case IDM_ABOUT:		// show the about dialog box 
            DialogBox( hInst, MAKEINTRESOURCE(DLG_ABOUT),
                       hWnd, About_DlgProc );
            break;
        
        case IDM_EXIT:		// exit this program
            bTerminate = TRUE;
            DestroyWindow( hWnd );
            break;

        default:
            break;
    }
    return;
    UNREFERENCED_PARAMETER( hWndCtl );
    UNREFERENCED_PARAMETER( uCode );
}

//==================================================================
//		CHILD_ONDESTROY
//		Post the quit message.

void Child_OnDestroy( HWND hWnd )
{
    PostQuitMessage(0);

    return;
    UNREFERENCED_PARAMETER( hWnd );
}

//==================================================================
//		PIPE THREAD
//		The WM_CREATE handler starts a thread with this procedure
//		to manage the pipe connection.

LONG PipeThread( LPVOID lpThreadData )
{
    char szBuffer[MAX_BUFFER];      // used for pipe name and window caption
    int iLen;                       // length of a string
    BOOL bConnected;                // TRUE when we have a pipe handle
    HANDLE hpipeRead;               // receiving handle to 1-way pipe

		// load the string that contains the name of the named pipe
    iLen = LoadString( hInst, IDS_PIPE, szBuffer, sizeof(szBuffer) );
    if( ! iLen )
        return( FALSE );

    // This while loop continues until an instance of the named
    // pipe becomes available.  bConnected is a global variable,
    // and while it is FALSE the primary thread paints "Waiting..."
    // in the window's client area.

    // If an unpredictable error occurs, the loop sets the bTerminate
    // flag, this procedure ends quickly, and it kills the program
    // on its way out.

    bConnected = FALSE;
    while( ! bConnected && ! bTerminate )
    {		// wait for a pipe instance to become available
        WaitNamedPipe( szBuffer, NMPWAIT_WAIT_FOREVER );
			// open the named pipe for reading
        hpipeRead = CreateFile( szBuffer,				// name of pipe
								GENERIC_READ,			// access mode
								0,						// share mode
								NULL,					// security descriptor
								OPEN_EXISTING,			// do not create a new object
								FILE_ATTRIBUTE_NORMAL,	// file attributes
								NULL );					// file from which to copy attributes
			// check that the pipe's handle is valid
        if( hpipeRead == INVALID_HANDLE_VALUE )
        {		// If CreateFile failed simply because
				// other waiting threads grabbed the
				// pipe, don't bother the user with
				// an error message.
            if( GetLastError() != ERROR_PIPE_BUSY )
            {		// an unpredictable error occured; show message
                ShowErrorMsg( );
                bTerminate = TRUE;		// break loop; end program
            }
        }
        else
            bConnected = TRUE;			// succeeded in connecting
    }
		//	change window caption to show this window is connected
    wsprintf( szBuffer, szTitle, (LPSTR)" (connected)" );
    SetWindowText( hwndMain, szBuffer );
		//	read messages from the pipe until we receive a terminate command
    while( ! bTerminate )
        DoRead( hpipeRead );
		// when bTerminate is TRUE, end the program
    FORWARD_WM_DESTROY( hwndMain, PostMessage );
    return( 0L );                               // implicit ExitThread()
    UNREFERENCED_PARAMETER( lpThreadData );
}

//==================================================================
//		DO READ
//		Read from the pipe and set the figure to be drawn.

void DoRead( HANDLE hpipeRead )
{       
    FIGURE figure;
    DWORD dwRead;
    BOOL bTest;
    
		// read from the pipe
    bTest = ReadFile( hpipeRead,        // place to read from
					  &figure,			// buffer to store input
					  sizeof(figure),	// bytes to read
					  &dwRead,			// bytes read
 					  NULL );
    if( bTest )
    {                                   // the read command succeeded
        if( figure.iShape == IDM_TERMINATE )	// if Terminate
            bTerminate = TRUE;					// set flag to end this thread
        else									// else continues; draw new shape
        {		// copy the new shape attributes to global variables
            iShape = figure.iShape;
            iColor = figure.iColor;
            iSize  = figure.iSize;
				// force the parent window to repaint itself
            InvalidateRect( hwndMain, NULL, TRUE );
            UpdateWindow( hwndMain );
        }
    }
    else                                // the read command failed
    {
        ShowErrorMsg( );                // tell user what happened
        bTerminate = TRUE;              // let the child end
			// Those two commands cause all connected instances
			// of Child to raise a message box and quit if the user
			// strands them by exiting from the Parent program first.
    }
    return;
}
    
//==================================================================
//		ABOUT BOX DIALOG PROCEDURE
//		Process messages for the About box window here.

BOOL WINAPI About_DlgProc( HWND hDlg,	  UINT uMessage,
						   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        case WM_INITDIALOG:
            return( TRUE );

        case WM_COMMAND:
            EndDialog( hDlg, TRUE );
            break;
    }
    return( FALSE );
    UNREFERENCED_PARAMETER( wParam );
    UNREFERENCED_PARAMETER( lParam );
}

//==================================================================
//		SHOW ERROR MESSAGE

void ShowErrorMsg()
{
    LPVOID lpvMessage;			// temporary message buffer

		// retrieve a message from the system message table
    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				   NULL, GetLastError(),
				   MAKELANGID( LANG_ENGLISH, SUBLANG_ENGLISH_US ),
				   (LPTSTR) &lpvMessage, 0, NULL );
		// display the message in a message box
    MessageBox( hwndMain, lpvMessage, "Child Message",
                MB_ICONEXCLAMATION | MB_OK );
		// release the buffer FormatMessage allocated
    LocalFree( lpvMessage );
    return;
}
