//===================================================================
//
//	PARENT.C  [named pipe version]
//		Contains the parent process for the PROCESS demo program.
//		This version communicates with its child process through
//		a named pipe.
//	FUNCTIONS
//		WinMain             receive and dispatch messages
//		Init                initialize program data
//		CreateMainWindow    create main overlapping window
//		Parent_WndProc      sort out messages for main window
//		Parent_OnCreate     create the named pipe
//		Parent_OnInitMenu   update status of all menu commands
//		Parent_OnCommand    process menu commands
//		Parent_OnDestroy    end program
//		ChangeShape         store new shape choice
//		ChangeSize          store new size choice
//		ChangeColor         store new color choice
//		StartProcess        start child process
//		MakePipeInstance    create a pipe and a thread to run it
//		PipeInstanceThread  starting procedure for pipe-writing threads
//		SendCommand         tell the child process what to draw
//		About_DlgProc       sort out messages for About box
//		ShowErrorMsg        calls GetLastError and displays the message

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include "parent.h"
#include "global.h"

//===================================================================
//      GLOBAL VARIABLES

HWND        hwndMain;               // main overlapping window
HINSTANCE   hInst;                  // current instance
FIGURE      figure;                 // stores current menu choices
int         iNumConnections = 0;    // number of connected children
int         iPrevShape;             // records shape before terminating
HANDLE      hCmdEvent;              // signals command from user
HANDLE      hNotTerminatingEvent;   // signals termination in progress

//===================================================================
//		WIN MAIN
//		Calls initializing procedures and runs the message loop

int WINAPI WinMain( HINSTANCE hinstThis, HINSTANCE hinstPrev,
				    LPSTR lpszCmdLine,   int iCmdShow )
{
    MSG msg;

    hInst  = hinstThis;
	    // This global FIGURE structure records whatever choices
		// the user makes to choose the shape, size, and color
		// of the figure drawn in the client's window.  Here
		// we initialize it to the program's startup defaults.
    figure.iShape = IDM_RECTANGLE;		// draw a rectangle
    figure.iSize  = IDM_SMALL;			// don't fill the whole window
    figure.iColor = IDM_RED;			// make the rectangle red
    if( ! Init() )						// register/create window
        return( FALSE );				// initialization failed
		// show the main window
    ShowWindow( hwndMain, iCmdShow );
    UpdateWindow( hwndMain );
		// receive and translate messages
    while (GetMessage( &msg, NULL, 0, 0 ))
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }
    return( msg.wParam );
    UNREFERENCED_PARAMETER( hinstPrev );
    UNREFERENCED_PARAMETER( lpszCmdLine );
}

//===================================================================
//		INIT
//		Register the window classes and the create the windows.

BOOL Init()
{
    WNDCLASS wc;
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle) );
		// register the main window class
    wc.lpszClassName    = szAppName;
    wc.hInstance        = hInst;
    wc.lpfnWndProc      = Parent_WndProc;
    wc.hCursor          = LoadCursor( NULL, IDC_ARROW );
    wc.hIcon            = LoadIcon( hInst, MAKEINTRESOURCE(ICON_APP) );
    wc.lpszMenuName     = MAKEINTRESOURCE( MENU_MAIN );
    wc.hbrBackground    = (HBRUSH)(COLOR_WINDOW + 1);
    wc.style            = CS_HREDRAW | CS_VREDRAW;
    wc.cbClsExtra       = 0;
    wc.cbWndExtra       = 0;
    if( ! RegisterClass( &wc ) ) return( FALSE );
    return( CreateMainWindow() );
}

//===================================================================
//		CREATE MAIN WINDOW
//		Create the program's main overlapping window.

BOOL CreateMainWindow()
{
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];

		// load the relevant strings
    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName));
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle));
		// create the parent window                       
    hwndMain = CreateWindow( szAppName, szTitle,
							 WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
							 CW_USEDEFAULT, CW_USEDEFAULT, 
							 CW_USEDEFAULT, CW_USEDEFAULT,
							 NULL, NULL, hInst, NULL );
		// return FALSE for an error
    return( hwndMain != NULL );
}

//===================================================================
//		PARENT_WNDPROC
//		This is where the messages for the main window are processed.

LRESULT WINAPI Parent_WndProc( HWND hWnd,     UINT uMessage,
							   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        HANDLE_MSG( hWnd, WM_CREATE, Parent_OnCreate );
        HANDLE_MSG( hWnd, WM_INITMENU, Parent_OnInitMenu );
        HANDLE_MSG( hWnd, WM_COMMAND, Parent_OnCommand );
        HANDLE_MSG( hWnd, WM_DESTROY, Parent_OnDestroy );
        default:
            return( DefWindowProc(hWnd, uMessage, wParam, lParam) );
    }
    return( 0L );
}

//===================================================================
//		PARENT_ONCREATE
//		Create all the pipe instances and the two event objects used
//		to synchronize the program's threads

BOOL Parent_OnCreate( HWND hWnd, LPCREATESTRUCT lpcs )
{
    int i;
    int iNumInstances = 0;              // counts instances created

    // Create all the instances of the named pipe.  The
    // MakePipeInstance command also starts up a new
    // thread to service each pipe instance.

    for( i=1; i<=NUM_PIPE_INSTANCES; i++ )
    {
        if( MakePipeInstance() )        // make one instance
            iNumInstances++;            // if successful, increment counter
    }

    if( iNumInstances != NUM_PIPE_INSTANCES )  // did we make all of them?
    {
        char szBuffer[128];
        wsprintf( szBuffer, "Created only %i instances\n\r", iNumInstances );
        return( FALSE );                // creation failed
    }

    // Create the event object used for signaling the pipe
    // instance threads when the user makes a command.

    hCmdEvent = CreateEvent( NULL,		// default security attributes
							 TRUE,		// manual reset event
							 FALSE,		// initially not signalled
							 NULL );	// no name
    if( hCmdEvent == NULL )
    {
        ShowErrorMsg( );                // event creation failed
        return( FALSE );
    }
		// Create the event that coordinates the pipe threads when
		// the program terminates all linked children.  The threads
		// block on this event until all the clients have received
		// the IDM_TERMINATE message.
    hNotTerminatingEvent = CreateEvent( NULL,	// default security attributes
										TRUE,	// manual reset event
										TRUE,	// initially signalled
										NULL );	// no name

    if (hNotTerminatingEvent == NULL)
    {
        ShowErrorMsg( );                // event creation failed
        return( FALSE );
    }

    return( TRUE );
    UNREFERENCED_PARAMETER( hWnd );
    UNREFERENCED_PARAMETER( lpcs );
}

//===================================================================
//		PARENT_ONINITMENU
//		Check whether the child process exists and enable or disable
//		the Start, Terminate, and Options commands accordingly.  Also
//		put check marks on the option commands that reflect the
//		user's most recent choices.

void Parent_OnInitMenu( HWND hWnd, HMENU hMenu )
{
    // While the child process does not exist, some of our
    // menu commands make no sense and should be disabled.
    // These include the Terminate command and the figure
    // options commands.

		// get a handle to the options popup menu
    HMENU hmenuOptions = GetSubMenu( hMenu, 1 );

    if( iNumConnections > 0 )
    {		// a client is listening; enable terminate and shape commands
        EnableMenuItem( hMenu, IDM_TERMINATE, MF_ENABLED );
        EnableMenuItem( hmenuOptions, 0, MF_ENABLED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 1, MF_ENABLED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 2, MF_ENABLED | MF_BYPOSITION );
    }
    else
    {		// no client is listening; disable terminate and shape commands
        EnableMenuItem( hMenu, IDM_TERMINATE, MF_GRAYED );
        EnableMenuItem( hmenuOptions, 0, MF_GRAYED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 1, MF_GRAYED | MF_BYPOSITION );
        EnableMenuItem( hmenuOptions, 2, MF_GRAYED | MF_BYPOSITION );
    }

		// set a check mark on one of the three shape commands
    CheckMenuItem( hMenu, IDM_ELLIPSE, ((figure.iShape == IDM_ELLIPSE)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );
    CheckMenuItem( hMenu, IDM_RECTANGLE, ((figure.iShape == IDM_RECTANGLE)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );
    CheckMenuItem( hMenu, IDM_TRIANGLE, ((figure.iShape == IDM_TRIANGLE)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );

		// set a check mark on one of the two size commands
    CheckMenuItem( hMenu, IDM_SMALL, ((figure.iSize == IDM_SMALL)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );
    CheckMenuItem( hMenu, IDM_LARGE, ((figure.iSize == IDM_LARGE)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );

	    // set a check mark on one of the three color commands
    CheckMenuItem( hMenu, IDM_RED, ((figure.iColor == IDM_RED)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );
    CheckMenuItem( hMenu, IDM_GREEN, ((figure.iColor == IDM_GREEN)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );
    CheckMenuItem( hMenu, IDM_BLUE, ((figure.iColor == IDM_BLUE)
                    ? (int)MF_CHECKED : (int)MF_UNCHECKED) );

    return;
    UNREFERENCED_PARAMETER( hWnd );
}

//===================================================================
//		PARENT_ONCOMMAND
//		Process commands from the menu.

void Parent_OnCommand( HWND hWnd, int iCmd, HWND hWndCtl, UINT uCode )
{
    switch (iCmd)
    {
        case IDM_START:			// start the drawing process
            StartProcess();
            break;

        case IDM_ELLIPSE:		// send the figure data through the pipe
        case IDM_RECTANGLE:
        case IDM_TRIANGLE:
        case IDM_TERMINATE:		// when child sees TERMINATE it quits
            ChangeShape( iCmd );
            break;

        case IDM_SMALL:			// adjust the size of the figure
        case IDM_LARGE:
            ChangeSize( iCmd );
            break;

        case IDM_RED:			// change the color of the figure
        case IDM_BLUE:
        case IDM_GREEN:
            ChangeColor( iCmd );
            break;

        case IDM_ABOUT:			// show the about dialog box
            DialogBox( hInst, MAKEINTRESOURCE(DLG_ABOUT),
                       hWnd, About_DlgProc );
            break;
        
        case IDM_EXIT:			// exit from the program
            DestroyWindow( hWnd );      // destroy program window
            break;

        default:
            break;
    }
    
    return;
    UNREFERENCED_PARAMETER( hWndCtl );
    UNREFERENCED_PARAMETER( uCode );
}

//===================================================================
//		PARENT_ONDESTROY
//		Post the quit message.

void Parent_OnDestroy ( HWND hWnd )
{
    PostQuitMessage(0);             // terminate program

    return;
    UNREFERENCED_PARAMETER( hWnd );
}

//===================================================================
//		CHANGE SHAPE
//		Record a shape command from the user.  If the user has chosen
//		a new shape, send the updated FIGURE structure to the child.

void ChangeShape ( int iCmd )
{
    if (iCmd != figure.iShape)      // new shape?
    {
        // After sending a terminate command, we need to
        // restore the last shape drawn so that newly
        // connected clients can still draw whatever the
        // user last chose.

        if (iCmd == IDM_TERMINATE)
        {
            iPrevShape = figure.iShape; // save old shape command
        }

        figure.iShape = iCmd;       // record new shape command
        PulseEvent( hCmdEvent );    // tell threads shape has changed
    }
    return;
}

//===================================================================
//		CHANGE SIZE
//		Record a size command from the user.  If the user has chosen
//		a new size, send the updated FIGURE structure to the child.

void ChangeSize ( int iCmd )
{
    if (iCmd != figure.iSize)       // new size?
    {
        figure.iSize = iCmd;        // record it
        PulseEvent( hCmdEvent );    // tell threads shape has changed
    }
    return;
}

//===================================================================
//		CHANGE COLOR
//		Record a color command from the user.  If the user has chosen
//		a new color, send the updated FIGURE structure to the child.

void ChangeColor ( int iCmd )
{
    if (iCmd != figure.iColor)      // new color?
    {
        figure.iColor = iCmd;       // record it
        PulseEvent( hCmdEvent );    // tell threads shape has changed
    }
    return;
}

//===================================================================
//		START PROCESS
//		In response to the IDM_START command, create a new
//		child process.  The user may create any number of children.

void StartProcess()
{
    STARTUPINFO			sui;		// info for starting a process
    PROCESS_INFORMATION pi;			// info returned about a process
    static	int iChildNum = 1;		// counts child processes
    char	szProcess[MAX_BUFFER];	// name of child process image
    char	szCmdLine[MAX_BUFFER];	// child's command line
    int		iLen;					// return value from LoadString
    BOOL	bTest;					// return value

		// load name of process image file from resources 
    iLen = LoadString( hInst, IDS_PROCESS, szProcess, sizeof(szProcess) );
    if( ! iLen )
    {
        return;                     // loading string failed
    }
		// fill in the process's startup information
    sui.cb               = sizeof(STARTUPINFO);
    sui.lpReserved       = NULL;            // must be NULL
    sui.lpDesktop        = NULL;            // starting desktop
    sui.lpTitle          = NULL;            // title for new console window
    sui.dwX              = 0;               // window starting offsets
    sui.dwY              = 0;
    sui.dwXSize          = 0;               // window starting size
    sui.dwYSize          = 0;
    sui.dwXCountChars    = 0;               // console screen buffer size
    sui.dwYCountChars    = 0;
    sui.dwFillAttribute  = 0;               // console text colors
    sui.dwFlags          = 0;               // flags to activate startup fields
    sui.wShowWindow      = 0;               // iCmdShow parameter
    sui.cbReserved2      = 0;
    sui.lpReserved2      = NULL;
		// prepare child's command line argument, a window caption string
    wsprintf( szCmdLine, "%s %i", (LPSTR)szProcess, iChildNum++ );
		// create the drawing process
    bTest = CreateProcess( NULL,            // .EXE image
						   szCmdLine,		// command line
						   NULL,			// process security
						   NULL,			// thread security
						   FALSE,			// inherit handles
						   0,				// creation flags
						   NULL,			// environment block
						   NULL,			// current directory
						   &sui,			// startup info
						   &pi );			// process info (returned)
    if (!bTest)
    {
        ShowErrorMsg( );                    // creation failed
        return;
    }
    WaitForInputIdle( pi.hProcess, 5000 );  // wait for child to start up
    CloseHandle( pi.hProcess );             // we don't need the handles
    CloseHandle( pi.hThread );
    return;
}

//===================================================================
//		MAKE PIPE INSTANCE
//		Create a new instance of the named pipe.
//		Return	TRUE if the procedure creates a new instance; 
//				FALSE if an error prevents creation.

BOOL MakePipeInstance()
{
    char	szPipe[MAX_BUFFER];	// name of pipe
    HANDLE	hPipe;				// handle to new pipe
    HANDLE	hThread;			// handle to new thread
    DWORD	dwThreadID;			// id of new thread
    
		// get name to use for sharing pipe
    if( ! LoadString( hInst, IDS_PIPE, szPipe, sizeof(szPipe) ) )
		return( FALSE );
	    // Create a new instance of the named pipe.  This command
		// will fail if two instances already exist.
    hPipe = CreateNamedPipe( szPipe,				// name
							 PIPE_ACCESS_OUTBOUND,	// open mode
							 PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
							 NUM_PIPE_INSTANCES,	// max instances
							 0,						// out buffer size
							 0,						// in buffer size
							 0,						// time out value
							 NULL );				// security attributes
	if( hPipe == INVALID_HANDLE_VALUE )
    {
        ShowErrorMsg();						// creation failed
        return( FALSE );
    }
    hThread = CreateThread( NULL,				// security attributes
							0,					// initial stack size
							(LPTHREAD_START_ROUTINE)PipeInstanceThread,
							(LPVOID) hPipe,		// argument for thread proc
							CREATE_SUSPENDED,	// creation flag
							&dwThreadID );		// new thread's id
    if( ! hThread )
    {
        ShowErrorMsg();
        CloseHandle( hPipe );
        return( FALSE );                    // thread creation failed
    }
		// lower the thread's priority and let it run
    SetThreadPriority( hThread, THREAD_PRIORITY_BELOW_NORMAL );
    ResumeThread( hThread );
		// let go of the handle, for which we have no further use
    CloseHandle( hThread );
    return( TRUE );
}

//===================================================================
//		PIPE INSTANCE THREAD
//		MakePipeInstance starts a thread with this procedure
//		to manage a connection for one instance of the pipe.

	// Tell the compiler not to complain about the "while (TRUE)" loop
#pragma warning (disable :4127)

LONG PipeInstanceThread( HANDLE hPipe )
{
    BOOL bConnected;    // true when a client connects to the pipe

    // This loop runs endlessly.  When a client disappears, the
    // loop simply waits for a new client to connect.  This thread is 
    // terminated automatically when the program's primary thread exits.

    while( TRUE )
    {		// wait for a connection with some client
        ConnectNamedPipe( hPipe, NULL );        
			// If other threads are terminating their clients, then
			// figure.iShape still holds the IDM_TERMINATE command.
			// The thread blocks here until the last client is
			// terminated.  The last terminating thread resets
			// figure.iShape to IDM_RECTANGLE.
        WaitForSingleObject( hNotTerminatingEvent, INFINITE );
			// now the connection is made and a command message is ready
        iNumConnections++;			// update global variable
        SendCommand( hPipe );		// give client its first command
			// send another message each time the Command event signals
        bConnected = TRUE;
        while (bConnected)
        {
            WaitForSingleObject( hCmdEvent, INFINITE ); // wait for signal
            if( ! SendCommand( hPipe ) )      // send new shape command
                bConnected = FALSE;
					// The connection failed--probably
					// we just sent IDM_TERMINATE or
					// the user exited from the client.
					// Show no error message.
        }
        FlushFileBuffers( hPipe );          // wait for child to read message
        DisconnectNamedPipe( hPipe );       // break connection
        iNumConnections--;                  // update global variable

        // The following if condition coordinates threads when they
        // are all terminating their clients. When a thread discovers
        // it has just sent the IDM_TERMINATE command, it sets the
        // hNotTerminatingEvent object to the non-signaled state.
        // Other threads will block until the last thread to disconnect
        // restores the signal.  The last thread also replaces the
        // IDM_TERMINATE command with IDM_RECTANGLE so that all the
        // threads will have a useful command to send in the first
        // message to their new clients.

        if( figure.iShape == IDM_TERMINATE )
            if( iNumConnections > 0 )		// everything terminated?
            				// NO; block other threads
							// while terminating proceeds
                ResetEvent( hNotTerminatingEvent );
            else
            {
                figure.iShape = iPrevShape;         // restore previous command
                SetEvent( hNotTerminatingEvent );   // unblock threads
            }
    }
    return( 0L );
}

	// allow the "conditional expression constant" warning again
#pragma warning (default :4127)

//===================================================================
//		SEND COMMAND
//		Tell the child program what to draw.  Write the current
//		contents of the global FIGURE variable into the pipe.
//		Return	TRUE indicates the write operation succeeded; FALSE means
//				an error occurred and we have lost contact with the child.

BOOL SendCommand ( HANDLE hPipe )
{
    BOOL bTest;                     // return value
    DWORD dwWritten;                // number of bytes written to pipe

		// pass the choices to the child through the pipe
    bTest = WriteFile( hPipe,			// named pipe (outbound handle)
					   &figure,			// buffer to write
					   sizeof(FIGURE),	// size of buffer
					   &dwWritten,		// bytes written
					   NULL );			// overlapping i/o structure

    if( ! bTest )					// did writing succeed?
    {
        // If the write operation failed because the user has
        // already closed the child program, then we don't need
        // to do anything special about the error.  If, however,
        // some less predictable error caused the failure, call
        // ShowErrorMsg as usual to display the system's error
        // message.

        DWORD dwResult = GetLastError();

        if ((dwResult != ERROR_BROKEN_PIPE)     // pipe has been ended
            && (dwResult != ERROR_NO_DATA))     // pipe close in progress
        {
            ShowErrorMsg( );                    // unpredictable error
        }
    }

    // SendCommand returns FALSE on errors to indicate that the
    // connection has failed.  SendCommand also returns FALSE after
    // it tells a child to quit because that too makes a connection
    // fail.

    return( (bTest) && (figure.iShape != IDM_TERMINATE) );
}

//===================================================================
//		ABOUT BOX DIALOG PROCEDURE
//		Process messages for the About box window here.

BOOL WINAPI About_DlgProc( HWND hDlg,	  UINT uMessage,
						   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        case WM_INITDIALOG:
            return( TRUE );

        case WM_COMMAND:
            EndDialog( hDlg, TRUE );
            return( TRUE );
    }
    return( FALSE );
    UNREFERENCED_PARAMETER( wParam );
    UNREFERENCED_PARAMETER( lParam );
}

//===================================================================
//		SHOW ERROR MESSAGE

void ShowErrorMsg()
{
    LPVOID lpvMessage;                      // temporary message buffer

		// retrieve a message from the system message table
    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				   NULL, GetLastError(),
				   MAKELANGID( LANG_ENGLISH, SUBLANG_ENGLISH_US ),
				   (LPTSTR) &lpvMessage, 0, NULL );
		// display the message in a message box 
    MessageBox( hwndMain, lpvMessage, "Parent Message",
                MB_ICONEXCLAMATION | MB_OK );
		// release the buffer FormatMessage allocated
    LocalFree( lpvMessage );
    return;
}
