//====================================================================
//
//	LIST.C
//
//		Creates a dynamic list using Windows NT's virtual memory
//		management functions.
//
//	FUNCTIONS
//		WinMain				receive and dispatch messages
//		Init				initialize program data
//		CreateWindows		make program window and list box
//		List_WndProc		sort out messages for main window
//		List_OnInitMenu		enable/disable menu choices
//		List_OnSize			make list box fit new window size
//		List_OnSetFocus		set focus on list box control
//		List_OnCommand		process menu commands
//		List_OnDestroy		end program
//		CreateList			initialize list structures
//		AddItem				create new list item
//		DeleteItem			delete selected item
//		DeleteList			destroy list structures
//		GetItemText			get text from user for new item
//		GetPageBaseEntry	find first entry on same page as given entry
//		AdjustLookupTable	update global status variables after changes
//		Text_DlgProc		text-entry dialog box
//		Text_DoCommand		copy text from edit control to buffer
//		About_DlgProc		About box
//		ShowErrorMsg		display message from system table
//
//		from Mastering Windows NT Programming
//		copyright 1993 by Brian Myers & Eric Hamer
//

#define STRICT

#pragma warning( disable: 4001 )        // nonstandard comment ("//")
#pragma warning( disable: 4201 )        // unnamed struct/union warning
#pragma warning( disable: 4214 )        // bit field warning
#include <windows.h>
#pragma warning( default: 4201 )
#pragma warning( default: 4214 )

#include <windowsx.h>
#include <string.h>
#include "list.h"


//====================================================================
//		GLOBAL VARIABLES

HINSTANCE hInst;                // program instance
HWND hwndMain;                  // main program window
HWND hwndList;                  // list box window
char szListText[ITEM_SIZE];     // holds text user enters
char *pBase;                    // bottom of list's virtual space
int  iListLookup[MAX_ITEMS];    // matches list box index to array position
BOOL bInUse[MAX_ITEMS];         // marks array positions that are in use
BOOL bListEmpty = TRUE;         // Add/DeleteItem update these
BOOL bListFull = FALSE;         //   global status flags

//====================================================================
//	WIN MAIN
//		Calls initializing procedures and runs the message loop

int WINAPI WinMain( HINSTANCE hinstThis, HINSTANCE hinstPrev,
					LPSTR lpszCmdLine,	 int iCmdShow )
{
    MSG		msg;

    hInst = hinstThis;
		// initialize the application
    if( !Init() ) return( FALSE );			// initialization failed
    ShowWindow( hwndMain, iCmdShow );		// show main window
    UpdateWindow( hwndMain );
    while( GetMessage( &msg, NULL, 0, 0 ) )	// pull messages from queue
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }
    return( msg.wParam );
    UNREFERENCED_PARAMETER(lpszCmdLine);
    UNREFERENCED_PARAMETER(hinstPrev);
}

//====================================================================
//	INIT
//		Register the window class and call a procedure to create
//		the program window and its list box.

BOOL Init()
{
    WNDCLASS wc;
    char szAppName[MAX_BUFFER];

    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName) );
		// register the main window class
    wc.lpszClassName = szAppName;
    wc.hInstance     = hInst;
    wc.lpfnWndProc   = List_WndProc;
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
    wc.hIcon         = LoadIcon( hInst, MAKEINTRESOURCE(ICON_APP) );
    wc.lpszMenuName  = szAppName;
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    if( ! RegisterClass( &wc ) ) return( FALSE );
		// create the windows
    if( ! CreateWindows() )		 return( FALSE );
		// reserve memory for the list 
    return( CreateList() );
}

//====================================================================
//	CREATE WINDOWS
//		Create the parent window and the listbox window.

BOOL CreateWindows()
{
    char szAppName[MAX_BUFFER];
    char szTitle[MAX_BUFFER];
    HMENU hMenu;

		// load the relevant strings
    LoadString( hInst, IDS_APPNAME, szAppName,  sizeof(szAppName));
    LoadString( hInst, IDS_TITLE,   szTitle,    sizeof(szTitle));
		// create the parent window
    hMenu = LoadMenu( hInst, MAKEINTRESOURCE(MENU_MAIN) );
    hwndMain = CreateWindow( szAppName, szTitle, WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
							 CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
							 NULL, hMenu, hInst, NULL);
		// an error occurred
    if( ! hwndMain ) return( FALSE );
		// create the list box
    hwndList = CreateWindow( "LISTBOX", NULL,
							 WS_CHILD | WS_VISIBLE | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT,
							 0, 0, 0, 0, hwndMain, (HMENU) 1, hInst, NULL );
    return( hwndList != NULL );             // FALSE for errors
}

//====================================================================
//	LIST_WNDPROC
//		Process messages for the main window here.

LRESULT WINAPI List_WndProc( HWND hWnd,		UINT uMessage,
							 WPARAM wParam,	LPARAM lParam )
{
    switch (uMessage)
    {
        HANDLE_MSG( hWnd, WM_INITMENU,	List_OnInitMenu );
        HANDLE_MSG( hWnd, WM_SIZE,		List_OnSize );
        HANDLE_MSG( hWnd, WM_SETFOCUS,	List_OnSetFocus );
        HANDLE_MSG( hWnd, WM_COMMAND,	List_OnCommand );
        HANDLE_MSG( hWnd, WM_DESTROY,	List_OnDestroy );
        default:
            return( DefWindowProc(hWnd, uMessage, wParam, lParam) );
    }
    return( 0L );
}

//====================================================================
//	LIST_ONINITMENU
//		When the user begins to make a menu choice, update the
//		status of menu items.

void List_OnInitMenu( HWND hWnd, HMENU hMenu )
{
    if( bListFull )											// is the list full?
        EnableMenuItem( hMenu, IDM_ADD, MF_GRAYED );		// disable Add Item
    else													// the list is not full
        EnableMenuItem( hMenu, IDM_ADD, MF_ENABLED );		// enable Add Item

    if( bListEmpty )										// is the list empty?
        EnableMenuItem( hMenu, IDM_DELETE, MF_GRAYED );		// disable Delete Item
    else													// list is not empty
        EnableMenuItem( hMenu, IDM_DELETE, MF_ENABLED );	// enable Delete Item
    return;
    UNREFERENCED_PARAMETER(hWnd);
}

//====================================================================
//	LIST_ONSIZE
//		Adjust the list box to fit in the main window's client area.

void List_OnSize( HWND hWnd, UINT uState, int cxClient, int cyClient )
{
    if (hwndList)               // does list box window exist?
    {
        /* reposition the list box and select the first item */
        MoveWindow( hwndList, 0, 0, cxClient, cyClient, TRUE );
        ListBox_SetCurSel( hwndList, 0 );
    }

    return;
    UNREFERENCED_PARAMETER(uState);
    UNREFERENCED_PARAMETER(hWnd);
}

//====================================================================
//	LIST_ONSETFOCUS
//		When main window receives the input focus, shift it to the
//		list box so the user can reach it with the keyboard.

void List_OnSetFocus( HWND hWnd, HWND hwndOldFocus )
{
    SetFocus( hwndList );
    return;
    UNREFERENCED_PARAMETER(hWnd);
    UNREFERENCED_PARAMETER(hwndOldFocus);
}

//====================================================================
//	LIST_ONCOMMAND
//		Process menu commands.

void List_OnCommand( HWND hWnd, int iCmd, HWND hwndCtl, UINT uCode )
{
    switch( iCmd )
    {
        case IDM_ADD:		AddItem();		break;		// add item to list
        case IDM_DELETE:	DeleteItem();	break;		// delete selected item
        case IDM_ABOUT:                     // show the About box
							DialogBox( hInst, MAKEINTRESOURCE(DLG_ABOUT),
									   hWnd, About_DlgProc );
				            break;
        case IDM_EXIT:		DestroyWindow( hWnd );		// end program
							break;
        default:			break;
    }
    return;
    UNREFERENCED_PARAMETER(hwndCtl);
    UNREFERENCED_PARAMETER(uCode);
}

//====================================================================
//	LIST_ONDESTROY
//		Post the quit message

void List_OnDestroy( HWND hWnd )
{
    DeleteList();                          // release list structures
    PostQuitMessage(0);
    return;
    UNREFERENCED_PARAMETER(hWnd);
}

//====================================================================
//	CREATE LIST
//		Reserve the memory for the list and initialize the variables
//		for the list box and memory status.

BOOL CreateList()
{
    int i;

		// reserve one meg of memory address space
    pBase = VirtualAlloc( NULL,						// starting address (anywhere)
						  MAX_ITEMS * ITEM_SIZE,	// one megabyte
						  MEM_RESERVE,				// reserve; don't commit
						  PAGE_NOACCESS );			// can't be touched
    if( pBase == NULL )
    {
        ShowErrorMsg( __LINE__ );
        return( FALSE );
    }
		// initialize the status variables
    for( i=0; i<MAX_ITEMS; i++ )
    {
        bInUse[i] = FALSE;			// show no entries in use
        iListLookup[i] = 0;
    }
    bListEmpty = TRUE;				// update global flags
    bListFull = FALSE;
    return TRUE;
}

//====================================================================
//	ADD ITEM
//		Add a list entry.  MAX_ITEMS and ITEM_SIZE are both set
//		to 1024.  With 1024 bytes per entry there are four entries
//		per page.  With 1024 entries, the list uses a maximum of
//		256 pages.

void AddItem()
{
    char	szText[ITEM_SIZE];	// text for one item
    int		iIndex;				// position in array
    int		iPos;				// position in list box
    int		iCount;				// count of entries in list box
    BOOL	bDone;				// TRUE when free array entry found

		// determine the location of the first free entry
    bDone = FALSE;
    iIndex = 0;
    while( ( ! bDone ) && ( iIndex < MAX_ITEMS ) )
    {
        if( ! bInUse[iIndex] )			// is entry in use?
            bDone = TRUE;				// found an empty entry
        else
            iIndex++;					// advance to next entry
    }

		// ask user for new text string
    if( ! GetItemText(szText) ) return;
		// The try block copies the new text to an empty place
		// in the item array.  If that memory page is uncommitted,
		// lstrcpy raises an exception.  The exception filter
		// commits the page and the command continues.
    __try
    {		// put text in the item
        lstrcpy( &pBase[iIndex * ITEM_SIZE], szText );
    }
    __except( CommitMemFilter( GetExceptionCode(), iIndex ) )
    {
        // the filter does all the work
    }
    bInUse[iIndex] = TRUE;	// mark this entry as in use
    bListEmpty = FALSE;
	    // Add the new text to the list box.  The string is inserted
		// at iPos.  Update iListLookup[iPos] to indicate where in the
	    // item array the new entry was stored (iIndex).
    iCount = ListBox_GetCount( hwndList );
    iPos = ListBox_InsertString( hwndList, iCount, szText );
    iCount++;
    ListBox_SetCurSel( hwndList, iPos );
    iListLookup[iPos] = iIndex;
    if( iCount == MAX_ITEMS )	// did we fill the last place?
        bListFull = TRUE;
    return;
}

//====================================================================
//	COMMIT MEM FILTER
//		Exception-handling filter for AddItem.  If a page fault
//		occurs, this function attempts to commit the page.  If
//		it succeeds, the program returns to lstrcpy and proceeds
//		from there.  If it fails, the search for an exception
//		handler continues.

LONG CommitMemFilter( DWORD dwExceptCode,	// code identifying the exception
					  int  iIndex )			// array element where fault occurred
{
    LPVOID lpvResult;

		// If the exception was not a page fault, then refuse
		// to handle it.  Make the system keep looking for
		// an exception handler.

    if( dwExceptCode != EXCEPTION_ACCESS_VIOLATION )
        return( EXCEPTION_CONTINUE_SEARCH );

		// Try to commit the missing page.
    lpvResult = VirtualAlloc(
        &( pBase[iIndex * ITEM_SIZE] ),	// bottom of area to commit
        ITEM_SIZE,						// size of area to commit
        MEM_COMMIT,						// new status flag
        PAGE_READWRITE );				// protection status

    if( ! lpvResult )			// did allocation fail?
    		// if we can't commit the page then we can't handle the exception
        return( EXCEPTION_CONTINUE_SEARCH );
    
		// The missing page is now in place.  Tell the
	    // system to go back and try again.
    return( EXCEPTION_CONTINUE_EXECUTION );
}

//====================================================================
//	DELETE ITEM
//		Delete the item containing the text selected in the
//		listbox.

void DeleteItem()
{
    int iCurSel;        // position of current selection in list box
    int iPlace;         // position of current selection in item array
    int iStart;         // first position on same memory page as selection
    int i;              // loop variable
    BOOL bFree;         // TRUE if all 4 entries on one page are unused
    BOOL bTest;         // for testing return results

		// retrieve the memory offset of the currently selected entry
    iCurSel = ListBox_GetCurSel( hwndList );
    iPlace = iListLookup[iCurSel];

	__try
	{
			// zero out the deleted entry 
		ZeroMemory( &(pBase[iPlace * ITEM_SIZE]), ITEM_SIZE );
	}
    __except( CommitMemFilter( GetExceptionCode(), iCurSel ) )
    {
        // the filter does all the work
    }
		// mark this entry as free
    bInUse[iPlace] = FALSE;
		// Figure out which entry number is first on the current page.
		// If all four entries on the page are empty, we'll uncommit
		// the page to release memory.
    iStart = GetPageBaseEntry( iPlace );
    bFree = TRUE;
    for( i=0; i<4; i++ )			// check four entries
        if( bInUse[i + iStart] )	// in use?
            bFree = FALSE;			// page is not free

		// if a whole memory page is now unused, free it
    if( bFree )						// is page free?
    {								// YES; release it
        bTest = VirtualFree( &pBase[iStart * ITEM_SIZE],
							 ITEM_SIZE, MEM_DECOMMIT );
        if( ! bTest )
        {
			ShowErrorMsg( __LINE__ );
			ExitProcess( (UINT)GetLastError() );
        }
    }
		// update the listbox display and the lookup table array
    ListBox_DeleteString( hwndList, iCurSel );
    AdjustLookupTable( iCurSel );
		// check whether any entries remain in the list
    bListEmpty = TRUE;
    i = 0;
    while( ( i < MAX_ITEMS ) && bListEmpty )
			// if the item is in use then the list is not empty
        bListEmpty = ! bInUse[i++];
		// reposition the selection marker in the list box
    if( ! bListEmpty )
    {
        if( iCurSel )				// did we delete the first item?
        {                           // no; select item above deletion
            ListBox_SetCurSel( hwndList, iCurSel - 1 );
        }
        else						// deleted item was at top
        {							// select new top entry
            ListBox_SetCurSel( hwndList, iCurSel );
        }
    }
    return;
}

//====================================================================
//	GET PAGE BASE ENTRY
//		Given an index into the list, figure out which entry
//		is first on the same page.  To do this, find the first
//		integer which is divisible by four and is less than or
//		equal to iPlace.

int GetPageBaseEntry( int iPlace )
{
    while( iPlace % 4 )
        iPlace--;
    return( iPlace );
}

//====================================================================
//	ADJUST LOOKUP TABLE
//		When a list box entry is deleted, the array which matches
//		the list box entries and the memory offsets must be updated.
//		The iStart parameter gives the position in the list box
//		from which a string was just deleted.

void AdjustLookupTable( int iStart )
{
    int i;

		// This loop starts at the position where an entry
		// was just deleted and works down the list, scooting
		// lower items up one space to fill in the gap.

    for( i=iStart; i<MAX_ITEMS-1; i++)
        iListLookup[i] = iListLookup[i+1];
    iListLookup[MAX_ITEMS-1] = 0;
    return;
}

//====================================================================
//	DELETE LIST
//		Delete all the entries in the list and free the memory
//		it occupied.  Called when the program ends.

void DeleteList()
{
		// decommit the memory and then release the address space
		// NOTE: we must supply a size for the MEM_DECOMMIT operation
	if( ! VirtualFree( (void*)pBase, MAX_ITEMS*ITEM_SIZE, MEM_DECOMMIT ) )
		ShowErrorMsg( __LINE__ );

		// now release the memory from the base address -- no size required
	if( ! VirtualFree( (void*)pBase, 0, MEM_RELEASE ) )
		ShowErrorMsg( __LINE__ );

    return;
}

//====================================================================
//	GET ITEM TEXT
//		Display a dialog box where the user can enter the text
//		for a new list item.  Returns the number of bytes in
//		the new string.

int GetItemText( char *pszText )
{
    int iRet;               // return value from dialog (1=OK)
    int iLen;               // length of user's new text string

		// show text entry dialog box
    iRet = DialogBox( hInst, MAKEINTRESOURCE(DLG_TEXTENTRY),
					  hwndMain, Text_DlgProc );
    if( iRet )								// did user cancel?
    {
        lstrcpy( pszText, szListText );		// no; store text
        iLen = lstrlen( pszText );
    }
    else
    {
        pszText = NULL;						// yes
        iLen = 0;							// say string was empty
    }
		// return the number of bytes in the text string
    return( iLen );
}

//====================================================================
//	TEXT_DLGPROC
//		Process messages for the text dialog box here

BOOL WINAPI Text_DlgProc( HWND hDlg,	 UINT uMessage,
						  WPARAM wParam, LPARAM lParam )
{
    switch( uMessage )
    {
        case WM_INITDIALOG:	return( TRUE );
        case WM_COMMAND:	Text_DoCommand( hDlg, LOWORD(wParam) );
							return( TRUE );
    }
    return( FALSE );
    UNREFERENCED_PARAMETER(lParam);
}

//====================================================================
//	TEXT_ONCOMMAND
//		If the user has entered text, store it in the global
//		variable szListText.

void Text_DoCommand( HWND hDlg,	int iCmd )
{
    HWND hwndEdit;

		// the user hit Cancel, so just end the dialog
    if( iCmd == IDCANCEL )
    {
        EndDialog( hDlg, FALSE );
        return;
    }

		// the user hit OK, so see if the edit box has a string
    if( iCmd == IDOK )
    {
        hwndEdit = GetDlgItem( hDlg, IDD_EDIT );
        if( hwndEdit )
        {
				// if there is no text return false 
            if( Edit_GetTextLength( hwndEdit ) == 0 )
            {
                EndDialog( hDlg, FALSE );
                return;
            }
				// store the text in the global buffer
            Edit_GetText( hwndEdit, szListText, sizeof(szListText) );
            EndDialog( hDlg, TRUE );
        }
    }
    return;
}

//====================================================================
//	ABOUT_DLGPROC
//		Process messages for the About box here.

BOOL WINAPI About_DlgProc( HWND hDlg,	  UINT uMessage,
						   WPARAM wParam, LPARAM lParam )
{
    switch (uMessage)
    {
        case WM_INITDIALOG:	return( TRUE );
        case WM_COMMAND:	EndDialog( hDlg, TRUE );
							return( TRUE );
    }
    return( FALSE );
    UNREFERENCED_PARAMETER(wParam);
    UNREFERENCED_PARAMETER(lParam);
}

//====================================================================
//	SHOW ERROR MESSAGE
//		Receives line number where error occured and displays it
//		as part of a system error message describing the most
//		recent error.  Relies on a global variable, hwndMain.

void ShowErrorMsg( int iLine )
{
    LPVOID lpvMessage;                      // temporary message buffer
    char *pMsg;

		// retrieve a message from the system message table
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,                               // ignored
        GetLastError(),                     // message id
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),   // message language
        (LPTSTR)&lpvMessage,                // address of buffer pointer
        0,                                  // minimum buffer size
        NULL );                             // no other arguments
		// allocate a larger buffer and append the line number to the end
    pMsg = LocalAlloc( LHND, lstrlen(lpvMessage) + 20 );
    wsprintf( pMsg, "%s (line %u)", (LPSTR)lpvMessage, iLine );

		// display the message in a message box 
    MessageBox( NULL, pMsg, "List Message", MB_ICONEXCLAMATION | MB_OK );
		// release the buffers allocated by FormatMessage and by us
    LocalFree( lpvMessage );
    LocalFree( pMsg );
    return;
}
