/*--------------------------------------------------------------------

	ENCRYPT.CPP

	Demonstrates Using Window the Crypto API to encrypt a file. 
	Takes an input and output filename from the command line. 
		
    FUNCTIONS USED:

	CryptAcquireContext // Gets handle to CSP.
	CryptGetUserKey		// Gets our default exchange key.
	CryptGenKey			// Generates a random key
	CryptExportKey		// Encrypts our random key as blob.
	CryptEncrypt		// Encrypts the data from our input file.
	CryptDestroyKey		// Cleans up our keys when were done.
	CryptReleaseContext	// Releases our CSP handle when were done.

	CreateFile			// Opens the input and output files.
	ReadFile			// Reads "raw" data from input file.	
	WriteFile			// Writes encrypted data to output file.
	CloseHandle			// Closes input and output file when done.
		
  

		NOTE: This program will work on both Windows 98 and NT4.0
		(or later).

		written/updated by Jim Blaney, 1998
		for Windows 98 Developers Handbook

  -------------------------------------------------------------------*/
#define _WIN32_WINNT 0x0400
#include <windows.h>
#include <wincrypt.h>
#include <stdio.h>

const IN_BUFFER_SIZE    = 2048;
const OUT_BUFFER_SIZE   = IN_BUFFER_SIZE + 64; // extra padding.

int __cdecl main(int argc, char * argv[])
{
	HANDLE			hInFile, hOutFile;
	BYTE			pbBuffer[OUT_BUFFER_SIZE];
	BOOL			finished;
	HCRYPTPROV  	hProvider = 0;
	HCRYPTKEY		hKey = 0, hExchangeKey = 0;
	DWORD 			dwByteCount, dwBytesWritten;

	if (argc != 3) {
		printf("Usage: ENCRYPT infile outfile\n");
		exit(0);
	}

	// Get handle for the default provider (use RSA encryption).
	CryptAcquireContext(&hProvider, NULL, NULL, PROV_RSA_FULL, 0);

	// Open infile and create outfile.
	hInFile	= CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	hOutFile = CreateFile(argv[2], GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	// Generate a random key (in "simple blob" format) to use for encryption.
	CryptGetUserKey(hProvider, AT_KEYEXCHANGE, &hExchangeKey);
	CryptGenKey(hProvider, CALG_RC2, CRYPT_EXPORTABLE, &hKey);

   // The first call to ExportKey with NULL gets the key size (in dwByteCount).
   dwByteCount=0;
	CryptExportKey(hKey, hExchangeKey, SIMPLEBLOB, 0, NULL, &dwByteCount);
	CryptExportKey(hKey, hExchangeKey, SIMPLEBLOB, 0, pbBuffer, &dwByteCount);

	// Write size of keyblob, then keyblob itself, to output file.
	WriteFile(hOutFile, &dwByteCount, sizeof(dwByteCount), &dwBytesWritten, NULL);
	WriteFile(hOutFile, pbBuffer, dwByteCount, &dwBytesWritten, NULL);

	// Now, read data in, encrypt it, and write encrypted data to output file.
	do {
		ReadFile(hInFile, pbBuffer, IN_BUFFER_SIZE, &dwByteCount, NULL);
		finished = (dwByteCount < IN_BUFFER_SIZE);
      CryptEncrypt(hKey, 0, finished, 0, pbBuffer, &dwByteCount, OUT_BUFFER_SIZE);
		WriteFile(hOutFile, pbBuffer, dwByteCount, &dwBytesWritten, NULL);
	} while (!finished);

	// Cleanup: release handles, close files.
	CryptDestroyKey(hKey);
	CryptDestroyKey(hExchangeKey);
	CryptReleaseContext(hProvider, 0);
	CloseHandle(hInFile);
	CloseHandle(hOutFile);

	return(0);
}


 