/*--------------------------------------------------------------------

		FILEUSER.CPP

        Demonstrates Using Window NT's Security API to Set File Permissions.
		
        FUNCTIONS
			Check							// *simple* error checking 
			LookupAccountName				// The rest of these
			GetFileSecurity					// are functions in NT's
			InitializeSecurityDescriptor	// Security API, and are 
			GetSecurityDescriptorDacl		// explained further in the
			GetAclInformation				// accompanying book.
			InitializeAcl
			IsValidAcl
			AddAccessDeniedAce
			AddAccessAllowedAce
			SetSecurityDescriptorDacl
			SetFileSecurity
			
    
		written/updated by Jim Blaney, 1998
		for Windows 98 Developers Handbook

  -------------------------------------------------------------------*/
#define _WIN32_WINNT 0x0400
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

const FILE_SD_SIZE = 65536 + SECURITY_DESCRIPTOR_MIN_LENGTH; // Max. possible SD size for a file.

// Since all these Security functions return BOOLEAN values, this function serves
// as a "quick and dirty" error-checking. It is certainly not sufficient for
// release-code, but does well here for debugging.
void Check( BOOL Success, char *szFunction )
{
	if (Success) return;

	char szErrorMessage[250];

	FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), 
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &szErrorMessage, sizeof(szErrorMessage), NULL );

	printf("%s failed.\n Error = (%d) %s\n", szFunction, GetLastError(), szErrorMessage );
	exit(1);
}

// takes filename, username, and whether to allow or deny READ access to given file.
int main( int argc, char *argv[] )
{
	SID_NAME_USE			SIDType;
	SECURITY_INFORMATION	SecInfo;
	PSECURITY_DESCRIPTOR	pFileSD,   pNewFileSD;
	PACL					pFileDACL, pNewFileDACL;
	ACL_SIZE_INFORMATION	ACLSizeInfo;
	LPVOID					pTempACE; 

	char	*lpSystemName = NULL;
	char	szDomainName[250];
	char	UserSID[2048];

	DWORD	dwSIDLength		= sizeof( UserSID );
	DWORD	dwDomainLength	= sizeof( szDomainName );
	DWORD	dwFileSDLength	= FILE_SD_SIZE;
	DWORD	dwLengthNeeded, dwNewACLSize;
	BOOL	bContainsDACL, bDefaultedDACL, bAllowAccess;
	UINT	iAceNum;
	
	if (argc != 4) { // show usage...
		printf("usage: fileuser FILENAME USER + or -\n");
		exit(1);
	}

	pFileSD		= malloc( FILE_SD_SIZE );
	pNewFileSD	= malloc( FILE_SD_SIZE );

	// determine whether we'll ALLOW or DENY access, based on cmd-line parameter.
	bAllowAccess = (argv[3][0] == '+');
	
	// get the SID of the user (or group)
	Check( LookupAccountName((LPSTR) NULL, argv[2], UserSID, &dwSIDLength,
		szDomainName, &dwDomainLength, &SIDType ), "LookupAccountName");

	// get the file's current SD.
	Check( GetFileSecurity(argv[1], SecInfo, pFileSD, dwFileSDLength, 
		&dwLengthNeeded ), "GetFileSecurity");
   
	// initialize a new SD (larger to hold the new user's SID)
	Check( InitializeSecurityDescriptor( pNewFileSD, SECURITY_DESCRIPTOR_REVISION ),
		"InitializeSecurityDescriptor");

	// get the file's DACL (from the SD)
	Check( GetSecurityDescriptorDacl( pFileSD, &bContainsDACL, &pFileDACL, &bDefaultedDACL ),
		"GetSecurityDescriptorDacl");

	// if file currently has a DACL, get the ACL size info.
	if (bContainsDACL)
		Check( GetAclInformation( pFileDACL, &ACLSizeInfo, sizeof(ACL_SIZE_INFORMATION), 
			AclSizeInformation),"GetAclInformation");

	// calculate size we'll need for current ACL, plus new ACE. 
	dwNewACLSize = ACLSizeInfo.AclBytesInUse + (bAllowAccess? 
		sizeof( ACCESS_ALLOWED_ACE ) : sizeof( ACCESS_DENIED_ACE ))   
		+ GetLengthSid( UserSID ) - sizeof( DWORD );


	// allocate and initialize new ACL -- note: don't forget to initialize!
	pNewFileDACL = (PACL) malloc( dwNewACLSize );
	Check( InitializeAcl( pNewFileDACL, dwNewACLSize, ACL_REVISION2 ), "InitializeAcl");
	
	// Verify we've set up our DACL correctly.
	Check( IsValidAcl( pNewFileDACL ), "IsValidAcl");

	// if DENYING access rights, place the new ACE first in the new file DACL.
	if (!bAllowAccess)
		Check( AddAccessDeniedAce( pNewFileDACL, ACL_REVISION2, GENERIC_READ,
			&UserSID ), "AddAccessDeniedAce");

	// copy each existing ACE (if any) over to the new DACL
	if (bContainsDACL) { // if current file ACL contains a DACL...
		if (ACLSizeInfo.AceCount > 0) { // and if the ACL had one or more ACEs...
			for (iAceNum=0; iAceNum<ACLSizeInfo.AceCount; iAceNum++) { // for each ACE...
				// read ACE from current ACL, write to new ACL.
				Check( GetAce(pFileDACL,	iAceNum, &pTempACE ), "GetAce");
				Check( AddAce(pNewFileDACL, ACL_REVISION, MAXDWORD, pTempACE,
					((PACE_HEADER) pTempACE)->AceSize ), 
					"AddAce");
			}
		}
	}
	
	// if GRANTING access rights, place the new ACE here at end of the new file DACL.
	if (bAllowAccess)
		Check( AddAccessAllowedAce( pNewFileDACL, ACL_REVISION2, GENERIC_READ,
			&UserSID ), "AddAccessAllowedAce");
	
	// set the new DACL to the new SD
	Check( SetSecurityDescriptorDacl( pNewFileSD, TRUE, pNewFileDACL, FALSE ), 
		"SetSecurityDescriptorAcl");
	
	// set our new SD to the object.
	Check( SetFileSecurity( argv[1], DACL_SECURITY_INFORMATION, pNewFileSD ), 
		"SetFileSecurity");

	// finally! cleanup, prepare to leave.
	free( pNewFileDACL );
	free( pNewFileSD );
	free( pFileSD );
	
	return(0);
}
	




	





