/*--------------------------------------------------------------------

	FTPGET.CPP

	Demonstrates using Win32 Internet API functions to perform FTP file 
   download. Takes two command-line parameters as follows:

      ftp.domain.com    \path\filename

   Connects to the supplied ftp domain, changes to specified directory on
   remote server, and downloads the specified file to current directory, 
   then disconnects and exits.
   	
   
     FUNCTIONS USED:

   LastErrorMsg            Returns last ftp error information. 
   InternetOpen            Initializes Internet API, returns handle.
   InternetConnect         Connects to a server and returns handle.
   FtpSetCurrentDirectory  Sets the current (remote) directory.
   FtpGetFile              Downloads file using FTP.
   InternetCloseHandle     Closes handles created by Internet API functions.

   
   NOTES: 
   - This program supports both Windows 95 and NT 4.0, but requires the
     presence of a copy of WININET.DLL (preferably in your \system32 or
     \system directory).
   - For non-anonymous FTP servers, you will need to change the 
     szUsername and szPassword global variables.

		written/updated by Jim Blaney, 1998
		for Windows 98 Developers Handbook

  -------------------------------------------------------------------*/

#include <windows.h>
#include <iostream.h>
#include <stdio.h>
#include <wininet.h>

// Customize login info as appropriate. 
char  szUsername[64]    = "anonymous";
char  szPassword[64]    = "user@domain.com";

// General Globals
char     szAppName[20]  = "FtpGet";
DWORD    dwBinary       = FTP_TRANSFER_TYPE_BINARY;

// Application exit codes.
const int   ERR_SUCCESS              = 0;
const int   ERR_USAGE                = 1;
const int   ERR_CONNECTING_INTERNET  = 2;
const int   ERR_CONNECTING_HOST      = 3;
const int   ERR_BAD_DIRECTORY        = 4;
const int   ERR_FTP_TRANSFER         = 5;


char *LastErrorMsg(void)
{
   static char szErrorMsg[4096];
   ULONG       dwErrorNum, dwLength;
   
   dwLength = sizeof( szErrorMsg );
   if (!InternetGetLastResponseInfo( &dwErrorNum, (LPSTR) szErrorMsg, &dwLength))
   {
      sprintf( szErrorMsg, "Unable to get error information, Error# = %d ", 
      GetLastError() );
   }
   return( szErrorMsg );
}


int main( int argc, char *argv[] )
{
   HINTERNET   hInternet, hHost; // handles.
   int         i;
   char        szHost[128];
   char        szPath[MAX_PATH];
   char        szFile[MAX_PATH];

   if (argc < 3) 
   {
      cout << "Usage: FtpGet ftp.domain.com \\path\\filename (or /path/filename)\n";
      return ERR_USAGE;
   }
   

   lstrcpy( szHost, argv[1] );
   lstrcpy( szFile, argv[2] );
   
   lstrcpy(szPath, "/"); // Use root as default path.
   // Get the path, if prepended to filename.
   for (i=strlen(szFile); i>0; --i)
   {
      if (szFile[i]=='/' || szFile[i]=='\\')
      {
         strncpy(szPath, szFile, ++i);
         szPath[i]= 0; // terminate.
         lstrcpy(szFile, szFile+i);
         break;
      }
   }
   
   hInternet = InternetOpen( szAppName, LOCAL_INTERNET_ACCESS , NULL, 0, 0 );
   if (!hInternet) 
   {
      cerr << "Error opening Internet: " << LastErrorMsg();
      InternetCloseHandle( hInternet );
      return ERR_CONNECTING_INTERNET;
   }

   hHost = InternetConnect( hInternet, szHost, INTERNET_INVALID_PORT_NUMBER, 
           szUsername,  szPassword, INTERNET_SERVICE_FTP, INTERNET_FLAG_PASSIVE , 0);
   if (!hHost)
   {
      cerr << "Error connecting to " << szHost << ": " << LastErrorMsg();
      InternetCloseHandle( hInternet );
      return ERR_CONNECTING_HOST;
   }
 
   if (!FtpSetCurrentDirectory( hHost, szPath ))
   {
      cerr << "Unable to find remote directory : " << szPath << endl;
      cerr << LastErrorMsg() << endl;
      InternetCloseHandle( hInternet );
      return ERR_BAD_DIRECTORY;
   }

   if (!FtpGetFile( hHost, szFile, szFile, FALSE, INTERNET_FLAG_RELOAD, dwBinary, 0 ))
   {
      cerr << "Unable to transfer file : " << szFile << endl; 
      cerr << LastErrorMsg() << endl;
      InternetCloseHandle( hInternet );
      return ERR_FTP_TRANSFER;
   }
 
   cout << "Transferred " << szFile << endl;
 
   // Finished, clean up and exit.
   InternetCloseHandle( hHost );
   InternetCloseHandle( hInternet );

   return ERR_SUCCESS;
}

 