        title   VFFT.ASM:  Perform Forward and Inverse FFT
        page    55,132

; Call with:    vfft0 (float* a, int str_a, float* b, int str_b, int n, int opt);
;
; INPUT:
; a   - array containing the real component of input
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of input
;       length = 2**n, stride = str_b
; n   - log, base 2, of number of samples to be transformed
; opt - options parameter       
;       inverse transform       if bit 0 is set
;       normalize               if bit 1 is set
;       skip bit-reversal step  if bit 8 is set
;
; OUTPUT:
; a   - array containing the real component of output
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of output
;       length = 2**n, stride = str_b
;
; Returns:      a [i*str_a] = real (fft (a [i*str_a], b [i*str_b])) for i = 0 to n-1
;               b [i*str_b] = imag (fft (a [i*str_a], b [i*str_b])) 
; 
; ERROR CODES:
; 0      - no error
; 0x01xx - error detected during butterfly
; 0x02xx - error detected during bit-reversal of real component
; 0x03xx - error detected during bit-reversal of imag component
;          (xx is error code returned from butterfly or bit-reverse routine)
;
; Destroys:     nothing
;
;
;
;
; Call with:    vfft1 (float* a, int str_a, float* b, int str_b, int n, int opt);
;
; INPUT:
; a   - array containing the even elements of real valued input
;       length = 2**n, stride = str_a
; b   - array containing the odd elements of real valued input
;       length = 2**n, stride = str_b
; n   - log, base 2, of number of samples to be transformed
; opt - options parameter       
;       inverse transform       if bit 0 is set
;       normalize               if bit 1 is set
;
; OUTPUT:
; a   - array containing the real component of output
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of output
;       length = 2**n, stride = str_b
; 
; ERROR CODES:
; 0      - no error
; 0x01xx - error detected during butterfly
; 0x02xx - error detected during bit-reversal of real component
; 0x03xx - error detected during bit-reversal of imag component
; 0x04xx - error detected during butterfly reordering routine
;          (xx is error code returned from butterfly or bit-reverse routine)
;
;
; Returns:      a [i*str_a] = real (fft (a [i*str_a], b [i*str_b])) for i = 0 to n-1
;               b [i*str_b] = imag (fft (a [i*str_a], b [i*str_b])) 
;
; Destroys:     nothing
;
;
;
;
; Call with:    vbfly0 (float* a, int str_a, float* b, int str_b, int n, int opt);
;
; INPUT:
; a   - array containing the real component of input
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of input
;       length = 2**n, stride = str_b
; n   - log, base 2, of number of samples to be transformed
; opt - options parameter       
;       inverse transform       if bit 0 is set
;       normalize               if bit 1 is set
;
; OUTPUT:
; a   - array containing the real component of output
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of output
;       length = 2**n, stride = str_b
; 
; ERROR CODES:
; 0      - no error
;
;
; Returns:      a [i*str_a] = real (butterfly (a [i*str_a], b [i*str_b])) for i = 0 to n-1
;               b [i*str_b] = imag (butterfly (a [i*str_a], b [i*str_b])) 
;
; Destroys:     nothing
;
;
;
;
; Call with:    vbfly1 (float* a, int str_a, float* b, int str_b, int n, int opt);
;
; INPUT:
; a   - array containing the even elements of real valued input
;       length = 2**n, stride = str_a
; b   - array containing the odd elements of real valued input
;       length = 2**n, stride = str_b
; n   - log, base 2, of number of samples to be transformed
; opt - options parameter       
;       inverse transform       if bit 0 is set
;       normalize               if bit 1 is set
;
; OUTPUT:
; a   - array containing the real component of output
;       length = 2**n, stride = str_a
; b   - array containing the imaginary component of output
;       length = 2**n, stride = str_b
; 
; ERROR CODES:
; 0      - no error
;
;
; Returns:      a [i*str_a] = real (butterfly (a [i*str_a], b [i*str_b])) for i = 0 to n-1
;               b [i*str_b] = imag (butterfly (a [i*str_a], b [i*str_b])) 
;
; Destroys:     nothing
;
;
;
;
; Call with:    vbitrev (float* a, int str_a, int n);
;
; INPUT:
; a   - array containing input
;       length = 2**n, stride = str_a
; n   - log, base 2, of number of samples to be transformed
;
; OUTPUT:
; a   - array containing output 
;       length = 2**n, stride = str_a
; 
; ERROR CODES:
; 0   - no error
;
;
; Returns:      a [i*str_a] = a [bit-reverse (i)*str_a]; for i = 0 to n-1
;
; Destroys:     nothing
;
;*******************************************************************************
; NOTE:  The agortithms in this file are impementations of algorithms found in
;        "Numerical Recipes in C" and "Digital Signal Processing" by Proakis and
;        Manolakis.  Refer to those books to aid in understanding what is done 
;        here.
;*******************************************************************************

.286
.287

;.MODEL SMALL
.MODEL LARGE

.DATA


tweedle   dq      0bff0000000000000r
          dq      00000000000000000r

          dq      00000000000000000r
          dq      0bff0000000000000r

          dq      03fe6a09e667f3bcdr
          dq      0bfe6a09e667f3bccr

          dq      03fed906bcf328d46r
          dq      0bfd87de2a6aea963r

          dq      03fef6297cff75cb0r
          dq      0bfc8f8b83c69a60ar

          dq      03fefd88da3d12526r
          dq      0bfb917a6bc29b42cr

          dq      03feff621e3796d7er
          dq      0bfa91f65f10dd814r

          dq      03feffd886084cd0dr
          dq      0bf992155f7a3667er

          dq      03fefff62169b92dbr
          dq      0bf8921d1fcdec784r

          dq      03fefffd8858e8a92r
          dq      0bf7921f0fe670071r

          dq      03feffff621621d02r
          dq      0bf6921f8becca4bar

          dq      03feffffd88586ee6r
          dq      0bf5921faaee6472dr

          dq      03fefffff62161a34r
          dq      0bf4921fb2aecb360r

          dq      03fefffffd8858675r
          dq      0bf3921fb49ee4ea6r

          dq      03feffffff621619cr
          dq      0bf2921fb51aeb57br

          dq      03feffffffd885867r
          dq      0bf1921fb539ecf31r

.CODE

          PUBLIC  _vbitrev__FPfii, _vbfly0__FPfiPfiii, _vbfly1__FPfiPfiii
          PUBLIC                    _vfft0__FPfiPfiii,  _vfft1__FPfiPfiii

F0_OPTN   equ     bp+6
F0_POWR   equ     bp+8
F0_BSTR   equ     bp+10
F0_BADD   equ     bp+12
F0_ASTR   equ     bp+16
F0_AADD   equ     bp+18

_vfft0__FPfiPfiii PROC
          enter   0, 0                    ; Entry sequence - save old BP

          push    es                      ; Save segments
          push    ds

          pusha                           ; Save registers and flags
          pushf

          lds     si, [F0_AADD]           ; Fetch parameters into registers
          les     di, [F0_BADD]

          mov     ax, [F0_OPTN]
          mov     bx, [F0_POWR]
          mov     cx, [F0_ASTR]
          mov     dx, [F0_BSTR]

          xor     ah, ah                  ; Clear flags in upper byte

          push    ds                      ; Push parameters for butterfly
          push    si
          push    cx
          push    es
          push    di
          push    dx
          push    bx
          push    ax

          call far ptr _vbfly0__FPfiPfiii ; Perform butterfly

          mov     ah, 01h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf0

          test    word ptr [F0_OPTN], 0100h ; Bypass bit reverse reordering?
          jnz     f0_a

          push    ds                      ; Push parameters for real bit reverse
          push    si
          push    cx
          push    bx

          call far ptr _vbitrev__FPfii    ; Perform bit reversal

          mov     ah, 02h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf0

          push    es                      ; Push parameters for imag bit reverse
          push    di
          push    dx
          push    bx

          call far ptr _vbitrev__FPfii    ; Perform bit reversal

          mov     ah, 03h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf0

f0_a:     xor     ax, ax                  ; Indicate no error

exitf0:   mov     ds, ax                  ; Save error code

          popf
          popa

          mov     ax, ds                  ; Fetch error code

          pop     ds                      ; Restore segments
          pop     es

          leave

          RET     16                      ; Return

_vfft0__FPfiPfiii ENDP

F1_OPTN   equ     bp+6
F1_POWR   equ     bp+8
F1_BSTR   equ     bp+10
F1_BADD   equ     bp+12
F1_ASTR   equ     bp+16
F1_AADD   equ     bp+18

_vfft1__FPfiPfiii PROC
          enter   0, 0                    ; Entry sequence - save old BP

          push    es                      ; Save segments
          push    ds

          pusha                           ; Save registers and flags
          pushf

          lds     si, [F1_AADD]           ; Fetch parameters into registers
          les     di, [F1_BADD]

          mov     ax, [F1_OPTN]
          mov     bx, [F1_POWR]
          mov     cx, [F1_ASTR]
          mov     dx, [F1_BSTR]

          xor     ah, ah                  ; Clear flags in upper byte

          test    ax, 0001h               ; Determine if inverse transform
          jz      f1_a

          push    ds                      ; Push parameters for single butterfly
          push    si
          push    cx
          push    es
          push    di
          push    dx
          push    bx
          push    ax

          call far ptr _vbfly1__FPfiPfiii ; Perform butterfly

          mov     ah, 04h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf1

          mov     ax, [F1_OPTN]

f1_a:     push    ds                      ; Push parameters for main butterfly
          push    si
          push    cx
          push    es
          push    di
          push    dx
          push    bx
          push    ax

          call far ptr _vbfly0__FPfiPfiii ; Perform butterfly

          mov     ah, 01h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf1

          push    ds                      ; Push parameters for real bit reverse
          push    si
          push    cx
          push    bx

          call far ptr _vbitrev__FPfii    ; Perform bit reversal

          mov     ah, 02h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf1

          push    es                      ; Push parameters for imag bit reverse
          push    di
          push    dx
          push    bx

          call far ptr _vbitrev__FPfii    ; Perform bit reversal

          mov     ah, 03h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf1

          mov     ax, [F1_OPTN]

          test    ax, 0001h               ; Determine if forward transform
          jnz     f1_b

          push    ds                      ; Push parameters for single butterfly
          push    si
          push    cx
          push    es
          push    di
          push    dx
          push    bx
          push    ax

          call far ptr _vbfly1__FPfiPfiii ; Perform butterfly

          mov     ah, 04h                 ; Did an error occur?
          cmp     al, 0
          jnz     exitf1

f1_b:     xor     ax, ax                  ; Indicate no error

exitf1:   mov     ds, ax                  ; Save error code

          popf
          popa

          mov     ax, ds                  ; Fetch error code

          pop     ds                      ; Restore segments
          pop     es

          leave

          RET     16                      ; Return

_vfft1__FPfiPfiii ENDP



BR_TOFF   equ     bp-10
BR_TSEG   equ     bp-8
BR_TEMP   equ     bp-6
BR_ELEM   equ     bp-2

BR_POWR   equ     bp+6
BR_ASTR   equ     bp+8
BR_AADD   equ     bp+10

_vbitrev__FPfii PROC
          enter   10, 0                   ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds
     
          shl     word ptr [BR_ASTR], 2   ; Compute stride

          cld                             ; Make sure SI and DI are incremented

          mov     word ptr [BR_TSEG], ss

          lea     ax, [BR_TEMP]
          mov     [BR_TOFF], ax

          mov     cx, [BR_POWR]           ; Load CX with index of last element
          mov     ax, 1
          shl     ax, cl
          mov     [BR_ELEM], ax

bitrev1:  dec     word ptr [BR_ELEM]      ; Compute index of current element

          xor     ax, ax                  ; Initialize registers for loop
          mov     bx, [BR_ELEM]
          mov     cx, [BR_POWR]

bitrev2:  rcr     bx, 1                   ; Reverse bits
          rcl     ax, 1
          loop    bitrev2

          cmp     ax, [BR_ELEM]           ; Bypass if no need to swap

          jge     bitrev3
          lds     bx, [BR_AADD]           ; Get address of first element

          mul     word ptr [BR_ASTR]
          add     ax, bx
          mov     si, ax                  ; Load SI with offset of bit-rev elem

          les     di, [BR_TOFF]           ; Get address of temp storage

          movsw                           ; Save first pair of bytes
          movsw                           ; Save second pair of bytes

          mov     di, ax                  ; Load DI with offset of bit-rev elem

          les     bx, [BR_AADD]           ; Get address of first element

          mov     ax, [BR_ELEM]
          mul     word ptr [BR_ASTR]
          add     ax, bx
          mov     si, ax                  ; Load SI with offset of non-rev elem

          movsw                           ; Move first pair of bytes
          movsw                           ; Move second pair of bytes

          mov     di, ax                  ; Load DI with offset of non-rev elem

          lds     si, [BR_TOFF]           ; Get address of temp storage

          movsw                           ; Finish first pair of bytes
          movsw                           ; Finish second pair of bytes

bitrev3:  cmp     word ptr [BR_ELEM], 0   ; Loop over all elements
          ja      bitrev1

          mov     word ptr [BR_ELEM], 0   ; Set return code

exitbr:   pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, word ptr [BR_ELEM]  ; Fetch error code

          leave

          RET     8                       ; Return

_vbitrev__FPfii ENDP



B0_OFFA   equ     bp-20
B0_OFFB   equ     bp-18
B0_OFF1   equ     bp-16
B0_OFF2   equ     bp-14
B0_DST2   equ     bp-12
B0_DST1   equ     bp-10
B0_TWDL   equ     bp-8
B0_PAIR   equ     bp-4
B0_SETS   equ     bp-2

B0_OPTN   equ     bp+6
B0_POWR   equ     bp+8
B0_BSTR   equ     bp+10
B0_BADD   equ     bp+12
B0_ASTR   equ     bp+16
B0_AADD   equ     bp+18

_vbfly0__FPfiPfiii PROC
          enter   20, 0                   ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds

          mov     ax, SEG DGROUP          ; Get segment of tweedle constants
          mov     [B0_TWDL+2], ax

          mov     ax, [B0_POWR]           ; Compute offset of tweedle constants
          dec     ax
          shl     ax, 4
          add     ax, offset tweedle
          mov     [B0_TWDL], ax
     
          shl     word ptr [B0_ASTR], 2   ; Compute strides 
          shl     word ptr [B0_BSTR], 2

          mov     ax, 1
          mov     [B0_SETS], ax

          mov     cx, word ptr [B0_POWR]  ; Compute number of pairs in a set
          dec     cx                      ;  for first butterfly
          shl     ax, cl
          mov     [B0_PAIR], ax

          mul     word ptr [B0_ASTR]      ; Compute distance between pairs of
          mov     [B0_DST1], ax           ;  real elements

          mov     ax, [B0_PAIR]
          mul     word ptr [B0_BSTR]      ; Compute distance between pairs of
          mov     [B0_DST2], ax           ;  imag elements

b0_a:     mov     dx, [B0_PAIR]           ; Set counter for outer loop

          les     bx, [B0_TWDL]           ; Load twiddle factor
          fld     qword ptr es:[bx]
          add     bx, 8
          fld     qword ptr es:[bx]

          mov     ax, [B0_OPTN]           ; Determine if inverse is requested
          and     ax, 0001
          jz      b0_b

          fchs                            ; Change sign  

b0_b:     mov     ax, [B0_DST1]           ; Load strides
          mov     bx, [B0_DST2]

          lds     si, [B0_AADD]           ; Point to first elements of real
          les     di, [B0_BADD]           ;  and imag array

          fld1                            ; Load initial constants
          fldz

          mov     [B0_OFFA], si           ; Init starting pointers for sets
          mov     [B0_OFFB], di

b0_c:     mov     cx, [B0_SETS]           ; Set counter for inner loop

          mov     [B0_OFF1], si           ; Initialize current real pointer
          add     si, [B0_DST1]

          mov     [B0_OFF2], di           ; Initialize current imag pointer
          add     di, [B0_DST2]       

b0_d:     fld     dword ptr ds:[si]
          xchg    si, word ptr [B0_OFF1]  ; Swap real offset

          fld     st(0)
          fld     dword ptr ds:[si]

          fsubr   st(2), st(0)
          fadd
          fstp    dword ptr ds:[si]

          fld     dword ptr es:[di]
          xchg    di, word ptr [B0_OFF2]  ; Swap imag offset

          fld     st(0)
          fld     dword ptr es:[di]

          fsubr   st(2), st
          fadd
          fstp    dword ptr es:[di]

          xchg    si, word ptr [B0_OFF1]  ; Swap real offset
          xchg    di, word ptr [B0_OFF2]  ; Swap imag offset

          fld     st(1)
          fmul    st(0), st(4)
          fld     st(1)
          fmul    st(0), st(4)
          fsub
          fstp    dword ptr ds:[si]

          fmul    st(0), st(3)
          fxch
          fmul    st(0), st(2)
          fadd
          fstp    dword ptr es:[di]

          add     si, ax
          mov     [B0_OFF1], si
          add     si, ax

          add     di, bx
          mov     [B0_OFF2], di
          add     di, bx

          loop    b0_d

          mov     si, [B0_OFFA]           ; Shift starting pointer for real sets
          add     si, [B0_ASTR]
          mov     [B0_OFFA], si

          mov     di, [B0_OFFB]           ; Shift starting pointer for imag sets
          add     di, [B0_BSTR]
          mov     [B0_OFFB], di

          fld     st(3)
          fld     st(3)

          fld     st(3)                   ; Compute new real twiddle component
          fmul    st(0), st(2)
          fld     st(3)
          fmul    st(0), st(2)
          fsub

          fxch    st(4)                   ; Compute new imag twiddle component
          fmul
          fxch    st(2)
          fmul
          fadd

          dec     dx
          jz      b0_e
          jmp     b0_c

b0_e:     fstp    st
          fstp    st
          fstp    st
          fstp    st
          
          shr     word ptr [B0_PAIR], 1
          jc      b0_f   

          sub     word ptr [B0_TWDL], 16

          shr     word ptr [B0_DST1], 1
          shr     word ptr [B0_DST2], 1
          shl     word ptr [B0_SETS], 1
          jmp     b0_a

b0_f:     mov     dx, [B0_OPTN]           ; Determine if scaling is requested
          and     dx, 2
          jz      b0_h

          lds     si, [B0_AADD]           ; Point to first elements of real
          les     di, [B0_BADD]           ;  and imag array

          mov     cx, [B0_POWR]           ; Load cx with number of elements
          mov     dx, 1
          shl     dx, cl
          mov     cx, dx

          fild    word ptr [B0_POWR]      ; Load scaling factor
          fchs

b0_g:     fld     dword ptr ds:[si]       ; Scale real element
          fscale
          fstp    dword ptr ds:[si]       ; Store result

          add     si, ax                  ; Increment pointer

          fld     dword ptr es:[di]       ; Scale imag element
          fscale
          fstp    dword ptr es:[di]       ; Store result

          add     di, bx                  ; Increment pointer

          loop    b0_g                    ; Repeat until cx is 0

          fstp    st                      ; Pop constant off stack
b0_h:     fwait                           ; Wait for processor to catch up

          mov     word ptr [B0_POWR], 0   ; Set error code
          
exitb0:   pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, word ptr [B0_POWR]  ; Fetch error code

          leave

          RET     16                      ; Return

_vbfly0__FPfiPfiii ENDP



B1_OPTN   equ     bp+6
B1_POWR   equ     bp+8
B1_BSTR   equ     bp+10
B1_BADD   equ     bp+12
B1_ASTR   equ     bp+16
B1_AADD   equ     bp+18

_vbfly1__FPfiPfiii PROC
          enter   0, 0                    ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds

          mov     ax, SEG DGROUP          ; Compute address of tweedle constants
          push    ax
          pop     es
          mov     bx, [B1_POWR]
          shl     bx, 4
          add     bx, offset tweedle

          fld     qword ptr es:[bx]       ; Load tweedle factor
          add     bx, 8
          fld     qword ptr es:[bx]

          fld     st(1)                   ; Initialize wr and wi
          fld     st(1)

          mov     ax, [B1_OPTN]           ; Determine if inverse is requested
          and     ax, 0001
          jz      b1_a

          fld1                            ; Load -1
          fchs

          fmul    st(3), st(0)            ; Change sign of tweedle values
          fmulp   st(2), st(0)
     
b1_a:     shl     word ptr [B1_ASTR], 2   ; Compute strides 
          shl     word ptr [B1_BSTR], 2

          mov     cx, word ptr [B1_POWR]  ; Compute number of elements
          mov     ax, 1
          shl     ax, cl

          push    ax                      ; Save number elements

          dec     ax                      ; Compute separation in elements
          dec     ax

          mov     bx, ax                  ; Compute offset between real elements
          mul     word ptr [B1_ASTR]

          xchg    bx, ax                  ; Compute offset between imag elements
          mul     word ptr [B1_BSTR]

          mov     cx, [B1_ASTR]           ; Load strides
          mov     dx, [B1_BSTR]

          lds     si, [B1_AADD]           ; Point to first elements of real
          les     di, [B1_BADD]           ;  and imag array

          add     si, cx
          add     di, dx

b1_b:     fld     dword ptr ds:[si+bx]    ; Compute sum and difference of reals
          fld     st(0)
          fld     dword ptr ds:[si]

          fsub    st(2), st(0)
          fadd
          fstp    dword ptr ds:[si]

          xchg    bx, ax

          fld     dword ptr es:[di]       ; Compute sum and difference of imags
          fld     st(0)
          fld     dword ptr es:[di+bx]

          fadd    st(2), st(0)
          fsub
          fstp    dword ptr es:[di]

          fld     st(0)                   ; Compute real and imag of W*F2
          fmul    st(0), st(3)
          fxch    st(1)
          fmul    st(0), st(4)

          fld     st(2)
          fmul    st(0), st(4)
          fsub

          fxch    st(2)
          fmul    st(0), st(4)
          fadd

          fld     st(0)                   ; Compute imag components
          fld     dword ptr es:[di]
          fadd    st(2), st(0)
          fsub
          fstp    dword ptr es:[di+bx]
          fstp    dword ptr es:[di]

          xchg    ax, bx

          fld     st(0)                   ; Compute real components
          fld     dword ptr ds:[si]
          fadd    st(2), st(0)
          fsubr
          fstp    dword ptr ds:[si+bx]
          fstp    dword ptr ds:[si]

          fld     st(3)
          fld     st(3)

          fld     st(3)                   ; Compute new real twiddle component
          fmul    st(0), st(2)
          fld     st(3)
          fmul    st(0), st(2)
          fsub

          fxch    st(4)                   ; Compute new imag twiddle component
          fmul
          fxch    st(2)
          fmul
          fadd

          add     di, dx                  ; Adjust pointers for imag elements
          sub     ax, dx
          sub     ax, dx

          add     si, cx                  ; Adjust pointers for real elements
          sub     bx, cx
          sub     bx, cx

          jge     b1_b

          fstp    st(0)                   ; Clean up residual values on stack
          fstp    st(0)
          fstp    st(0)
          fstp    st(0)

          mov     ax, cx
          mov     bx, dx

          lds     si, [B1_AADD]           ; Point to first elements of real
          les     di, [B1_BADD]           ;  and imag array

          fld1                            ; Load scaling factor
          fchs

          fld     dword ptr ds:[si]       ; Compute values for zero freq.
          fld     st(0)
          fld     dword ptr es:[di]
          fadd    st(2), st(0)
          fsub
          fstp    dword ptr es:[di]
          fstp    dword ptr ds:[si]

          pop     cx                      ; Fetch number of elements

          mov     dx, [B1_OPTN]           ; Determine if inverse is requested
          and     dx, 0001
          jnz     b1_c

          dec     cx

          add     si, ax                  ; Increment pointers
          add     di, bx

b1_c:     fld     dword ptr ds:[si]       ; Scale real element
          fscale
          fstp    dword ptr ds:[si]       ; Store result

          add     si, ax                  ; Increment pointer

          fld     dword ptr es:[di]       ; Scale imag element
          fscale
          fstp    dword ptr es:[di]       ; Store result

          add     di, bx                  ; Increment pointer

          loop    b1_c                    ; Repeat until cx is 0

          fstp    st                      ; Pop constant off stack
          fwait                           ; Wait for processor to catch up

          mov     word ptr [B1_POWR], 0   ; Set return code

exitb1:   pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, word ptr [B1_POWR]  ; Fetch error code

          leave

          RET     16                      ; Return

_vbfly1__FPfiPfiii ENDP
          END

