; Copyright (C)  Michael Holzrichter 1992,  All rights reserved.

        title   VSPLINE.ASM:  Create and Evaluate Cubic Splines
        page    55,132

; Call with:    vsplin (float* a, int str_a, float* b, int str_b, float* c, int str_c, int n, float* s1, float* s2, int opt, float* w);
;
; INPUT:
; a     - array containing the x coordinates
;         length = n, stride = str_a
; b     - array containing the y coordinates
;         length = n, stride = str_b
; w     - work array
;         length = 3 * n (must be contiguous)
; s1    - slope at x [0]   (not used if not clamped at x [0])
; s2    - slope at x [n-1] (not used if not clamped at x [n-1])
; opt   - clamped ends parameter
;         clamp spline at x [0]   if bit 0 is set
;         clamp spline at x [n-1] if bit 8 is set
;
; OUTPUT:
; c     - array with parameter for spline at each input point
;         length = n, stride = str_c
;
; Returns:      c [i*str_c] = spline_coef (a [i*str_a], b [i*str_b]) for i = 0 to n-1

;
;
; Call with:    vspval0 (float* a, int str_a, float* b, int str_b, float* c, int str_c, int n, float d, float* e);
;
; INPUT:
; a     - array containing the x coordinates
;         length = n, stride = str_a
; b     - array containing the y coordinates
;         length = n, stride = str_b
; c     - array with parameter for spline at each input point
;         length = n, stride = str_c
; d     - x coordinate where spline is to be evaluated
;
; OUTPUT:
; e     - value of spline at d
; 
; ERROR CODES:
; 0     - no error
; 1     - n < 2
; 2     - d < x [0]
; 3     - d > x [n-1]
;
; Returns:      *e = spline_eval (a, b, c, d)
;
; Destroys:     nothing
;
;
;
; Call with:    vspval1 (float* a, int str_a, float* b, int str_b, float* c, int str_c, int n, float* d, int str_d, float* e, int str_e, int m);
;
; INPUT:
; a     - array containing the x coordinates
;         length = n, stride = str_a
; b     - array containing the y coordinates
;         length = n, stride = str_b
; c     - array with parameter for spline at each input point
;         length = n, stride = str_c
; d     - array with x coordinates where spline is to be evaluated
;         length = m, stride = str_d
;
; OUTPUT:
; e     - array with values of spline at x coordinates in d
;         length = m, stride = str_e
; 
; ERROR CODES:
; 0     - no error
; 1     - n < 2
; 2     - m < 1
; 3     - d [0] < x [0]   or d [m-1] < x [0]
; 4     - d [0] > x [n-1] or d [m-1] > x [n-1]
; 5     - x [i] < x [i-1]
;
; Returns:      *e [i*str_e] = spline_eval (a, b, c, d) for i = 0 to m-1
;
; Destroys:     nothing
;
;
;
; Call with:    vspval2 (float* a, int str_a, float* b, int str_b, float* c, int str_c, int n, float d_0, float d_inc, float* e, int str_e, int m);
;
; INPUT:
; a     - array containing the x coordinates
;         length = n, stride = str_a
; b     - array containing the y coordinates
;         length = n, stride = str_b
; c     - array with parameter for spline at each input point
;         length = n, stride = str_c
; d_0   - first point to evaluate spline
; d_inc - increment between points where spline is evaluated
;
; OUTPUT:
; e     - array with values of spline at x coordinates d0+i*d_inc
;         length = m, stride = str_e
; 
; ERROR CODES:
; 0     - no error
; 1     - n < 2
; 2     - m < 1
; 3     - d0 < x [0]   or dm1 < x [0]   (where dm1 = d0 + (m-1)*d_inc)
; 4     - d0 > x [n-1] or dm1 > x [n-1] (where dm1 = d0 + (m-1)*d_inc)
; 5     - x [i] < x [i-1]
; 6     - d_inc < 0.0
;
; Returns:      *e [i*str_e] = spline_eval (a, b, c, d0+i*d_inc) for i = 0, m-1
;
; Destroys:     nothing
;
;*******************************************************************************
; NOTE:  The agortithms in this file are impementations of algorithms found in
;        "Numerical Recipes in C".  Refer to that book to aid in understanding
;        what is done here.  Significant changes have been made to optimize
;        for the 80287 coprocessor and the specific uses I want to use this for.
;*******************************************************************************

.286
.287

;.MODEL SMALL
.MODEL LARGE
.DATA

.CODE

W_OFF     equ     bp+6
W_SEG     equ     bp+8
OPTN      equ     bp+10
SLP_2     equ     bp+12
SLP_1     equ     bp+16
N_ELM     equ     bp+20
C_STR     equ     bp+22
C_ADD     equ     bp+24
B_STR     equ     bp+28
B_ADD     equ     bp+30
A_STR     equ     bp+34
A_ADD     equ     bp+36

          PUBLIC  _vsplin__FPfiPfiPfiiffiPf, _vspval0__FPfiPfiPfiifPf, _vspval1__FPfiPfiPfiiPfiPfii, _vspval2__FPfiPfiPfiiffPfii

_vsplin__FPfiPfiPfiiffiPf PROC
          push    bp                      ; Entry sequence - save old BP
          mov     bp, sp                  ; Set stack framepointer

          push    si                      ; Save registers
          push    di
          push    es
          push    ds

          push    dx
          push    cx
          push    bx

          mov     cx, [N_ELM]             ; Get number of elements

          cmp     cx, 2                   ; Check for positive vector length
          jg      oksp
          jmp     error

oksp:     sub     cx, 2                   ; Loop n-2 times

          lds     si, [A_ADD]             ; Get address of A array

          les     di, [W_OFF]             ; Get address of W array

          mov     ax, [A_STR]             ; Compute stride of source A
          shl     ax, 1
          shl     ax, 1

          mov     bx, 8                   ; Load strides for work array
          mov     dx, 4

          fld     dword ptr ds:[si]       ; Load x0 and x1
          add     si, ax
          fld     dword ptr ds:[si]       ; x1        x0
          add     si, ax
          
loop1:    fld     st                      ; xi        xi        xi-1
          fsub    st, st(2)               ; xi-xi-1   xi        xi-1
          fstp    dword ptr es:[di]       ; xi        xi-1
          add     di, bx                  ; Skip one element in W array

          fxch                            ; xi-1      xi
          fld     dword ptr ds:[si]       ; xi+1      xi-1      xi
          fxch                            ; xi-1      xi+1      xi
          fsubr   st, st(1)               ; xi+1-xi-1 xi+1      xi
          fstp    dword ptr es:[di]       ; xi+1      xi

          add     si, ax                  ; Point to next element in A array
          add     di, dx                  ; Point to next elemtnt in W array

          loop    loop1

          fsubr                           ; Compute last difference 
          fstp    dword ptr es:[di]

          mov     cx, [N_ELM]             ; Get number of elements
          dec     cx                      ; Loop n-1 times

          lds     si, [B_ADD]             ; Get address of B array

          les     di, [W_OFF]             ; Get address of W array
          add     di, 4                   ; Point to open slot for yi-yi-1

          mov     ax, [B_STR]             ; Compute stride of source B
          shl     ax, 1
          shl     ax, 1

          mov     bx, 12                  ; Load stride for work

          fld     dword ptr ds:[si]       ; Load y0
          add     si, ax
          
loop2:    fld     dword ptr ds:[si]       ; yi        yi-1
          add     si, ax                  ; Point to next element in B array
          fxch                            ; yi-1      yi
          fsubr   st, st(1)               ; yi-yi-1   yi
          fstp    dword ptr es:[di]       ; yi
          add     di, bx                  ; Point to next element in W array

          loop    loop2

          fstp    st                      ; Pop residual value off stack

          mov     dx, [OPTN]              ; Get option parameter

          mov     cx, [N_ELM]             ; Get number of elements
          sub     cx, 2                   ; Loop n-2 times

          mov     ax, 4                   ; Load stride for both in and out

          fld1                            ; Get value of 2.0 on stack
          fadd    st, st

          lds     si, [W_OFF]             ; Load address of W array

          fld     dword ptr ds:[si]       ; Load x2-x1
          add     si, ax
          fld     dword ptr ds:[si]       ; Load y2-y1
          add     si, ax

          fxch
          fdiv    st(1), st

          cmp     dl, 0                   ; Test which end condition requested
          je      a0

          fld1                            ; Initialize for clamped end condition
          fdiv    st, st(3)
          fchs
          fxch    st(2)

          fld     st
          fsub    dword ptr [SLP_1]
          fdiv    st, st(4)
          fdiv    st, st(2)
          fxch    st(2)

          jmp     a1

a0:       fldz                            ; Initialize for free end conditions
          fxch    st(2)
          fldz
          fxch    st(2)

a1:       les     di, [W_OFF]             ; Get address of W array for output

          fld     st(2)                   ; Save u(1) and y2(1)
          fstp    dword ptr es:[di]
          add     di, ax
          fld     st(3)
          fstp    dword ptr es:[di]
          add     di, ax


loop3:    fld     dword ptr ds:[si]       ; Compute u(i) and y2(i)
          add     si, ax
          fxch
          fdiv    st, st(1)
          fmul    st(3), st
          fxch    st(4)
          fmul    st, st(4)
          fadd    st, st(5)
          fdiv    st(3), st
          fld1
          fsubp   st(5), st
          fdiv    st(4), st
          fmul

          fld     dword ptr ds:[si]
          add     si, ax
          fxch
          fld     dword ptr ds:[si]
          add     si, ax
          fdiv    st, st(2)
          fxch    st(3)
          fsubr   st, st(3)
          fdivr
          fsubrp  st(3), st

          fld     st(2)                   ; Save u(i) & y2(i) values
          fstp    dword ptr es:[di]
          add     di, ax
          fld     st(3)
          fstp    dword ptr es:[di]
          add     di, ax

          loop    loop3

          cmp     dh, 0                   ; Test which end condition requested
          je      b0


          fxch
          fsubr   dword ptr [SLP_2]
          fdivr
          fsubr
          fxch    st(2)
          fadd
          fdiv
          
          jmp     b1          

b0:       fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fldz

b1:       mov     cx, [N_ELM]             ; Get number of elements
          dec     cx                      ; Loop n-1 times

          mov     ax, di                  ; Get address of intermediate values
          mov     si, ax
          mov     ds, [W_SEG]

          mov     ax, [C_STR]             ; Compute stride of C array
          shl     ax, 1
          shl     ax, 1

          mov     bx, ax                  ; Compute address of end of C array
          mul     cx
          les     di, [C_ADD]
          add     di, ax
          mov     ax, bx

          mov     bx, 4                   ; Load stride of work array
          sub     si, bx                  ; Backup one element to last written

          fst     dword ptr es:[di]       ; Save last element of C array
          sub     di, ax

loop4:    fmul    dword ptr ds:[si]       ; Compute elements of C array
          sub     si, bx
          fadd    dword ptr ds:[si]
          sub     si, bx

          fst     dword ptr es:[di]
          sub     di, ax

          loop    loop4

          fstp    st                      ; Pop residual value off stack

          fwait                           ; Wait for processor to catch up

          xor     ax, ax                  ; Return code is 0

exit:     pop     bx                      ; Restore registers
          pop     cx
          pop     dx

          pop     ds
          pop     es
          pop     di
          pop     si

          pop     bp

          RET     34                      ; Return

error:    mov     ax, 1                   ; Load error code
          jmp     exit

_vsplin__FPfiPfiPfiiffiPf ENDP

U_NDX0    equ     bp-6
L_NDX0    equ     bp-4
E_COD0    equ     bp-2

E_ADD0    equ     bp+6
X_VAL0    equ     bp+10
N_ELM0    equ     bp+14
C_STR0    equ     bp+16
C_ADD0    equ     bp+18
B_STR0    equ     bp+22
B_ADD0    equ     bp+24
A_STR0    equ     bp+28
A_ADD0    equ     bp+30
        
_vspval0__FPfiPfiPfiifPf PROC
          enter   6, 0                    ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds

          xor     ax, ax                  ; Store lower index
          mov     [L_NDX0], ax
          mov     ax, [N_ELM0]            ; Store upper index
          dec     ax
          mov     [U_NDX0], ax

          jle     err_0_a                 ; Check for positive vector length

          shl     word ptr [A_STR0], 2    ; Compute strides
          shl     word ptr [B_STR0], 2
          shl     word ptr [C_STR0], 2

          fld     dword ptr [X_VAL0]      ; Load evaluation point

          lds     bx, [A_ADD0]            ; Load address of first element
          mov     cx, [A_STR0]            ; Load stride of array A

          mov     ax, [U_NDX0]            ; Compute offset of last coordinate
          mul     cx
          mov     si, ax

          fld     dword ptr ds:[bx]       ; Load first x-coordinate

          fcom    st(1)                   ; Make sure not to left
          fstsw   ax
          sahf
          ja      err_0_b

          fld     dword ptr ds:[bx+si]    ; Load last x-coordinate

          fcom    st(2)                   ; Make sure not to right
          fstsw   ax
          sahf
          jb      err_0_c

lbl_0_b:  mov     ax, [U_NDX0]            ; Compute index of midpoint
          add     ax, [L_NDX0]
          shr     ax, 1                   ; Divide by 2 to get average

          mov     di, ax                  ; Save index

          cmp     ax, [L_NDX0]            ; Is U_NDX0 - L_NDX0 <= 1?
          jle     lbl_0_d

          mul     cx                      ; Find offset of midpoint element
          mov     si, ax

          fld     dword ptr ds:[bx+si]    ; Load midpoint value

          fcom    st(3)                   ; Compare with requested coordinate
          fstsw   ax
          sahf
          ja      lbl_0_c

          fstp    st(2)                   ; Save new lower bound
          mov     [L_NDX0], di
          jmp     lbl_0_b

lbl_0_c:  fstp    st(1)                   ; Save new upper bound
          mov     [U_NDX0], di
          jmp     lbl_0_b

;*******************************************************************************
;         Handle error conditions
;*******************************************************************************

err_0_a:  mov     word ptr [E_COD0], 1    ; Load error code
          jmp     exit0

err_0_b:  fstp    st                      ; Pop values off stack
          fstp    st
          mov     word ptr [E_COD0], 3    ; Load error code
          jmp     exit0

err_0_c:  fstp    st                      ; Pop values off stack
          fstp    st
          fstp    st
          mov     word ptr [E_COD0], 4    ; Load error code
          jmp     exit0

;*******************************************************************************
; Precompute constants
;*******************************************************************************

lbl_0_d:  fsubr   st(2), st
          fsubr

          lds     bx, [B_ADD0]            ; Compute address of B
          mov     si, [B_STR0]
          mov     ax, di
          mul     si
          add     bx, ax

          fld     dword ptr ds:[bx+si]    ; Load Y (upper)
          fdiv    st, st(1)

          fld     dword ptr ds:[bx]       ; Load Y (lower)
          fdiv    st, st(2)

          lds     bx, [C_ADD0]            ; Compute address of C
          mov     si, [C_STR0]
          mov     ax, di
          mul     si
          add     bx, ax

          fld     dword ptr ds:[bx+si]    ; Load Y2 (upper)
          fld     st
          fmul    st, st(4)
          fsubp   st(3), st
          fdiv    st, st(3)

          fld     dword ptr ds:[bx]       ; Load Y2 (lower)
          fld     st
          fmul    st, st(5)
          fsubp   st(3), st
          fdiv    st, st(4)

          lds     bx, [E_ADD0]            ; Load destination address

          fld     st
          fmul    st, st(6)
          fmul    st, st(6)
          fadd    st, st(3)
          fmul    st, st(6)

          fxch    st(6)
          fsubr   st, st(5)
          fxch    st(6)

          fld     st(2)
          fmul    st, st(7)
          fmul    st, st(7)
          fadd    st, st(5)
          fmul    st, st(7)

          fadd

          fstp    dword ptr ds:[bx]       ; Save value

          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st

          fwait                           ; Wait for processor to catch up

          mov     word ptr [E_COD0], 0    ; Return code is 0

exit0:    pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, [E_COD0]            ; Get return code

          leave

          RET     28                      ; Return
_vspval0__FPfiPfiPfiifPf ENDP


X_LST1    equ     bp-16
X_REF1    equ     bp-12
A_END1    equ     bp-8
U_NDX1    equ     bp-6
L_NDX1    equ     bp-4
E_COD1    equ     bp-2

M_ELM1    equ     bp+6 
E_STR1    equ     bp+8
E_ADD1    equ     bp+10
D_STR1    equ     bp+14
D_ADD1    equ     bp+16
N_ELM1    equ     bp+20
C_STR1    equ     bp+22
C_ADD1    equ     bp+24
B_STR1    equ     bp+28
B_ADD1    equ     bp+30
A_STR1    equ     bp+34
A_ADD1    equ     bp+36
        
_vspval1__FPfiPfiPfiiPfiPfii PROC
          enter   16, 0                   ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds
     
          shl     word ptr [A_STR1], 2    ; Compute strides
          shl     word ptr [B_STR1], 2
          shl     word ptr [C_STR1], 2
          shl     word ptr [D_STR1], 2
          shl     word ptr [E_STR1], 2

          xor     ax, ax                  ; Store lower index
          mov     [L_NDX1], ax
          mov     ax, [N_ELM1]            ; Store upper index
          dec     ax
          mov     [U_NDX1], ax

          jle     err_1_a                 ; Check for positive spline vector length

          mov     ax, [M_ELM1]
          dec     ax
          jl      err_1_b                 ; Check for positive x-coord vector length

          lds     bx, [D_ADD1]            ; Load first and last evaluation point

          mul     word ptr [D_STR1]
          mov     si, ax

          fld     dword ptr ds:[bx]
          fst     dword ptr [X_LST1]

          fld     dword ptr ds:[bx+si]

          lds     bx, [A_ADD1]            ; Load address of first element of A
          mov     cx, [A_STR1]            ; Load stride of array A

          mov     ax, [U_NDX1]            ; Compute offset of last coordinate
          mul     cx
          mov     si, ax

          mov     [A_END1], si            ; Save offset of end of array
          add     [A_END1], bx

          fld     dword ptr ds:[bx]       ; Load first x-coordinate

          fcom    st(2)                   ; Make sure first coordinate to left
          fstsw   ax                      ;  of first evaluation point
          sahf
          ja      err_1_c

          fcom    st(1)                   ; Make sure first coordinate to left
          fstsw   ax                      ;  of last evaluation point
          sahf
          ja      err_1_c

          fxch

          fld     dword ptr ds:[bx+si]    ; Load last x-coordinate

          fcom    st(3)                   ; Make sure last coordinate to right
          fstsw   ax                      ;  of first evaluation point
          sahf
          jb      err_1_d

          fcom    st(1)                   ; Make sure last coordinate to right
          fstsw   ax                      ;  of last evaluation point
          sahf
          jb      err_1_d

          fxch
          fstp    st

lbl_1_b:  mov     ax, [U_NDX1]            ; Compute index of midpoint
          add     ax, [L_NDX1]
          shr     ax, 1                   ; Divide by 2 to get average

          mov     di, ax                  ; Save index

          cmp     ax, [L_NDX1]            ; Is U_NDX1 - L_NDX1 <= 1?
          jle     lbl_1_d

          mul     cx                      ; Find offset of midpoint element
          mov     si, ax

          fld     dword ptr ds:[bx+si]    ; Load midpoint value

          fcom    st(3)                   ; Compare with requested coordinate
          fstsw   ax
          sahf
          ja      lbl_1_c

          fstp    st(2)                   ; Save new lower bound
          mov     [L_NDX1], di
          jmp     lbl_1_b

lbl_1_c:  fstp    st(1)                   ; Save new upper bound
          mov     [U_NDX1], di
          jmp     lbl_1_b

;*******************************************************************************
;         Handle error conditions
;*******************************************************************************

err_1_a:  mov     word ptr [E_COD1], 1    ; Load error code
          jmp     exit1

err_1_b:  mov     word ptr [E_COD1], 2    ; Load error code
          jmp     exit1

err_1_c:  fstp    st                      ; Pop values off stack
          fstp    st
          fstp    st
          mov     word ptr [E_COD1], 3    ; Load error code
          jmp     exit1

err_1_d:  fstp    st                      ; Pop values off stack
          fstp    st
          fstp    st
          fstp    st
          mov     word ptr [E_COD1], 4    ; Load error code
          jmp     exit1

;*******************************************************************************
; Precompute constants
;*******************************************************************************

lbl_1_d:  sub     [N_ELM1], di            ; Decrement number of elements
          dec     word ptr [N_ELM1]       

          fst     dword ptr [X_REF1]      ; Save current upper bound
          fsubr   st(2), st
          fsubr

          mov     ax, di                  ; Compute new starting address of A
          inc     ax
          mul     cx
          add     bx, ax
          mov     [A_ADD1], bx

          lds     bx, [B_ADD1]            ; Compute new starting address of B
          mov     si, [B_STR1]
          mov     ax, di
          mul     si
          add     bx, ax
          add     si, bx
          mov     [B_ADD1], si

          fld     dword ptr ds:[si]       ; Load Y (upper)
          fdiv    st, st(1)

          fld     dword ptr ds:[bx]       ; Load Y (lower)
          fdiv    st, st(2)

          lds     bx, [C_ADD1]            ; Compute new starting address of C
          mov     si, [C_STR1]
          mov     ax, di
          mul     si
          add     bx, ax
          add     si, bx
          mov     [C_ADD1], si

          fld     dword ptr ds:[si]       ; Load Y2 (upper)
          fld     st
          fmul    st, st(4)
          fsubp   st(3), st
          fdiv    st, st(3)

          fld     dword ptr ds:[bx]       ; Load Y2 (lower)
          fld     st
          fmul    st, st(5)
          fsubp   st(3), st
          fdiv    st, st(4)

          lds     si, [D_ADD1]            ; Reload source & destination address
          mov     bx, [D_STR1]
          les     di, [E_ADD1]
          mov     cx, [E_STR1]

lbl_1_e:  fld     st
          fmul    st, st(6)
          fmul    st, st(6)
          fadd    st, st(3)
          fmul    st, st(6)

          fxch    st(6)
          fsubr   st, st(5)
          fxch    st(6)

          fld     st(2)
          fmul    st, st(7)
          fmul    st, st(7)
          fadd    st, st(5)
          fmul    st, st(7)

          fadd

          fstp    dword ptr es:[di]       ; Save value and point to next location
          add     di, cx

          dec     word ptr [M_ELM1]       ; Decrement number of evaluation points
          jz      lbl_1_j

          add     si, bx                  ; Get next evaluation point
          fld     dword ptr ds:[si]

          fcom    dword ptr [X_LST1]

          fstsw   ax                      ; Error if points are not increasing
          sahf
          jb      err_1_e

          fst     dword ptr [X_LST1]      ; Save current evaluation point

          fcom    dword ptr [X_REF1]      ; Subract from current upper bound

          fstsw   ax
          sahf
          ja      lbl_1_f                 ; Jump if outside of current interval

          fsubr   dword ptr [X_REF1]      ; Subract from current upper bound
          fstp    st(6)
          jmp     lbl_1_e                 ; Jump if inside of current interval

;*******************************************************************************
; Get ready to exit
;*******************************************************************************

lbl_1_j:  fstp    st                      ; Pop residual values off stack
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st

          fwait                           ; Wait for coprocessor to catch up

          mov     word ptr [E_COD1], 0    ; Return code is 0

          jmp     exit1

;*******************************************************************************
;         Handle error condition
;*******************************************************************************

err_1_e:  fstp    st                      ; Pop value off stack
          mov     word ptr [E_COD1], 5    ; Load error code
          jmp     exit1

;*******************************************************************************
; Search for interval containing next point
;*******************************************************************************

lbl_1_f:  fstp    st(6)                   ; Get rid of old constants
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st

          mov     [D_ADD1], si            ; Save offsets of D and E arrays
          mov     [E_ADD1], di

          lds     bx, [A_ADD1]            ; Reload address of spline x-coordinates
          mov     si, [A_STR1]
          mov     di, 1
          mov     dx, [N_ELM1]
          xor     cx, cx

lbl_1_g:  add     cx, di
          cmp     dx, cx                  ; Make sure next interval does not go past end of spline
          jbe     lbl_1_i

          fcom    dword ptr ds:[bx+si]

          fstsw   ax
          sahf
          jb      lbl_1_h

          add     bx, si

          shl     si, 1                   ; Double interval size
          shl     di, 1

          jmp     lbl_1_g

lbl_1_h:  sub     cx, di
          sub     dx, cx

          mov     [N_ELM1], dx            ; Initialize for contraction phase
          mov     [A_ADD1], bx

          mov     ax, [B_STR1]            ; Increment pointer to B array
          mul     cx
          add     [B_ADD1], ax

          mov     ax, [C_STR1]            ; Increment pointer to C array
          mul     cx
          add     [C_ADD1], ax

          xor     ax, ax                  ; Initialize indices of interval
          mov     word ptr [L_NDX1], ax
          mov     [U_NDX1], di

          mov     cx, [A_STR1]

          fld     dword ptr ds:[bx]       ; Initialize coprocessor stack
          fld     dword ptr ds:[bx+si]

          jmp     lbl_1_b                 ; Jump to contraction loop

lbl_1_i:  sub     di, cx                  ; Jump here if interval goes past end
          mov     cx, dx
          dec     cx
          add     di, cx

          mov     si, [A_END1]
          sub     si, bx

          jmp     lbl_1_h

exit1:    pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, [E_COD1]            ; Get return code

          leave

          RET     34                      ; Return

_vspval1__FPfiPfiPfiiPfiPfii ENDP



I_CNT2    equ     bp-14
X_REF2    equ     bp-12
A_END2    equ     bp-8
U_NDX2    equ     bp-6
L_NDX2    equ     bp-4
E_COD2    equ     bp-2

M_ELM2    equ     bp+6 
E_STR2    equ     bp+8
E_ADD2    equ     bp+10
D_INC2    equ     bp+14
D_BAS2    equ     bp+18
N_ELM2    equ     bp+22
C_STR2    equ     bp+24
C_ADD2    equ     bp+26
B_STR2    equ     bp+30
B_ADD2    equ     bp+32
A_STR2    equ     bp+36
A_ADD2    equ     bp+38
        
_vspval2__FPfiPfiPfiiffPfii PROC
          enter   14, 0                   ; Entry sequence - save old BP

          pusha                           ; Save registers and flags
          pushf

          push    es                      ; Save segments
          push    ds
     
          shl     word ptr [A_STR2], 2    ; Compute strides
          shl     word ptr [B_STR2], 2
          shl     word ptr [C_STR2], 2
          shl     word ptr [E_STR2], 2

          xor     ax, ax                  ; Store lower index
          mov     [L_NDX2], ax
          mov     ax, [N_ELM2]            ; Store upper index
          dec     ax
          mov     [U_NDX2], ax

          jle     err_2_a                 ; Check for positive spline vector length

          mov     cx, [M_ELM2]
          cmp     cx, 1
          jl      err_2_b                 ; Check for positive x-coord vector length

          lds     bx, [E_ADD2]
          mov     dx, [E_STR2]

          fld     dword ptr [D_BAS2]

          fld     dword ptr [D_INC2]

          ftst                            ; Check that increment is not negative
          fstsw   ax
          sahf
          jb      err_2_c

          xor     si, si
          mov     word ptr [I_CNT2], si

          jmp     lbl_2_a

;*******************************************************************************
;         Handle error conditions
;*******************************************************************************

err_2_a:  mov     word ptr [E_COD2], 1    ; Load error code
          jmp     exit2

err_2_b:  mov     word ptr [E_COD2], 2    ; Load error code
          jmp     exit2

err_2_c:  fstp    st                      ; Pop values off stack
          mov     word ptr [E_COD2], 6
          jmp     exit2

err_2_d:  fstp    st                      ; Pop values off stack
          fstp    st
          fstp    st
          mov     word ptr [E_COD2], 3    ; Load error code
          jmp     exit2

err_2_e:  fstp    st                      ; Pop values off stack
          fstp    st
          fstp    st
          fstp    st
          mov     word ptr [E_COD2], 4    ; Load error code
          jmp     exit2

;*******************************************************************************
; Compute coordinates at which to evaluate spline
;*******************************************************************************

lbl_2_a:  fld     st                      ; Compute evaluation coordinate
          fimul   word ptr [I_CNT2]
          fadd    st,st(2)

          fstp    dword ptr ds:[bx+si]    ; Store value and increment offset
          add     si, dx

          inc     word ptr [I_CNT2]       ; Increment counter

          loop    lbl_2_a

          fstp    st

          sub     si, dx
          fld     dword ptr ds:[bx+si]

          lds     bx, [A_ADD2]            ; Load address of first element of A
          mov     cx, [A_STR2]            ; Load stride of array A

          mov     ax, [U_NDX2]            ; Compute offset of last coordinate
          mul     cx
          mov     si, ax

          mov     [A_END2], si            ; Save offset of end of array
          add     [A_END2], bx

          fld     dword ptr ds:[bx]       ; Load first x-coordinate

          fcom    st(2)                   ; Make sure first coordinate to left
          fstsw   ax                      ;  of first evaluation point
          sahf
          ja      err_2_d

          fcom    st(1)                   ; Make sure first coordinate to left
          fstsw   ax                      ;  of last evaluation point
          sahf
          ja      err_2_d

          fxch

          fld     dword ptr ds:[bx+si]    ; Load last x-coordinate

          fcom    st(3)                   ; Make sure last coordinate to right
          fstsw   ax                      ;  of first evaluation point
          sahf
          jb      err_2_e

          fcom    st(1)                   ; Make sure last coordinate to right
          fstsw   ax                      ;  of last evaluation point
          sahf
          jb      err_2_e 

          fxch
          fstp    st

lbl_2_b:  mov     ax, [U_NDX2]            ; Compute index of midpoint
          add     ax, [L_NDX2]
          shr     ax, 1                   ; Divide by 2 to get average

          mov     di, ax                  ; Save index

          cmp     ax, [L_NDX2]            ; Is U_NDX1 - L_NDX1 <= 1?
          jle     lbl_2_d

          mul     cx                      ; Find offset of midpoint element
          mov     si, ax

          fld     dword ptr ds:[bx+si]    ; Load midpoint value

          fcom    st(3)                   ; Compare with requested coordinate
          fstsw   ax
          sahf
          ja      lbl_2_c

          fstp    st(2)                   ; Save new lower bound
          mov     [L_NDX2], di
          jmp     lbl_2_b

lbl_2_c:  fstp    st(1)                   ; Save new upper bound
          mov     [U_NDX2], di
          jmp     lbl_2_b

;*******************************************************************************
; Precompute constants
;*******************************************************************************

lbl_2_d:  sub     [N_ELM2], di            ; Decrement number of elements
          dec     word ptr [N_ELM2]       

          fst     dword ptr [X_REF2]      ; Save current upper bound
          fsubr   st(2), st
          fsubr

          mov     ax, di                  ; Compute new starting address of A
          inc     ax
          mul     cx
          add     bx, ax
          mov     [A_ADD2], bx

          lds     bx, [B_ADD2]            ; Compute new starting address of B
          mov     si, [B_STR2]
          mov     ax, di
          mul     si
          add     bx, ax
          add     si, bx
          mov     [B_ADD2], si

          fld     dword ptr ds:[si]       ; Load Y (upper)
          fdiv    st, st(1)

          fld     dword ptr ds:[bx]       ; Load Y (lower)
          fdiv    st, st(2)

          lds     bx, [C_ADD2]            ; Compute new starting address of C
          mov     si, [C_STR2]
          mov     ax, di
          mul     si
          add     bx, ax
          add     si, bx
          mov     [C_ADD2], si

          fld     dword ptr ds:[si]       ; Load Y2 (upper)
          fld     st
          fmul    st, st(4)
          fsubp   st(3), st
          fdiv    st, st(3)

          fld     dword ptr ds:[bx]       ; Load Y2 (lower)
          fld     st
          fmul    st, st(5)
          fsubp   st(3), st
          fdiv    st, st(4)

          les     di, [E_ADD2]            ; Reload source & destination address
          mov     cx, [E_STR2]

lbl_2_e:  fld     st
          fmul    st, st(6)
          fmul    st, st(6)
          fadd    st, st(3)
          fmul    st, st(6)

          fxch    st(6)
          fsubr   st, st(5)
          fxch    st(6)

          fld     st(2)
          fmul    st, st(7)
          fmul    st, st(7)
          fadd    st, st(5)
          fmul    st, st(7)

          fadd

          fstp    dword ptr es:[di]       ; Save value and point to next location
          add     di, cx

          dec     word ptr [M_ELM2]       ; Decrement number of evaluation points
          jz      lbl_2_j

          fld     dword ptr es:[di]       ; Get next evaluation point

          fcom    dword ptr [X_REF2]      ; Subract from current upper bound

          fstsw   ax
          sahf
          ja      lbl_2_f                 ; Jump if outside of current interval

          fsubr   dword ptr [X_REF2]      ; Subract from current upper bound
          fstp    st(6)
          jmp     lbl_2_e                 ; Jump if inside of current interval

;*******************************************************************************
; Get ready to exit
;*******************************************************************************

lbl_2_j:  fstp    st                      ; Pop residual values off stack
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st

          fwait                           ; Wait for coprocessor to catch up

          mov     word ptr [E_COD2], 0    ; Return code is 0

          jmp     exit2

;*******************************************************************************
; Search for interval containing next point
;*******************************************************************************

lbl_2_f:  fstp    st(6)                   ; Get rid of old constants
          fstp    st
          fstp    st
          fstp    st
          fstp    st
          fstp    st

          mov     [E_ADD2], di            ; Save offsets of D and E arrays

          lds     bx, [A_ADD2]            ; Reload address of spline x-coordinates
          mov     si, [A_STR2]
          mov     di, 1
          mov     dx, [N_ELM2]
          xor     cx, cx

lbl_2_g:  add     cx, di
          cmp     dx, cx                  ; Make sure next interval does not go past end of spline
          jbe     lbl_2_i

          fcom    dword ptr ds:[bx+si]

          fstsw   ax
          sahf
          jb      lbl_2_h

          add     bx, si

          shl     si, 1                   ; Double interval size
          shl     di, 1

          jmp     lbl_2_g

lbl_2_h:  sub     cx, di
          sub     dx, cx

          mov     [N_ELM2], dx            ; Initialize for contraction phase
          mov     [A_ADD2], bx

          mov     ax, [B_STR2]            ; Increment pointer to B array
          mul     cx
          add     [B_ADD2], ax

          mov     ax, [C_STR2]            ; Increment pointer to C array
          mul     cx
          add     [C_ADD2], ax

          xor     ax, ax                  ; Initialize indices of interval
          mov     word ptr [L_NDX2], ax
          mov     [U_NDX2], di

          mov     cx, [A_STR2]

          fld     dword ptr ds:[bx]       ; Initialize coprocessor stack
          fld     dword ptr ds:[bx+si]

          jmp     lbl_2_b                 ; Jump to contraction loop

lbl_2_i:  sub     di, cx                  ; Jump here if interval goes past end
          mov     cx, dx
          dec     cx
          add     di, cx

          mov     si, [A_END2]
          sub     si, bx

          jmp     lbl_2_h

exit2:    pop     ds                      ; Restore segments
          pop     es

          popf
          popa

          mov     ax, [E_COD2]            ; Get return code

          leave

          RET     36                      ; Return

_vspval2__FPfiPfiPfiiffPfii ENDP
          END

