/* ----------------------------------------------------
 *  LISTING 4
 *
 *  Filename:           DRVMAN.C
 *  Summary:            Manager code for dynamic-link
 *                      DOS device drivers
 *  Author:             T.W. Nelson
 *  Compiler:           BORLAND
 *  Compile options:    Large model only
 *  Version:            1.00
 *  Date:               31-Jan-1993
 *  Notes:
 *
 *  Source code Copyright (c) 1992 T.W. Nelson. May
 *  be used only with appropriate acknowledgement of
 *  copyright.
 * ------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>
#include <dos.h>
#include <io.h>
#include <dir.h>
#include <assert.h>
#include "drv.h"

#define MAX_MODULES 30
#define SEG_PARAS   0x0fff  //#pgraphs in a segment

static struct ModuleDescriptor  mDesc[MAX_MODULES];
static int NumModules = 0;     //#modules loaded

//parameter block for DOS int 21/4bh, load overlay ....
struct dos_exec {
    unsigned seg;
    unsigned relocator;
    };

/* ------------------------------------------------- */
flag_t DrvLoadModule ( char *modname,
                        unsigned *handle,
                        void *args   )
{
   /* Load and initialize the requested module and
    * assign module handle. Return TRUE on success
    * (and also if module was previously loaded).
    */

    size_t pseg, rval;
    void *pmblock;
    register int i;
    union REGS regs;
    struct SREGS sregs;
    struct dos_exec parms;
    int (*OvEntry)( struct ModuleDescriptor * );
    struct ModuleDescriptor *pmDesc;
    char EXTname[MAXPATH], cwd[MAXPATH], *drvpath;

    //Check load status of module ...
    if( DrvGetModHandle( modname, handle) == TRUE )
            return TRUE;    //already loaded

    *handle = -1;       //pre-set for error return

    //Find open slot in mDesc[] array .......
    for( i = 0; i < MAX_MODULES ; i++ ) {
            if( mDesc[i].mseg == 0 )
                    break;      //found slot ......
    }

    if( i >= MAX_MODULES )
            return ERR_MOD_LIMIT;

    pmDesc = &mDesc[i];     //point to open slot

    //reset all data items to be assigned by
    //user driver initializer function ......
    pmDesc->modsize = 0;
    pmDesc->fTab = 0;
    pmDesc->exit_f = 0;
    pmDesc->id = 0;
    pmDesc->v_major = -1;
    pmDesc->v_minor = -1;

    assert( strlen( modname ) < MAX_NAME );
    strcpy( pmDesc->mname, modname );
    pmDesc->args = args;    //setup arguments

    //Access DRVPATH in environment and build pathname
    //string for extension module. Use CWD by default
    //if no DRVPATH variable exists.......
    drvpath = getenv("DRVPATH");
    fnmerge( EXTname, "",
             drvpath ? drvpath : getcwd(cwd, MAXPATH),
                                    modname, MOD_EXT);

    //Verify that driver extension module exists ....
    if( access( EXTname, 0 ) != 0 )
        return ERR_MOD_ACCESS;

    //Attempt to allocate up to a segment (64k bytes)
    //for the module. Calling allocmem() here always
    //fails .........
    if( (rval = allocmem( 0xffff, &pseg)) > SEG_PARAS )
         rval = SEG_PARAS;

    if((pmblock = malloc( rval << 4 )) == 0)
            return ERR_NO_MEM;       //out of memory

    //Normalize malloc() return and adjust module
    //load segment to next highest pgraph boundary ...
    rval = FP_OFF( pmblock );
    pseg = FP_SEG( pmblock ) + (rval >> 4) +
                      ((rval & 0x000f) ? 1 : 0);

    //Load driver module as an overlay ...........
    parms.seg = pseg;               //load parms block
    parms.relocator = pseg;
    sregs.es =  FP_SEG( (void *) &parms );
    regs.x.bx = FP_OFF( (void *) &parms );
    sregs.ds =  FP_SEG( (void *) EXTname );
    regs.x.dx = FP_OFF( (void *) EXTname );
    regs.x.ax = 0x4b03;             //DOS load overlay
    intdosx( &regs, &regs, &sregs );

    if( regs.x.cflag )  {
            free( pmblock );
            return ERR_BAD_LOAD;
    }

    OvEntry = MK_FP(pseg, 0);   //set up entry point
    rval = OvEntry( pmDesc );   //xfer to overlay

    if( rval )  {
            free( pmblock );
            return ERR_INIT_FAIL;
    }

    //Shrink mallocated block.  Add 1 to request size
    //to adjust for normalizing.  This code assumes
    //request size is always less than the size of
    //the original malloced block.......
    pmblock = realloc( pmblock,
                      (pmDesc->modsize + 1) << 4);

    //Update module descriptor array .....
    pmDesc->mseg = pseg;   //indicate module in use
    pmDesc->pblock = pmblock;  //use later for free()
    ++NumModules;           //keep running total
    *handle = i;            //assign handle

    return TRUE;
}

flag_t DrvGetProcAddr( unsigned handle,
                        char *procname,
                        DRVFUNC (**ppfn)())
{
   /* Given module handle and ASCIIZ procedure name,
    * return a pointer to indicated function in the
    * driver module.  Return NULL pointer if function
    * can't be found. Search is case-sensitive.
    */

    struct expTab *tabp;
    int i;

    if( handle >= MAX_MODULES ||
        mDesc[handle].mseg == 0 )     {
            *ppfn = 0L;         //return NULL ->
            return ERR_BAD_HANDLE;
    }

    tabp = mDesc[handle].fTab;

    for( i = 0; tabp[i].name ; i++ ) {
        if( strcmp( procname, tabp[i].name ) == 0 )  {
                *ppfn = tabp[i].pfun;
                return TRUE;
        }
    }

    *ppfn = 0L;         //return NULL ->
    return ERR_BAD_LINK;       //no match found
}

flag_t DrvGetModID( unsigned handle,
                     int *major,
                     int *minor,
                     char *idstr    )
{
    //Return ID information for an active module.
    struct ModuleDescriptor *pmd;

    if( handle >= MAX_MODULES ||
        mDesc[handle].mseg == 0 )     {
        return ERR_BAD_HANDLE;
    }

    pmd = &mDesc[handle];
    idstr = pmd->id;
    *major = pmd->v_major;
    *minor = pmd->v_minor;

    return TRUE;
}

flag_t DrvFreeModule(unsigned handle)
{
   /* Given module handle, release module and free
    * associated DOS memory.
    */

    struct ModuleDescriptor *pmd;

    if( handle >= MAX_MODULES ||
        mDesc[handle].mseg == 0 )     {
        return ERR_BAD_HANDLE;
    }

    pmd = &mDesc[handle];

    if( pmd->exit_f )
        (*pmd->exit_f)();      //do on-exit duties

    free( pmd->pblock );
    pmd->mseg = 0;           //indicate inactive
    pmd->pblock = NULL;
    pmd->mname[0] = '\0';
    --NumModules;           //keep count
    return TRUE;
}

flag_t DrvGetModName( unsigned handle,
                       char **name      )
{
   /* Given module handle, return ASCIIZ name of
    * active extension module, NULL if handle is
    * invalid or if module not active.
    */

    if( handle >= MAX_MODULES ||
        mDesc[handle].mseg == 0 )     {
        *name = NULL;
        return ERR_BAD_HANDLE;
    }

    *name = mDesc[handle].mname;
    return TRUE;
}

flag_t DrvGetModHandle( char *mname,
                         unsigned *handle)
{
   /* Given ASCIIZ name of active driver module,
    * return module handle.  Returns error if module
    * can't be located. Search is case sensitive.
    */

    register int i;

    for( i = 0; i < MAX_MODULES ; i++ )     {
        if( strcmp( mname, mDesc[i].mname ) == 0 ) {
                if( mDesc[i].mseg != 0 )    {
                        *handle = i;
                        return TRUE;
                }
        }
    }

    return ERR_MOD_NAME;
}

/* ----- End of File ------------------------------- */
