/*////////////////////////////////////////////////////////////////////////////

    MODULE:     PICEDIT.CPP
    
    Description:
        This is the implementation of the CPicEdit class.  This has all the
        methods that will subclass a CEdit control and respond to all key
        strokes with accordance to the set edit mask.
        

    Method List:
    
    History:                           
        Date        Email       Action
        ----------------------------------------------
        08/17/93    daveste     created
                
*/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#include "picdll.h"
#include "picedit.h"

#undef AFXAPP_DATA
#define AFXAPP_DATA AFXAPI_DATA

IMPLEMENT_DYNAMIC(CPicEdit, CEdit)

//MessageMap
BEGIN_MESSAGE_MAP(CPicEdit, CEdit)
    //{{AFX_MSG_MAP(CPicEdit)
    ON_WM_SETFOCUS()
    ON_WM_KILLFOCUS()
    ON_WM_KEYDOWN()
    ON_WM_CHAR()    
//    ON_WM_MOUSEACTIVATE()
    ON_WM_LBUTTONDOWN()
    ON_WM_MOUSEMOVE()
    ON_WM_LBUTTONUP()
    ON_MESSAGE(WM_COPY, OnEditCopy)
    ON_MESSAGE(WM_PASTE, OnEditPaste)
    ON_MESSAGE(WM_CUT, OnEditCut)
    ON_MESSAGE(EM_UNDO, OnEditUndo)
    ON_MESSAGE(EM_CANUNDO, OnEditCanUndo)
    ON_MESSAGE(EM_EMPTYUNDOBUFFER, OnEditEmptyUndoBuffer)
    ON_WM_PAINT()
    ON_MESSAGE(WM_CLEAR, OnClear)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

//Contructors and Destructors

PICAPI CPicEdit::CPicEdit()
{              
    //do an initializing here        
    
    //empty the data and mask strings
    m_strDisplay.Empty();
    m_strMask.Empty();
    
    //set the default prompt character
    m_chPrompt = ' ';
    
    //set the subclassed flag
    m_fSubclassed = FALSE;

    //initialize the mask characters string
    m_strMaskChars = "9ANLUX";
    
    //set the literal character
    m_chLiteralChar = '/';
    
    m_fAutoAdvance = FALSE;
    m_fAutoSelect = TRUE;
    m_fInsertModeCapable = FALSE;
    m_fInsertMode = FALSE;
    m_fMouseActivate = FALSE;
    m_fReadOnly = FALSE;
    m_fEatMouseMove = FALSE;
    
    m_sUndoBuffer.Empty();
    m_iLastKey = 0;

    m_nRepaintPos = -1;

} // CPicEdit()

PICAPI CPicEdit::~CPicEdit()
{
    //do any clean-up work
    
} // ~CPicEdit()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SubclassEdit

    Description:
        This method will perform the subclassing of the edit control.
        
        
    Inputs:     id of the edit control and a pointer to its parent window
    Outputs:    TRUE if successful and false if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SubclassEdit(UINT iCtlID, CWnd* pParentWnd)
{
    LONG lStyle = 0;        //style for this window
 
    //test our inputs
    ASSERT(this);
    if(!iCtlID || ! pParentWnd)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SubclassEdit -- Control iD or parent window class is invalid!\n");
        return FALSE;
    }                
    
    //this can only work if it's on a form view or dialog
    if(pParentWnd->IsKindOf(RUNTIME_CLASS(CFormView)) ||
       pParentWnd->IsKindOf(RUNTIME_CLASS(CDialog)))
    {
        //subclass the control
        if(SubclassDlgItem(iCtlID, pParentWnd))
        {
            m_fSubclassed = TRUE;
            //test for the read-only style
            lStyle = GetWindowLong(GetSafeHwnd(), GWL_STYLE);
            if(lStyle & ES_READONLY)
                m_fReadOnly = TRUE;
                
            return TRUE;
        }
        else
        {
            TRACE("CPicEdit::SubclassEdit -- Could not subclass edit control!\n");
            return FALSE;
        }
    }
    else    //parent is not of CDialog or CFormView class
    {  
        TRACE("CPicEdit::SubclassEdit -- Parent class is not a CFormView or CDialog!\n");
        return FALSE;
    }  
    
} // SubclassEdit()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SubclassEdit

    Description:
        This method will perform the subclassing of the edit control.  This 
        version will take just the handle to the window.
        
        
    Inputs:     window handle of the edit box
    Outputs:    TRUE if successful and false if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SubclassEdit(HWND hEdit)
{   
    LONG lStyle = 0;    //style for this window
    
    //test our inputs
    ASSERT(this);
    if(!IsWindow(hEdit))
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SubclassEdit -- window handle is invalid!\n");
        return FALSE;
    }                
    
    //subclass the control
    if(SubclassWindow(hEdit))
    {
        m_fSubclassed = TRUE;
        //test for the read-only style
        lStyle = GetWindowLong(GetSafeHwnd(), GWL_STYLE);
        if(lStyle & ES_READONLY)
            m_fReadOnly = TRUE;
        return TRUE;
    }
    else
    {
        TRACE("CPicEdit::SubclassEdit -- Could not subclass edit control!\n");
        return FALSE;
    }
    
} // SubclassEdit()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetMask

    Description:
        This method sets the edit mask for the control.
        
    Inputs:     reference to a mask string
    Outputs:    true if all is ok and false if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SetMask(LPCSTR lpszMask)
{                 
    //assign the mask to our member
    m_strMask = lpszMask;
    //and then parse the mask string
    ParseMaskString(m_strMask);
    
    //Build the display string
    BuildDisplayStr(m_strMask, m_strDisplay);
    
    RefreshDisplay();    
    
    //reset the modify flag so that we start clean
    SetModify(FALSE);
    
    //select the first editable character
    SetCursorPos(-1, 1);
    
    return TRUE;
} // SetMask()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetMask

    Description:
        This method sets the edit mask for the control.  However, this
        differs from the other SetMask in that it takes an ID to a pre-
        defined mask string.  CPicEdit comes with some predefined mask
        strings that the user of this class can choose from.  They are:
        
    ID                  Mask                    Description
    ---------------------------------------------------------------------------
    PEM_US_PHONE        (999) 999-9999          Standard US/Canada phone number
    PEM_US_PHONE_EX4    (999) 999-9999 [9999]   Phone with 4 digit extension
    PEM_US_PHONE_EX5    (999) 999-9999 [99999]  Phone with 5 digit extension
    PEM_SSN             999-99-9999             Social Security Number
    PEM_US_ZIP          99999                   US ZIP code (5 digit only)
    PEM_US_ZIP_4        99999-9999              US ZIP + 4 code
        
    Inputs:     ID to a predefined mask string
    Outputs:    true if all is ok and false if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SetMask(int iMaskID)
{                 
    CString sMask;
    
    switch(iMaskID)
    {             
        case PEM_US_PHONE:
            sMask = "(999) 999-9999";
            break;
        case PEM_US_PHONE_EX4:
            sMask = "(999) 999-9999 [9999]";
            break;
        case PEM_US_PHONE_EX5:
            sMask = "(999) 999-9999 [99999]";
            break;
        case PEM_SSN:
            sMask = "999-99-9999";
            break;
        case PEM_US_ZIP:
            sMask = "99999";
            break;
        case PEM_US_ZIP_4:
            sMask = "99999-9999";
            break;
        case PEM_US_ZIP_6:
            sMask = "99999-999999";
            break;

        //if we came to the default, then they didn't pass a valid ID in
        default:
            TRACE("CPicEdit::SetMask() -- Bad ID '%i' passed in!\n", iMaskID);
            return FALSE;
            break;
    }
    
    return SetMask(sMask);
} // SetMask()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetPromptChar

    Description:
        This method sets the prompt character for the edit.  This is the
        character that is displayed in an empty data position.
        
    Inputs:     new prompt char
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::SetPromptChar(char chNewPrompt)
{
    int iPosition = 0;
    int iMaskLen = m_strMask.GetLength();
    
    //test inputs
    ASSERT(this);
    if(!chNewPrompt)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SetPromptChar -- Invalid new prompt character\n");
        return;
    }
    
    //replace all instances of the previous prompt char in display string
    //with the new prompt char
    for (iPosition = 0; iPosition < iMaskLen; iPosition++)
    {
        if(!IsMaskLiteral(m_strMask.GetAt(iPosition)))
        {
            if(m_chPrompt == m_strDisplay.GetAt(iPosition))
                m_strDisplay.SetAt(iPosition, chNewPrompt);
        }    
    }
    
    m_chPrompt = chNewPrompt;
    
    RefreshDisplay();
    
} // SetPromptChar

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetEditData

    Description:
        This method allows the user of the class to set raw data into the
        control in one function.  It will clear out what it is in the control
        currenly, and put as much of the input text into the control as
        possible.  If there are more data characters sent in than in the
        mask, it will fill from left to right and discard the extras.
        
        The iStartPos is used to denote a starting position in the mask
        string.
        
    Inputs:     const reference to the string of data, starting position
                for the data, a flag for insert or overtype, and a flag
                to treat data that matches the prompt character as a space
                
    Outputs:    TRUE if successful and FALSE if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SetData(const CString& rstrData, int iStartPos, BOOL fInsert,
                                BOOL fIgnorePromptChars)
{                 
    int iDataCt = 0;                        //counter for the data string
    int iMaskCt = iStartPos;                //counter for the mask string
    int iMaskLen = m_strMask.GetLength();   //mask length
    int iDataLen = rstrData.GetLength();    //data length
    char chData;                            //data character buffer
    
    //check inputs
    ASSERT(this); 
    
    if(iMaskLen == 0)
        return FALSE;
        
    if(iStartPos < 0 || iStartPos >= iMaskLen)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SetData() -- Start position '%i' is invalid!\n", iStartPos);
        return FALSE;
    }
    
    //if the starting position is 0, rebuild the display string from scratch
    if(iStartPos == 0)
        BuildDisplayStr(m_strMask, m_strDisplay);
    
    //if the starting position isn't a data position, get the
    //next data position after the starting position
    if(IsMaskLiteral(m_strMask.GetAt(iMaskCt)))
        iMaskCt = GetNextDataPos(iMaskCt);
    
    //check to make sure we found a data position!
    if(iMaskCt >= iMaskLen || iMaskCt == -1)
        return FALSE;
    
    //now loop through the data string
    for(iDataCt = 0; iDataCt < iDataLen; iDataCt++)
    {
        if(iMaskCt >= iMaskLen || iMaskCt == -1)
            break;
        
        if(!IsMaskLiteral(m_strMask.GetAt(iMaskCt)))             
        {
            chData = rstrData.GetAt(iDataCt);
            
            if(IsCharValid(iMaskCt, chData, m_strMask))
            {
                if(fInsert)
                    //push char into string 
                    PushChar(iMaskCt, chData, FALSE);
                else
                    //overtype existing char
                    m_strDisplay.SetAt(iMaskCt, chData);
                
                //get the next data position
                iMaskCt = GetNextDataPos(iMaskCt);
            } //if(IsCharValid())
            else
            {
                //if the fIgnorePromptChars flag has been set to FALSE, then we will
                //treat any data character that matches the current prompt character as an
                //implicit blank data position.
                if(!fIgnorePromptChars && chData == m_chPrompt)
                    //get the next data position
                    iMaskCt = GetNextDataPos(iMaskCt);
                else
                    TRACE("CPicEdit::SetData -- Invalid character '%c' in input string\n", chData);
            }
        } //if(!IsMaskLiteral())
        
    } //end for
    
    //now refresh the display
    RefreshDisplay();
    
    return TRUE;
    
} // SetEditData()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetData

    Description:
        This method allows the user of the class to set raw data into the
        control in one function.  It will clear out what it is in the control
        currenly, and put as much of the input text into the control as
        possible.  If there are more data characters sent in than in the
        mask, it will fill from left to right and discard the extras.
        
        The iStartPos is used to denote a starting position in the mask
        string.
        
    Inputs:     LPCSTR to data and starting position
    Outputs:    TRUE if successful and FALSE if not

*/////////////////////////////////////////////////////////////////////////////
BOOL PICAPI CPicEdit::SetData(LPCSTR lpszData, int iStartPos, BOOL fInsert,
                                BOOL fIgnorePromptChars)
{                 
    int iDataCt = 0;                        //counter for the data string
    int iMaskCt = iStartPos;                //counter for the mask string
    int iMaskLen = m_strMask.GetLength();   //mask length
    int iDataLen = 0;                       //data length
    char chData;                            //data character buffer
    
    //check inputs
    ASSERT(this); 

    if(iMaskLen == 0)
        return FALSE;
        
    if(iStartPos < 0 || iStartPos >= iMaskLen)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SetData() -- Start position '%i' is invalid!\n", iStartPos);
        return FALSE;
    }
    
    if(!AfxIsValidString(lpszData))
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::SetData() -- Data string is invalid!\n");
        return FALSE;
    }

    //get the length of the data string
    iDataLen = lstrlen(lpszData);
    
    //if the starting position is 0, rebuild the display string from scratch
    if(iStartPos == 0)
        BuildDisplayStr(m_strMask, m_strDisplay);
    
    //if the starting position isn't a data position, get the
    //next data position after the starting position
    if(IsMaskLiteral(m_strMask.GetAt(iMaskCt)))
        iMaskCt = GetNextDataPos(iMaskCt);
    
    //check to make sure we found a data position!
    if(iMaskCt >= iMaskLen || iMaskCt == -1)
        return FALSE;

    //now loop through the data string
    for(iDataCt = 0; iDataCt < iDataLen; iDataCt++)
    {
        if(iMaskCt >= iMaskLen || iMaskCt == -1)
            break;
        
        if(!IsMaskLiteral(m_strMask.GetAt(iMaskCt)))             
        {
            chData = lpszData[iDataCt];
            
            if(IsCharValid(iMaskCt, chData, m_strMask))
            {
                if(fInsert)
                    //push char into string         
                    PushChar(iMaskCt, chData, FALSE);
                else
                    //overtype existing char
                    m_strDisplay.SetAt(iMaskCt, chData);
                
                //get the next data position
                iMaskCt = GetNextDataPos(iMaskCt);
            }
            else
            {
                //if the fIgnorePromptChars flag has been set to FALSE, then we will
                //treat any data character that matches the current prompt character as an
                //implicit blank data position.
                if(!fIgnorePromptChars && chData == m_chPrompt)
                    //get the next data position
                    iMaskCt = GetNextDataPos(iMaskCt);
                else
                    TRACE("CPicEdit::SetData -- Invalid character '%c' in input string\n", chData);
            }
            
        } //if(!IsMaskLiteral())
    } //end for
    
    //now refresh the display
    RefreshDisplay();
    
    return TRUE;
    
} // SetData()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     GetData

    Description:
        This method returns just the data from the edit control, not any
        of the mask characters.  The trim flag indicates if they want the
        trailing spaces trimmed from the data string.
        
    Inputs:     pointer to a string buffer
    Outputs:    filled string buffer

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::GetData(LPSTR lpszData, int iBufferLen, BOOL fTrim)
{                 
    int iMaskCt = 0;                        //mask string counter
    int iMaskLen = m_strMask.GetLength();   //length of mask string
    int iDataCt = 0;                        //data string counter
    
    //check inputs
    ASSERT(this);
    if(!lpszData)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::GetData() -- Invalid pointer to output string!\n");
        return;
    }
    
    if(iBufferLen < 1)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::GetData() -- Buffer length is less than 1 character!\n");
        return;
    }

    //clear buffer
    *lpszData = '\0';
    
    for (iMaskCt = 0; iMaskCt < iMaskLen; iMaskCt++)
    {
        if(iDataCt >= iBufferLen - 1)
            break;

        if(!IsMaskLiteral(m_strMask.GetAt(iMaskCt)))
        {
            //if the data value is just the prompt char,
            //then just add a space
            if(m_strDisplay.GetAt(iMaskCt) == m_chPrompt)
                *(lpszData + iDataCt) = ' ';
            else
                *(lpszData + iDataCt) = m_strDisplay.GetAt(iMaskCt);
                
            iDataCt++;
        }
    }   
    
    *(lpszData + iDataCt) = '\0';
    
    if(fTrim)
    {        
        //trim the trailing spaces
        iDataCt--;
        
        while(iDataCt >= 0 && *(lpszData + iDataCt) == ' ')
        {
            *(lpszData + iDataCt) = '\0';
            iDataCt--;
        }
    }
    
} // GetData()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     GetData

    Description:
        This method returns just the data from the edit control, not any
        of the mask characters.  The trim flag indicates if they want the 
        trailing spaces trimmed from the data string.
        
    Inputs:     pointer to a string buffer
    Outputs:    filled string buffer

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::GetData(CString& rsData, BOOL fTrim)
{                 
    int iMaskCt = 0;                        //mask string counter
    int iMaskLen = m_strMask.GetLength();   //length of mask string
    
    //check inputs
    ASSERT(this);
    
    //clear buffer
    rsData.Empty();
    
    for (iMaskCt = 0; iMaskCt < iMaskLen; iMaskCt++)
    {
        if(!IsMaskLiteral(m_strMask.GetAt(iMaskCt)))            
        {
            //if the data is the prompt char, just add a space
            if(m_strDisplay.GetAt(iMaskCt) == m_chPrompt)
                rsData += ' ';
            else
                rsData += m_strDisplay.GetAt(iMaskCt);
        }
    }
    
    if(fTrim)
    {
        int iTrimPos = rsData.GetLength() - 1;
        
        //find where to trim
        while(iTrimPos >= 0 && rsData.GetAt(iTrimPos) == ' ')
            iTrimPos--;
            
        //trim the string
        rsData = rsData.Left(iTrimPos + 1);
    }
    
} // GetData()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     ClearData()

    Description:
        This method will clear all the data out of the current control.  The
        control must be subclassed before this will have any effect.  It will
        essentially build the display string from scratch and refresh the
        display.
        
    Inputs:     none
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::ClearData(void)
{               
    //check inputs
    ASSERT(this);
    if(!m_fSubclassed)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::ClearData() -- Control is not subclassed yet!\n");
        return;
    }
    
    //rebuild the display string
    BuildDisplayStr(m_strMask, m_strDisplay);    
    
    //refresh the display
    RefreshDisplay();
    
} // ClearData()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     IsCharValid

    Description:
        This method will validate an incoming character with the input mask.
        
    Inputs:     character position, and character
    Outputs:    TRUE if valid, FALSE if not

*/////////////////////////////////////////////////////////////////////////////
BOOL CPicEdit::IsCharValid(UINT nPos, char& ch, const CString& rstrMask)
{
    char chMask;
    BOOL bRet = FALSE;
    int iSpread = ('a' - 'A');
    
    //check inputs
    ASSERT(this);
    if(ch == '\0')
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::IsCharValid -- Null character passed in!\n");
        return FALSE;
    }

    //if we are at the end of the input mask, this char is not valid
    if((int)nPos >= rstrMask.GetLength() || nPos < 0)
        return FALSE;
                     
                     
    chMask = rstrMask.GetAt(nPos);
    //if the position is on a literal char it's not valid
    if(IsMaskLiteral(chMask))
        bRet = FALSE;
    else
    {
        switch ((unsigned char)chMask - INFLATION_FACTOR)
        {
            case '9': //numerics only
            {
                if(ch >= '0' && ch <= '9')
                    bRet = TRUE;
                else
                    bRet = FALSE;
                break;
            }   
            
            case 'A': //alphas only
            {
                if((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z'))
                    bRet = TRUE;
                else
                    bRet = FALSE;
                break;
            }
            
            case 'N': //alpanumerics only
            {         
                if(   (ch >= 'a' && ch <= 'z') 
                   || (ch >= 'A' && ch <= 'Z')
                   || (ch >= '0' && ch <= '9')
                   )
                    bRet = TRUE;
                 else
                    bRet = FALSE;
                break;
            }
                
            case 'L': //force lower case
            {         
                if((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z'))
                {
                    if(ch >= 'A' && ch <= 'Z')
                        ch += iSpread;
                    bRet = TRUE;
                }
                else
                    bRet = FALSE;
                
                break;
            }

            case 'U': //force upper case
            {         
                if((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z'))
                {
                    if(ch >= 'a' && ch <= 'z')
                        ch -= iSpread;
                    bRet = TRUE;
                }
                else
                    bRet = FALSE;
                    
                break;
            }
            
            case 'X': //any ASCII character
            {         
                if(ch >= 0 || ch <= 128)
                    bRet = TRUE;
                else
                    bRet = FALSE;
                
                break;
            }
            
        }
    }
    
    return bRet;
    
} // IsCharValid

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     ParseMaskString

    Description:
        This method does the grunt work of parsing the mask string into our
        internal format.  The escape characters are dealt with here, and all
        mask characters are increased by the inflation factor to make them 
        easy to distinguish
        
    Inputs:     reference to the mask string
    Outputs:    changed mask string

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::ParseMaskString(CString& rstrMask)
{
    UINT iCounter = 0;
    UINT iMaskLen = 0;
    BOOL fEscapeChar = FALSE;
    char ch;
    CString sOutMask;
    
    ASSERT(this);

    //clear out the mask string
    sOutMask.Empty();
    
    iMaskLen = rstrMask.GetLength();
    
    //now go through the strMask character by character
    for(iCounter=0; iCounter < iMaskLen; iCounter++)
    {
        ch = rstrMask.GetAt(iCounter);
        //if the char is one of the mask chars and if it didn't have an escape
        //char before it, add the inflation factor to it
        if(m_strMaskChars.Find(ch) > -1 || fEscapeChar)
        {
            if(fEscapeChar)     
            {
                sOutMask += ch;
                fEscapeChar = FALSE;
            }
            else
                sOutMask += (ch + INFLATION_FACTOR);
        }    
        //if it is a literal escape character, skip it and set the flag
        else 
        {
            if(ch == m_chLiteralChar)
                fEscapeChar = TRUE;
            else
                sOutMask += ch;
        }
    }
    
    //copy the completed string into the master string
    rstrMask = sOutMask;
    
} // ParseMaskString()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     BuildDisplayStr

    Description:
        This builds the data string that will be used for display.  This should
        be called anytime the mask is reset.
        
    Inputs:     none
    Outputs:    DisplayString

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::BuildDisplayStr(CString& rstrMask, CString& rstrDisplay)
{               
    UINT iMaskCt = 0;   //counter in the mask string
    UINT iMaskLen = 0;  //length of mask string
    char ch = '\0';     //char buffer
    
    //empty the out string just in case
    rstrDisplay.Empty();
    
    //get the length of the mask and data strings
    iMaskLen = rstrMask.GetLength();
    
    //go through the mask string
    for (iMaskCt = 0; iMaskCt < iMaskLen; iMaskCt++)
    {
        //get the character in the iMaskCt position
        ch = rstrMask.GetAt(iMaskCt);
        
        //if the character is a literal, add it to the out string
        if(IsMaskLiteral(ch))
            rstrDisplay += ch;
        else    //else add the prompt character
            rstrDisplay += m_chPrompt;
    }
    
} // BuildDisplayStr  

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     SetCursorPos

    Description:
        This method sets the new cursor position in the edit box.  It will
        set it forward or backward of it's present placement depending on
        the flag passed in.  It will only place the cursor on positions
        that are legal for input (i.e. not on literal characters).
        
    Inputs:     old position, and direction to go (1 = forward, -1 = back)
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::SetCursorPos(int nOldPos, int iDirection, BOOL fSelect)
{
    int nMaskLen = m_strMask.GetLength();      //length of mask string
    int nNewPos =  nOldPos;                    //new position for cursor
    int nPosHolder = 0;                         //position holder
    
    //test inputs
    ASSERT(this);
    ASSERT(nOldPos >= 0 || nOldPos <= nMaskLen);
    
    //if the mask string is empty, just return
    if(nMaskLen < 1)    
        return nOldPos;
    
    //direction must be 1 or -1 (1=forward, -1=back)
    switch (iDirection)
    {
        case 1:
        {
            nNewPos = GetNextDataPos(nOldPos);

            //if we fell off the end, just set the position to the old one
            if(nNewPos == -1)
            {
                if(nOldPos != -1)
                {
                    //if the position right before where we are now is a
                    //literal, then keep it there.
                    if(IsMaskLiteral(m_strMask.GetAt(nOldPos)))
                        SetSel((int)nOldPos, (int)(nOldPos + fSelect));
                    else
                        SetSel((int)(nOldPos + 1), (int)(nOldPos + 1 + fSelect));
                }
                else
                    SetSel(0, 0);        
            }
            else
                SetSel(nNewPos, nNewPos + fSelect);
    
            break;
        }
        
        case -1:
        {         
            nNewPos = GetPrevDataPos(nOldPos);

            //if the new position is less than 0, we ran off the begining
            //so move to the previous control
            if(nNewPos < 0)
            {
                //set focus to previous control
                if(m_fAutoAdvance)
                    AdvanceFocusPrevious();
            }
            else //set the selection
            {
                SetSel(nNewPos, nNewPos + fSelect);
            }
            
            break;
        }
    }

    if(nNewPos < 0 || nNewPos >= nMaskLen)
        return -1;
    else
        return nNewPos;
    
    
} // SetCursorPos                  

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     GetNextDataPos

    Description:
        This method will return the next available data position in the
        mask string given an existing position.  If there are no more data
        positions, it will return -1.
        
    Inputs:     current position
    Outputs:    next data position or -1 if there are no more.

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::GetNextDataPos(int nOldPos)
{
    int nMaskLen = m_strMask.GetLength();      //length of mask string
    int nNewPos = nOldPos;                     //new position for cursor

    //test inputs
    ASSERT(this);
    ASSERT(nOldPos >= -1 || nOldPos <= nMaskLen);
    
    //if the mask string is empty, just return
    if(nMaskLen < 1)    
        return nOldPos;

    //if we are at the end of the mask, skip this code
    if(nNewPos < nMaskLen)
    {
        nNewPos++;
                
        //try positions forward until you find one or fall off
        while(nNewPos < nMaskLen)
        {     
            if(IsMaskLiteral(m_strMask.GetAt(nNewPos)))
                nNewPos++;
            else
                break;
        }
        
        //if we fell off the end without finding a data position
        //return -1
        if(nNewPos >= nMaskLen)
        {
            nNewPos = -1;
        }
    }

    return nNewPos;
    
} // GetNextDataPos()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     GetPrevDataPos

    Description:
        This method will return the previous available data position in the
        mask string given an existing position.  If there are no more data
        positions to the left of the old position, it will return the current
        position.
        
    Inputs:     current position
    Outputs:    previous data position or the current position if there are no
                more.

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::GetPrevDataPos(int nOldPos)
{
    int nMaskLen = m_strMask.GetLength();      //length of mask string
    int nNewPos = nOldPos;                     //new position for cursor
    
    //test inputs
    ASSERT(this);
    ASSERT(nOldPos >= 0 || nOldPos <= nMaskLen);
    
    //if the mask string is empty, just return
    if(nMaskLen < 1)    
        return nOldPos;

    nNewPos--;
            
    //try positions backward until you find one or fall off
    while(nNewPos >= 0)
    {     
        if(IsMaskLiteral(m_strMask.GetAt(nNewPos)))
            nNewPos--;
        else
            break;
    }

    return nNewPos;
    
} // GetPrevDataPos()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     ClearDataRange

    Description:
        This method will clear the data in the edit mask given a range to clear.
        This would be called when the user has selected more than one character
        with the mouse or keyboard and then pressed delete or backspace.  This
        method will clear the data, but not refresh the display--it is the
        responsibility of the calling function to refresh the display.
        
    Inputs:     start and end positions
    Outputs:    new start position

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::ClearDataRange(int iStart, int iEnd)
{
    int iMaskCt = 0;                        //counter in mask string
    int iMaskLen = m_strMask.GetLength();   //length of mask string
    
    //check inputs
    ASSERT(this);
    if (iStart < 0 || iEnd < 0
        || iEnd > iMaskLen
        || iEnd < iStart)
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::ClearDataRange -- Start and/or end positions are invalid!\n");
        return -1;
    }
    
    //if in insert mode, pull the characters back
    if(m_fInsertMode)
    {
        PullChar(iEnd, iStart);
    }
    else
    {       
        //go through the mask string and replace all data value with the prompt char
        for(iMaskCt = iStart; iMaskCt < iEnd; iMaskCt++)
        {
            if(!IsMaskLiteral(m_strMask.GetAt(iMaskCt)))
                m_strDisplay.SetAt(iMaskCt, m_chPrompt);
        }
    }
    
    return iStart;
    
} // ClearDataRange

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnKeyDown

    Description:
        This method responds to all the KeyDown events for the subclassed control
        
    Inputs:     char, repeat count and flags
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
    int iPosition = 0;          //position of the cursor in the edit box

    //Get the Sel position
    iPosition = LOWORD(GetSel());
    
    //processing for virtual keys
    switch(nChar)
    {
        case VK_BACK:
        {
            if(m_fReadOnly)
            {
                m_fGotKey = TRUE;
                break;
            }
            
            int iEndPos = HIWORD(GetSel());

            //if we have a multiple selection, handle it differently
            if(iPosition != iEndPos)
            {
                if(m_iLastKey == VK_BACK)
                    PullChar(iEndPos, iPosition, FALSE);
                else
                    PullChar(iEndPos, iPosition);
                    
                RefreshDisplay();
                iPosition = GetNextDataPos(iPosition - 1);
                if(iPosition == -1)
                    iPosition == GetPrevDataPos(m_strMask.GetLength()) + 1;
                SetSel(iPosition, iPosition);                    
            }
            else
            {
                int iPrevPos = GetPrevDataPos(iPosition);
                    
                //if this is the leftmost data position,
                //then ignore the backspace key
                if(iPrevPos != -1)
                {
                     //set the repaint position
                    if(m_nRepaintPos == -1)
                        m_nRepaintPos = GetRepaintPos(iPrevPos);

                    if(m_iLastKey == VK_BACK)
                        PullChar(iPosition, iPrevPos, FALSE);
                    else
                        PullChar(iPosition, iPrevPos);

                    RefreshDisplay();
                    SetSel(iPrevPos, iPrevPos);
                }
            }    
            
            m_fGotKey = TRUE;
            break;
        }
        
        case VK_DELETE:
        {
            if(m_fReadOnly)
            {
                m_fGotKey = TRUE;
                break;
            }
            
            int iEndPos = HIWORD(GetSel());
            int iCounter = 0;
            
            if(m_nRepaintPos == -1)
                m_nRepaintPos = GetRepaintPos(iPosition);
                
            //if we have a multiple selection, handle it differently
            if(iPosition != iEndPos)
            {
                if(m_iLastKey == VK_DELETE)
                    PullChar(iEndPos, iPosition, FALSE);
                else
                    PullChar(iEndPos, iPosition);
                 
                RefreshDisplay();
                iPosition = GetNextDataPos(iPosition - 1);
                if(iPosition == -1)
                    iPosition = GetPrevDataPos(m_strMask.GetLength()) + 1;
                SetSel(iPosition, iPosition);                    
            }
            else
            {
                int iNextPos = GetNextDataPos(iPosition);
                if(iNextPos != -1)
                {
                    //make sure we are not at the end of the mask
                    if(iPosition < m_strMask.GetLength())
                    {
                        if(m_iLastKey == VK_DELETE)
                            PullChar(iNextPos, iPosition, FALSE);
                        else
                            PullChar(iNextPos, iPosition);
                            
                        RefreshDisplay();
                        SetSel(iPosition, iPosition);
                    }
                    else //if were at the end of the mask, then reset the repaint pos
                        m_nRepaintPos = -1;
                }
                else
                {
                    if(m_iLastKey == VK_DELETE)
                        PullChar(iPosition + 1, iPosition, FALSE);
                    else
                        PullChar(iPosition + 1, iPosition);
                        
                    RefreshDisplay();
                    SetSel(iPosition, iPosition);
                }
            } //else

            m_fGotKey = TRUE;
            break;
        }
            
        case VK_LEFT:
        {
            int iEndPos = HIWORD(GetSel());

            //test for a multi select
            if(iPosition != iEndPos)
            {
                //test for the shift key
                if((::GetKeyState(VK_SHIFT)) < 0)
                    CEdit::OnKeyDown(nChar, nRepCnt, nFlags);  
                else
                    SetCursorPos(iEndPos, -1);
                
            }
            else    //single select state
            {
                //test for the shift key
                if((::GetKeyState(VK_SHIFT)) < 0)
                    CEdit::OnKeyDown(nChar, nRepCnt, nFlags);
                else
                    SetCursorPos(iPosition, -1);
            }        

            m_fGotKey = TRUE;
            break;
        }
        
        case VK_RIGHT:
        {
            int iEndPos = HIWORD(GetSel());

            //test for a multi select
            if(iPosition != iEndPos)
            {
                //test for the shift key
                if((::GetKeyState(VK_SHIFT)) < 0)
                    CEdit::OnKeyDown(nChar, nRepCnt, nFlags);
                else                     
                    SetCursorPos(iEndPos, 1);
            }
            else    //single select state
            {
                //test for the shift key
                if((::GetKeyState(VK_SHIFT)) < 0)
                    CEdit::OnKeyDown(nChar, nRepCnt, nFlags);
                else
                    SetCursorPos(iPosition, 1);
            }        
    
            m_fGotKey = TRUE;
            break;
        }    

        case VK_END:
            //test for shift
            if((::GetKeyState(VK_SHIFT)) < 0)
            {   
                SetSel(iPosition, m_strMask.GetLength());                                                     
            }
            else
            {   
                //set the cursor just after the last data position.
                //iPosition = SetCursorPos(m_strMask.GetLength(), -1);
                //SetSel(iPosition + 1, iPosition + 1);
                iPosition = FindLastFilledDataPos(m_strMask.GetLength());
                SetSel(iPosition, iPosition);
            }
                
            m_fGotKey = TRUE;
            break;
        
        case VK_HOME:
            //test for shift
            if((::GetKeyState(VK_SHIFT)) < 0)
            {   
                SetSel(0, iPosition);
            }
            else
                SetCursorPos(-1, 1);

            m_fGotKey = TRUE;
            break;
        
        case VK_INSERT:
            //if insert mode capable is true then toggle insert mode
            if(m_fInsertModeCapable && !m_fReadOnly)
            {
                m_fInsertMode = !m_fInsertMode;
                m_fGotKey = TRUE;
            }    
            break;    
        
        default:
            m_fGotKey = FALSE;
    }
    
    m_iLastKey = nChar;
} // OnKeyDown()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnSetFocus

    Description:
        This method responds to the set focus message from windows
        
    Inputs:     previous focused window
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::OnSetFocus(CWnd* pOldWnd)
{
    CEdit::OnSetFocus(pOldWnd);
/*    
        //if the auto select property is true, select the entire text
        //if the controls was activated by a mouse click, then don't select
        //the contents, regardless of what m_fAutoSelect is.
        if(m_fAutoSelect && !m_fMouseActivate)
                SetSel(0, GetWindowTextLength());
        else
            SetCursorPos(-1, 1);
*/
        if(m_fAutoSelect)
                SetSel(0, GetWindowTextLength());
        else
            SetCursorPos(-1, 1);
            
} // OnSetFocus

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnLButtonDown

    Description:
        This method responds to the WM_LBUTTONDOWN message from windows
        
    Inputs:     see windows sdk
    Outputs:    see windows sdk

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::OnLButtonDown(UINT nFlags, CPoint point)
{
    //call the base class so the caret gets set right
    CEdit::OnLButtonDown(nFlags, point);
    
    //now check to see if the cursor was put after the last piece
    //of data.  if so then we need to put it after the last filled data
    //position or if the whole thing is empty, we need to put it at the
    //first data position.
    int nStartSel, nEndSel;
    GetSel(nStartSel, nEndSel);
    //only do this if there isn't a multi select now.
    if(nStartSel == nEndSel)
    {
        //find the last filled data position to the left
        //we will re-use nEndSel since its original value
        //isn't needed anymore
        nEndSel = FindLastFilledDataPos(nStartSel);
        if(nEndSel != nStartSel)
        {   
            //if the new position is different from the current one,
            //set the selection.
            SetSel(nEndSel, nEndSel);
            //set the eat mouse move flag on so that we eat any mouse moves
            //that occur before the next lbuttonup
            //this is needed since a little jiggle of the mouse before
            //the lbuttonup and it would select from the nEndSel to the
            //nStartSel, which would look bad.
            m_fEatMouseMove = TRUE;
            m_ptMouseMoveRef = GetCaretPos();
        }//if(nEndSel != nStartSel)
    }//if(nStartSel == nEndSel)

} // OnLButtonDown()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     FindLastFilledDataPos

    Description:
        This method will search the current display string starting with
        the position passed in and find the last filled data position to
        the left of the starting position.  If the starting position is
        filled in, then it will just return the starting position.
        
    Inputs:     starting position
    Outputs:    last filled data pos

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::FindLastFilledDataPos(int nStartPos)
{
    int nNextPos = 0;
    
    //if the start pos is greater than the length of the mask,
    if(nStartPos >= m_strMask.GetLength())
    {
        //check to see if the last position in the mask is a data position
        //if so then if it's filled, return the length of the mask
        //if not then subtract 1 from the start pos, so we don't assert
        //when checking the mask later
        if(!IsMaskLiteral(m_strMask.GetAt(nStartPos - 1)))
            if(m_strDisplay.GetAt(nStartPos - 1) != m_chPrompt)
                return nStartPos;

        nStartPos--;
    }
    
    //check to see if the next data position to the right is filled
    //if so, then don't change the position
    if(IsMaskLiteral(m_strMask.GetAt(nStartPos)))
        nNextPos = GetNextDataPos(nStartPos);
    else
        nNextPos = nStartPos;
    
    if(nNextPos != -1)
        if(m_strDisplay.GetAt(nNextPos) != m_chPrompt)
            return nStartPos;
    
    //check to see if this position is a data position.
    if(IsMaskLiteral(m_strMask.GetAt(nStartPos)))
        nStartPos = GetPrevDataPos(nStartPos);
    
    //keep checking in a loop for a filled data pos
    while(nStartPos > 0 && m_strDisplay.GetAt(nStartPos) == m_chPrompt)
    {
        nStartPos = GetPrevDataPos(nStartPos);
    }
    
    //if there were no data positions to the left of the start pos
    //then return the first data position
    if(nStartPos <= 0)
        if(nStartPos == 0 && m_strDisplay.GetAt(nStartPos) != m_chPrompt)
            return 1;
        else
            return GetNextDataPos(-1);
    else
        //return one greater so that it will line up on the
        //right side of the position
        return nStartPos + 1;
        
} //FindLastFilledDataPos()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnMouseMove

    Description:
        This method will respond to mouse move messages for the control.  It
        will pass them on if the m_fEatMouseMove flag is not turned on.
        
    Inputs:     see windows sdk
    Outputs:    see windows sdk

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::OnMouseMove(UINT nFlags, CPoint point)
{
    if(m_fEatMouseMove && point.x >= m_ptMouseMoveRef.x)
        return;
    else
        CEdit::OnMouseMove(nFlags, point);
} //OnMouseMove()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnLButtonUp

    Description:
        This method responds to the WM_LBUTTONUP message from windows
        
    Inputs:     see windows sdk
    Outputs:    see windows sdk

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::OnLButtonUp(UINT nFlags, CPoint point)
{
    m_fEatMouseMove = FALSE;
    CEdit::OnLButtonUp(nFlags, point);
}

#if 0
/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnMouseActivate

    Description:
        This method responds to the On Mouse Activate message from windows
        
    Inputs:     see windows sdk
    Outputs:    see windows sdk

*/////////////////////////////////////////////////////////////////////////////
/*
int PICAPI CPicEdit::OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message)
{
    //mouse activate happens more that on the set focus operation, so make
    //sure that we don't set mouse activate to true when it is
    if(!m_fMouseActivate)
        m_fMouseActivate = TRUE;
    return (CEdit::OnMouseActivate(pDesktopWnd, nHitTest, message));
} // OnMouseActivate()
*/
#endif

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnKillFocus

    Description:
        This method responds to the WM_KILLFOCUS message from windows
        
    Inputs:     see windows sdk
    Outputs:    see windows sdk

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::OnKillFocus(CWnd* pNewWnd)
{
    m_fMouseActivate = FALSE;    
    CEdit::OnKillFocus(pNewWnd);
}                               

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnChar

    Description:
        This method responds to all the char events for the subclassed control
        
    Inputs:     char, repeat count and flags
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void PICAPI CPicEdit::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
    if(!m_fGotKey && !m_fReadOnly)
    {
        //set the last key member
        m_iLastKey = nChar;
        
        int iPosition = 0;                      //position of selection start
        int iEndPos = 0;                        //end selection position
        //char str[2];                            //temporary string variable
        int iMaskLen = m_strMask.GetLength();   //length of input mask
    
        
        //Get the Sel position
        iPosition = LOWORD(GetSel());
        iEndPos   = HIWORD(GetSel());
        
        if(iPosition != iEndPos)
        {
            //make sure that the starting position is on a valid data field.
            //if iPosition is not on a data field, then advance it
            if(IsMaskLiteral(m_strMask.GetAt(iPosition)))
                iPosition = SetCursorPos(iPosition, 1);
                
            //if iPosition >= iEndPos, they weren't highlighting a
            //data field at all.  Beep and Get out
            if(iPosition >= iEndPos || iPosition == -1)
            {
                MessageBeep(MB_ICONEXCLAMATION);
                m_fGotKey = FALSE;
                return;
            }    
        }    
        else //not in multi-select
        {
            //make sure that the starting position is on a valid data field.
            //if iPosition is not on a data field, then advance it
            iPosition = SetCursorPos(iPosition - 1, 1);
                
            //if iPosition = -1 then there wasn't a data position to fill
            if(iPosition == -1)
            {
                MessageBeep(MB_ICONEXCLAMATION);
                m_fGotKey = FALSE;
                return;
            }    
        }
        
        //if it's not a virtual key, then see if it's valid for it's position
        if(IsCharValid(iPosition, (char&)nChar, m_strMask))   
        {         
            //get the repaint position
            if(m_nRepaintPos == -1)
                m_nRepaintPos = GetRepaintPos(iPosition);
            
            //if we have a multi-select state, clear the text first
            if(iPosition != iEndPos)
            {
                PullChar(iEndPos, iPosition);
            }
                
            //if in insert mode, push characters down one
            if(m_fInsertMode)
            {
                //if we can't push anymore (mask is full), then beep
                if(!PushChar(iPosition, nChar))
                    MessageBeep(MB_ICONEXCLAMATION);
                else // push was successful
                {
                    //if we are in the last data position,
                    //then auto-advance.
                    if(m_fAutoAdvance && (GetNextDataPos(iPosition) == -1))
                        AdvanceFocusNext();
                }     
                
            }
            else // do normal overtyping
            {
                m_strDisplay.SetAt(iPosition, nChar);

                RefreshDisplay();

                //set the caret to one place past where they were
                iPosition = SetCursorPos(iPosition, 1);
                //if we are at the end of the mask and the auto advance flag is true
                //advance to the next dialog control
                if(iPosition == -1 && m_fAutoAdvance)
                    AdvanceFocusNext();
            }
        } //if(IsCharValid(iPosition, (char&)nChar, m_strMask))   
        else
            MessageBeep(MB_ICONEXCLAMATION);
    } //if(!m_fGotKey && !m_fReadOnly)                 
    
    m_fGotKey = FALSE;
}

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     PushChar

    Description:
        This method will push all data characters after the push point down
        one.  If there is no more room to push, it will return FALSE.
        
    Inputs:     first character to push and a flag if we should refresh the
                display or not.
    Outputs:    TRUE if all chars are pushed, FALSE if not

*/////////////////////////////////////////////////////////////////////////////
BOOL CPicEdit::PushChar(int iStartPos, char ch, BOOL fRefreshDisplay)
{
    int iLastDataPos;
    int iPosCt;
    
    //check inputs
    ASSERT(this);
    if(iStartPos < 0 || iStartPos > m_strMask.GetLength())
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::PushChar() -- Start position '%i' is invalid!\n", iStartPos);
        return FALSE;
    }
    
    //get the index of the last data position
    iLastDataPos = GetPrevDataPos(m_strMask.GetLength());
    
    //now go backwards until you find an actual data character
    iPosCt = iLastDataPos;
    while (iPosCt >= 0 && m_strDisplay.GetAt(iPosCt) == m_chPrompt)
        iPosCt = GetPrevDataPos(iPosCt);
    
    //if we are still in the last data position, return false and get out
    if (iPosCt == iLastDataPos)
        return FALSE;
    else
    {
        while (iPosCt >= iStartPos)
        {
            //shift the character forward one data place
            m_strDisplay.SetAt(GetNextDataPos(iPosCt),
                                m_strDisplay.GetAt(iPosCt));
            //move iPosCt to the previous data position                                
            iPosCt = GetPrevDataPos(iPosCt);
        }
        
        //set the input char in the starting position
        m_strDisplay.SetAt(iStartPos, ch);
        
        if(fRefreshDisplay)
        {
            //refresh the display
            RefreshDisplay();
            
            //set the selection foward one data position
            int iNextPos = GetNextDataPos(iStartPos);
            if(iNextPos != -1)
                SetSel(iNextPos, iNextPos);
            else
                SetSel(iStartPos + 1, iStartPos + 1);
        }        
        
        return TRUE;
    }
} // PushChar()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     PullChar

    Description:
        This method, called only when in insert mode will pull the data 
        characters backward.  It will start at the from position and begin
        shifting the characters to the To Position until it reaches the end
        of the string.
        
    Inputs:     Starting position
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void CPicEdit::PullChar(int iFromPos, int iToPos, BOOL fSetUndoBuffer)
{              
    int iToCt;              //a position index counter

    //check inputs
    ASSERT(this);
    if(iFromPos < 0 || iFromPos > m_strMask.GetLength())
    {
        ASSERT(FALSE);
        TRACE("CPicEdit::PullChar() -- Start position '%i' is invalid!\n", iFromPos);
        return;
    }

    //if the to position is on a literal position,
    //then try to move it foward
    if(IsMaskLiteral(m_strMask.GetAt(iToPos)))
    {
        iToPos = GetNextDataPos(iToPos);
        //if the to pos is now greater than the from pos,
        //there was no data to pull
        if(iToPos > iFromPos)
            return;
    }

    //if the from pos is = to the length of m_strMask, set to -1
    if(iFromPos == m_strMask.GetLength())
        iFromPos = -1;
    else        
    {
        //if the from position is on a literal position,
        //then try to move it foreward
        if(IsMaskLiteral(m_strMask.GetAt(iFromPos)))
        {
            iFromPos = GetNextDataPos(iFromPos);
        }
    }
    
    if(fSetUndoBuffer)
        //set the undo buffer
        m_sUndoBuffer = m_strDisplay;
        
    iToCt = iToPos;
    
    while (iFromPos >= 0)
    {                                                 
        //move the character
        m_strDisplay.SetAt(iToCt, m_strDisplay.GetAt(iFromPos));
        //move the pointers
        iFromPos = GetNextDataPos(iFromPos);
        iToCt = GetNextDataPos(iToCt);
    }
    
    //clear out the rest of the data from the iToCt on
    while (iToCt >= 0)
    {
        m_strDisplay.SetAt(iToCt, m_chPrompt);
        iToCt = GetNextDataPos(iToCt);
    }
    

} // PullChar()       

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnEditCopy

    Description:
        This method will handle copy requests from the control.  It will
        copy just the data to the clipboard, so that it can be pasted back
        into the control again
        
    Inputs:     wParam and lParam are not used.
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditCopy(WPARAM, LPARAM)
{              
    int nSelStart = 0;          //selection starting point
    int nSelEnd = 0;            //selection ending point
    int nSelCt = 0;             //selection counter for loop
    int nDataCt = 0;            //data counter for loop
    LPSTR lpszSelData = NULL;   //string variable to hold the selected data
    HGLOBAL hMem = NULL;        //memory module for clipboard
    
    //get the current selection
    GetSel(nSelStart, nSelEnd);
    
    //allocate memory for the data string and lock it
    hMem = GlobalAlloc(GHND, nSelEnd - nSelStart);
    if(hMem == NULL)
    {
        TRACE("CPicEdit::OnEditCopy()--Couldn't allocate memory!\n");
        return 0;
    }    
        
    lpszSelData = (LPSTR)GlobalLock(hMem);
    if(lpszSelData == NULL)
    {
        TRACE("CPicEdit::OnEditCopy()--Couldn't lock memory!\n");
        GlobalFree(hMem);
        return 0;
    }    
            
    //get just the selected data
    for (nSelCt = nSelStart; nSelCt < nSelEnd; nSelCt++)
    {
        if(!IsMaskLiteral(m_strMask.GetAt(nSelCt)))            
        {
            //if the data is the prompt char, just add a space
            if(m_strDisplay.GetAt(nSelCt) == m_chPrompt)
                lpszSelData[nDataCt] = ' ';
            else
                lpszSelData[nDataCt] = m_strDisplay.GetAt(nSelCt);
            
            nDataCt++;    
        }
    }

    //unlock the data
    GlobalUnlock(hMem);
        
    //open the clipboard
    if(OpenClipboard())
    {
        //empty the clipboard
        EmptyClipboard();
        //set the new data into the clipboard
        if(SetClipboardData(CF_TEXT, hMem) == NULL)
            //setdata failed! free the memory
            GlobalFree(hMem);
            
        CloseClipboard();    
    }
    
    return 1;
} // OnEditCopy()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnEditPaste

    Description:
        This method will handle pasting data from the clipboard into the
        control.  It will get the data out of the clipboard and use SetData()
        to put the data into the control.
        
    Inputs:     wParam and lParam are not used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditPaste(WPARAM, LPARAM)
{
    int nSelStart = 0;          //selection start
    int nSelEnd = 0;            //selection end
    int nNextDataPos = 0;       //next data position after selection
    LPSTR lpszData = NULL;      //pointer to data in clipboard
    HMODULE hClipMem = NULL;     //memory handle of clipboard data
    
    //if the control is read-only, don't do anything!
    if(m_fReadOnly)
        return 0;
   
    //check to see if there is data to paste!
    if(!IsClipboardFormatAvailable(CF_TEXT))
        return 0;
    
    //get the current selection
    GetSel(nSelStart, nSelEnd);
    
    //if the start sel is at the end of the mask, just return
    if(nSelStart < 0 || nSelStart >= m_strMask.GetLength())
        return 0;
    
    //if there is a multi-selection clear the data first
    if(nSelEnd != nSelStart)
    {
        PullChar(nSelEnd, nSelStart);
    }
    
    //get the data from the clipboard
    if(OpenClipboard())
    {
        hClipMem = (HMODULE)GetClipboardData(CF_TEXT);
        if(hClipMem != NULL)
            lpszData = (LPSTR)GlobalLock(hClipMem);
        
        if(lpszData != NULL)
        {
            //save the current text in the undo buffer
            m_sUndoBuffer = m_strDisplay;
            
            //set the data into the control
            SetData((LPCSTR)lpszData, nSelStart, TRUE);
        }
        
        //unlock memory and close the clipboard
        GlobalUnlock(hClipMem);
        CloseClipboard();
    }
    
    RefreshDisplay();
    
    //if the old selection was not on a data position,
    //get the next data position, and if there is one,
    //set the selection there
    if(IsMaskLiteral(m_strMask.GetAt(nSelStart)))
    {
        nNextDataPos = GetNextDataPos(nSelStart);
        if(nNextDataPos != -1)
            nSelStart = nNextDataPos;
    }

    SetSel(nSelStart, nSelStart);
        
    return 1;
} // OnEditPaste()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnEditCut

    Description:
        This method will cut data from the control and put it into the
        clipboard.  It will first copy the data selected and then clear it.
        
    Inputs:     none used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditCut(WPARAM, LPARAM)
{
    int nSelStart = 0;          //selection start
    int nSelEnd = 0;            //selection end
    CString sDisplaySave;       //a backup of the current display string

    OnEditCopy(0, 0);

    //if the control is read-only, don't remove the data!
    if(m_fReadOnly)
        return 0;
    
    //save off the current display string
    TRY{sDisplaySave = m_strDisplay;}
    CATCH_ALL(e) {return 0;}
    END_CATCH_ALL
        
    //get the current selection
    GetSel(nSelStart, nSelEnd);
    
    //if there is a multi-selection clear the data
    if(nSelEnd != nSelStart)
    {
        PullChar(nSelEnd, nSelStart);
    }
    
    //if the display string didn't change, don't refresh the display
    //because that will set the modify flag to TRUE
    if(m_strDisplay != sDisplaySave)
        RefreshDisplay();
    
    SetSel(nSelStart, nSelStart);
    return 0;
        
} // OnEditCut()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnClear

    Description:
        This method will handle the WM_CLEAR messages that come our way.
        it will do what the normal delete key would do.
        
    Inputs:     none used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnClear(WPARAM, LPARAM)
{
    OnKeyDown(VK_DELETE, 0, 0);
    return 1;
} // OnClear()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnEditUndo

    Description:
        This method will handle doing an undo for the control.  It will
        set the m_strDisplay to be the m_sUndoBuffer and refresh the display.
        
    Inputs:     none used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditUndo(WPARAM, LPARAM)
{
    //don't undo if the undo buffer is empty or the control is read-only
    if(m_sUndoBuffer.IsEmpty() || m_fReadOnly)
        return 0;
        
    CString sCurDisplay;
    //swap the current display with the undo buffer    
    sCurDisplay = m_strDisplay;
    m_strDisplay = m_sUndoBuffer;
    m_sUndoBuffer = sCurDisplay;
    
    RefreshDisplay();
    //select all the text in the control
    SetSel(0, -1);
    return 1;
} // OnEditUndo()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnCanUndo

    Description:
        This method will test to see if we can undo something.
        
    Inputs:     none used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditCanUndo(WPARAM, LPARAM)
{
    //they can undo if the undo buffer is not empty and the
    //control is not read-only
    return (LRESULT)(!m_sUndoBuffer.IsEmpty() && !m_fReadOnly);
} // OnEditCanUndo()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnEditEmptyUndoBuffer

    Description:
        This method handles emptying the undo buffer for the control.
        
    Inputs:     not used
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg LRESULT CPicEdit::OnEditEmptyUndoBuffer(WPARAM, LPARAM)
{
    m_sUndoBuffer.Empty();
    return 0;
} // OnEditEmptyUndoBuffer()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     OnPaint

    Description:
        This method will override the default painting of the control and
        add a little optimization.  If the person is just typing into the
        control, it will only repaint the portion that it needs to, namely
        the area to the right of the caret.  The portion to the left stays
        the same, so there is no reason to redraw it, so this will validate
        that area before calling the default paint procedure.
        
    Inputs:     none
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
afx_msg void CPicEdit::OnPaint()
{
    //if the nRepaintPos member is not -1, then
    //do our optimization.  The -1 is a cue that the whole control
    //needs to be repainted.
    if(m_nRepaintPos != -1)
    {
        //RECT rcUpdate = {0,0,0,0};
        RECT rcEdit = {0,0,0,0};
        
        //GetUpdateRect(&rcUpdate, FALSE);
        GetRect(&rcEdit);
        
        rcEdit.right = rcEdit.left + m_nRepaintPos - 1;
        ValidateRect(&rcEdit);
        
        m_nRepaintPos = -1;
    }    
    
    CEdit::OnPaint();
} // OnPaint()

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     GetRepaintPos

    Description:
        This methos will determine the repaint position which is the left
        most point that needs to be repainted.  To do some optimization, we
        will only repaint from this position to the right, leaving the area
        to the left of the repaint position untouched.  This will eliminate
        unecessary repainting and annoying flicker.  The method will
        translate the index to a coordinate using GetTextExtent
        
    Inputs:     index of left-most selection
    Outputs:    repaint position

*/////////////////////////////////////////////////////////////////////////////
int CPicEdit::GetRepaintPos(int iIndex)
{
    HDC hdc = NULL;
    int iRetPos = -1;
    
    //because edit controls don't keep their current font in the HDC,
    //we need to get it and select it into the HDC before doing the
    //calculations
    HFONT hCurFont = (HFONT)SendMessage(WM_GETFONT, 0, 0);
    
    //get the current device context
    hdc = ::GetDC(m_hWnd);                                            
    
    if(hdc != NULL)
    {
        //select the current font into the HDC
        HFONT hOldFont = (HFONT)::SelectObject(hdc, hCurFont);
        
        //now get the width of the text left of the index value
        iRetPos = LOWORD(::GetTextExtent(hdc, 
                                        (LPCSTR)(m_strDisplay.Left(iIndex)),
                                        iIndex));
        
        //restore the old font
        SelectObject(hdc, hOldFont);
        
        //release DC
        ::ReleaseDC(m_hWnd, hdc);
    }

    return iRetPos;
} // GetRepaintPos()

/*////////////////////////////////////////////////////////////////////////////
    DDX Functions for Class Wizard    
*/////////////////////////////////////////////////////////////////////////////
/*////////////////////////////////////////////////////////////////////////////
    METHOD:     DDX_PicEditControl

    Description:
        This global function can be used in the DoDataExchange member
        function of a window or dialog.  It will call the subclass edit
        for a control if it has not already been subclassed.
        
    Inputs:     data exchange obj, control ID, and ref to control obj
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void AFXAPI DDX_PicEditControl(CDataExchange* pDX, int nIDC, CPicEdit& rPicEdit,
                                int iMaskID)
{   
    if (rPicEdit.m_hWnd == NULL)    // not subclassed yet
    {
        ASSERT(!pDX->m_bSaveAndValidate);
        HWND hWndCtrl = pDX->PrepareCtrl(nIDC);
        if (!rPicEdit.SubclassEdit(hWndCtrl))
        {
            ASSERT(FALSE);      // possibly trying to subclass twice ?
            AfxThrowNotSupportedException();
        }
        else //subclassed successfully
        {
            if(iMaskID != NULL)
            {
                if(!rPicEdit.SetMask(iMaskID))
                {
                    ASSERT(FALSE);      // set mask failed! Illegal mask.
                    AfxThrowNotSupportedException();
                }
            } //if(iMaskID != NULL)
        } //end else
    } //if(rPicEdit.m_hWnd == NULL)

} //DDX_PicEditControl

//This version takes a string for a mask
void AFXAPI DDX_PicEditControl(CDataExchange* pDX, int nIDC, CPicEdit& rPicEdit,
                                LPCSTR lpszMask)
{   
    if (rPicEdit.m_hWnd == NULL)    // not subclassed yet
    {
        ASSERT(!pDX->m_bSaveAndValidate);
        HWND hWndCtrl = pDX->PrepareCtrl(nIDC);
        if (!rPicEdit.SubclassEdit(hWndCtrl))
        {
            ASSERT(FALSE);      // possibly trying to subclass twice ?
            AfxThrowNotSupportedException();
        }
        else //subclassed successfully
        {
            if(lpszMask != NULL)
            {
                if(!rPicEdit.SetMask(lpszMask))
                {
                    ASSERT(FALSE);      // set mask failed! Illegal mask.
                    AfxThrowNotSupportedException();
                }
            } //if(lpszMask != NULL)
        } //end else
    } //if(m_hWnd != NULL)
} //DDX_PicEditControl

/*////////////////////////////////////////////////////////////////////////////
    METHOD:     DDX_PicEditText

    Description:
        This global method is for use in a DoDataExchange block to set
        and get the data from a CPicEdit control.  It will call GetData()
        or SetData() depending on which direction the data exchange is
        going.
        
    Inputs:     data exchange obj, ref to a CPicEdit object, and a ref
                to the CString obj.
    Outputs:    none

*/////////////////////////////////////////////////////////////////////////////
void AFXAPI DDX_PicEditText(CDataExchange* pDX, CPicEdit& rPicEdit, CString& value,
                            BOOL fTrim)
{
    if(rPicEdit.m_hWnd != NULL)
    {
        if (pDX->m_bSaveAndValidate)
            rPicEdit.GetData(value, fTrim);
        else
            rPicEdit.SetData(value);
    }
} //DDX_PicEditText
