// File:   ccache.cpp
// Author: Pete Rail November 1994
// Copyright (c) EDS, 1994.
// Readers of Windows/DOS Developer's Journal are hereby
// granted a nonexclusive, royalty-free license to use, modify 
// and distribute this program.

#include "stdafx.h"
#include "ccache.h"

CCache :: CCache( const char* name, int numEntries /* =1 */ )
{ 
    m_Name       = name;
    m_EntryLimit = __max( numEntries, 1 );  // allow no less than 1 entry 
in cache
    InitStats();
}

CCache :: ~CCache()
{            
    RemoveAll();
}

void CCache :: GetStats( int& hits, int& misses, int& inserts, int& removes 
) const
{              
    hits = m_Hits; misses = m_Misses; inserts = m_Inserts, removes = 
m_Removes;
}

void CCache :: InitStats()   
{
    m_Hits       = 0;        
    m_Misses     = 0;
    m_Inserts    = 0;
    m_Removes    = 0;
}

BOOL CCache :: IsFull() const
{
    return m_LRU.GetCount() >= m_EntryLimit;
}
                                                        
CCacheObject* CCache :: Lookup( const char* key )
{                                                        
    BOOL b;                                    
    CCacheObject* rValue;
    
    if( (b = CMapStringToOb::Lookup( key, (CObject*&)rValue )) == TRUE )
    {   // make this object most-recently used, promote to head of m_LRU
        POSITION pos = m_LRU.Find( rValue );   // linear search
        ASSERT( pos );
        m_LRU.RemoveAt( pos );
        m_LRU.AddHead( rValue );

        // make a new object and copy object from cache to new object
        // this is the address of the one to return.
        rValue = rValue->Duplicate();   
        m_Hits++;
    }
    else          
    {
        m_Misses++;              
    }

    return (b == TRUE) ? rValue : NULL;
}


void CCache :: RemoveAll()
{                   
    POSITION pos;
    
    for( pos = m_LRU.GetHeadPosition(); pos != NULL; )
    {   // release memory for all objects in cache
        delete m_LRU.GetNext( pos );        
    }

    CMapStringToOb::RemoveAll(); // release memory for map
    m_LRU.RemoveAll();           // ... and list
    InitStats();                 // clear statistics 
}

BOOL CCache :: RemoveKey( const char* key )
{
    CCacheObject* rValue;
    
    if( (rValue = Lookup( key )) != NULL )
    {    
        CMapStringToOb::RemoveKey( key );   // remove obj from map
        delete m_LRU.RemoveHead();          // & cleanup LRU too
        m_Removes++;
    }              
    
    return rValue != NULL;
}

void CCache :: SetAt( const char* key, CCacheObject* newValue )
{                         
    CString oldKey;
    CCacheObject* myObj;
    
    if( CMapStringToOb::Lookup( key, (CObject*&)myObj ) )                   
                                 
        RemoveKey( key );  // try to remove in case it's a duplicate

    if( IsFull() )
    {
        CCacheObject* old = (CCacheObject*) m_LRU.RemoveTail();
        ASSERT(old);
        // clean up map and release cached object
        CMapStringToOb::RemoveKey( old->GetKey( oldKey ) ); 
        delete old; 
        m_Removes++;
    }            

    // make a copy for the cache and make it most-recently-used 
    CMapStringToOb::SetAt( key, myObj = newValue->Duplicate() );
    m_LRU.AddHead( myObj );
    m_Inserts++;
}                            

// shrink the cache if necessary (min size = 1 entry)
void CCache :: SetEntryLimit( int numEntries )
{   
    CString oldKey;
    m_EntryLimit = __max( numEntries, 1 );

    while( m_EntryLimit < m_LRU.GetCount() )
    {
        CCacheObject* old = (CCacheObject*) m_LRU.RemoveTail();
        ASSERT(old);

        CMapStringToOb::RemoveKey( old->GetKey( oldKey ) );   
        delete old;
        m_Removes++;
    }            
}


