#include <windows.h>
#include <mmsystem.h>
#include "video.h"
#include <digitalv.h>

/*
Opens the given digital video file and returns an
MCIDEVICEID. This ID must be passed to all other video
functions. Returns 0 if an error occurred (such as invalid
file name), non-zero otherwise.
*/
MCIDEVICEID videoOpen(LPSTR lpszFileName, HWND hwndParent,
					LPRECT lprcRect, HWND hwndSignalCallback)
{
	MCI_DGV_OPEN_PARMS	mciOpen;
	MCI_DGV_WINDOW_PARMS mciWindow;
	MCI_DGV_SET_PARMS		mciSet;
	MCI_DGV_SIGNAL_PARMS mciSignal;
	MCI_DGV_PUT_PARMS    mciPut;

	mciOpen.lpstrElementName = lpszFileName;

	if ( mciSendCommand(0, MCI_OPEN,
			 MCI_WAIT | MCI_OPEN_ELEMENT, (DWORD) &mciOpen) )
		return 0;

	if (hwndParent != NULL)
	{
		mciWindow.hWnd       = hwndParent;
		mciWindow.nCmdShow   = SW_SHOWNORMAL;
		mciWindow.lpstrText  = NULL;

		if ( mciSendCommand(mciOpen.wDeviceID, MCI_WINDOW,
								  MCI_WAIT | MCI_DGV_WINDOW_HWND,
								  (DWORD) &mciWindow) )
			return 0;
	}
  
	if ( hwndSignalCallback != NULL)
	{
		// Switch to millisecond time format
		
		mciSet.dwTimeFormat = MCI_FORMAT_MILLISECONDS;

		if ( mciSendCommand(mciOpen.wDeviceID, MCI_SET,
								  MCI_WAIT | MCI_SET_TIME_FORMAT,
								  (DWORD) &mciSet) )
			return FALSE;
	
		// Send an MM_MCISIGNAL message to hwndCallback once
		// every second
		
		mciSignal.dwPeriod 	= VIDEO_SIGNAL_INTERVAL;
		mciSignal.dwCallback = (DWORD) hwndSignalCallback;
		mciSignal.dwUserParm = 0;

		if ( mciSendCommand(mciOpen.wDeviceID, MCI_SIGNAL,
								MCI_WAIT | MCI_DGV_SIGNAL_EVERY |
								MCI_DGV_SIGNAL_POSITION,
								(DWORD) &mciSignal) )
			return FALSE;

		// Switch back to the frame time format
		
		mciSet.dwTimeFormat = MCI_FORMAT_FRAMES;

		if ( mciSendCommand(mciOpen.wDeviceID, MCI_SET,
				 				  MCI_WAIT | MCI_SET_TIME_FORMAT,
				  				  (DWORD) &mciSet) )
			return FALSE;
	}
	
	if ( lprcRect != NULL )
	{
		CopyRect(&mciPut.rc, lprcRect);

		if ( mciSendCommand(mciOpen.wDeviceID, MCI_PUT,
		   MCI_WAIT | MCI_DGV_RECT | MCI_DGV_PUT_DESTINATION,
			(DWORD) &mciPut) )
			return 0;
	}

	return mciOpen.wDeviceID;
}

/*
Closes the open video device.
*/
void videoClose(MCIDEVICEID mciDeviceID)
{
	mciSendCommand(mciDeviceID, MCI_CLOSE, MCI_WAIT, 0);
}

/*
Begins playback of the video in the given window.
lprcRect is the rect in which to scale the video,
where (0,0) is the upper-left corner of the hwndParent
client rect. If lprcRect is NULL, the video is played at
the default size. Returns TRUE if successful, FALSE
otherwise.
*/
BOOL videoPlay(MCIDEVICEID mciDeviceID,
					HWND hwndPlayCallback, BOOL fLoop)
{
	MCI_DGV_PLAY_PARMS mciPlay;
	DWORD dwPlayCommand;

	if (hwndPlayCallback != NULL)
	{
		mciPlay.dwCallback = (DWORD) hwndPlayCallback;
		dwPlayCommand = MCI_NOTIFY;
	}
	else
	{
		dwPlayCommand = 0;
	}

	if (fLoop)
		dwPlayCommand |= MCI_DGV_PLAY_REPEAT;

	if ( mciSendCommand(mciDeviceID, MCI_PLAY,
						 	  dwPlayCommand, (DWORD) &mciPlay) )
		return FALSE;

	return TRUE;
}

/*
Stops the playback of the video.
*/
void videoStop(MCIDEVICEID mciDeviceID)
{
	mciSendCommand(mciDeviceID, MCI_STOP, 0, 0);
}

/*
If the video is playing, pauses the playback.
If the video is paused, resumes the playback.
Otherwise, nothing.
*/
void videoPause(MCIDEVICEID mciDeviceID)
{
	MCI_DGV_STATUS_PARMS mciStatus;

	mciStatus.dwItem = MCI_STATUS_MODE;

	if ( mciSendCommand(mciDeviceID, MCI_STATUS,
			 MCI_WAIT | MCI_STATUS_ITEM, (DWORD) &mciStatus) )
		return;

	switch ( mciStatus.dwReturn )
	{
		case MCI_MODE_PLAY:
			mciSendCommand(mciDeviceID, MCI_PAUSE,
								MCI_WAIT, 0);
			break;

		case MCI_MODE_PAUSE:
			mciSendCommand(mciDeviceID, MCI_RESUME, 0, 0);
			break;
	}
}

/*
Seeks to the given frame.
Returns FALSE if an error occurred (such as position
out-of-range), TRUE otherwise.
*/
BOOL videoSeek(MCIDEVICEID mciDeviceID, DWORD dwPos)
{
	MCI_SEEK_PARMS mciSeek;

	mciSeek.dwTo = dwPos;

	if ( mciSendCommand(mciDeviceID, MCI_SEEK,
						 MCI_WAIT | MCI_TO, (DWORD) &mciSeek) )
		return FALSE;

	return TRUE;
}

/*
Returns the total number of frames in the video
*/
DWORD videoGetFrameCount(MCIDEVICEID mciDeviceID)
{
	MCI_DGV_STATUS_PARMS mciStatus;

	mciStatus.dwItem = MCI_STATUS_LENGTH;

	if ( mciSendCommand(mciDeviceID, MCI_STATUS,
			 MCI_WAIT | MCI_STATUS_ITEM, (DWORD) &mciStatus) )
		return 0;

	return mciStatus.dwReturn;
}

/*
Returns the current frame position.
*/
DWORD videoGetCurrentFrame(MCIDEVICEID mciDeviceID)
{
	MCI_DGV_STATUS_PARMS mciStatus;

	mciStatus.dwItem = MCI_STATUS_POSITION;

	if ( mciSendCommand(mciDeviceID, MCI_STATUS,
			 MCI_WAIT | MCI_STATUS_ITEM, (DWORD) &mciStatus) )
		return 0;

	return mciStatus.dwReturn;
}

/*
Sets the given RECT to the dimensions of the default
rectangle. The default rectangle is the video dimensions
before they are stretched, originating at (0,0).
*/
void videoGetDefaultRect(MCIDEVICEID mciDeviceID,
								 LPRECT lprcVideo)
{
	MCI_DGV_RECT_PARMS mciRect;

	if ( mciSendCommand(mciDeviceID, MCI_WHERE,
							  MCI_WAIT | MCI_DGV_WHERE_SOURCE,
							  (DWORD) &mciRect) )
		SetRectEmpty(lprcVideo);
	else
		CopyRect(lprcVideo, &mciRect.rc);
}

/*
Updates the display of the video.  Call this function
after receiving WM_PAINT.
*/
void videoUpdate(MCIDEVICEID mciDeviceID, HDC hdc)
{
	MCI_DGV_UPDATE_PARMS mciUpdate;

	mciUpdate.hDC = hdc;

	mciSendCommand(mciDeviceID, MCI_UPDATE,
					 	MCI_WAIT | MCI_DGV_UPDATE_HDC,
					 	(DWORD) &mciUpdate);
}
