/*
    File: ZPinup.cpp
    Author: V.Ramachandran
    Date: September 08, 1995

    ZPinupDialog: Toggles between modal and modeless states.
*/

#include "stdafx.h"
#include "windowsx.h"       // For GetInstanceModule
#include "zpinup.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

HHOOK ZPinupDialog::m_hOldHook = NULL;  // Hook procedure.
int ZPinupDialog::m_nNofDialogs = 0;

IMPLEMENT_DYNAMIC (ZPinupDialog, CDialog)

/**************************************************************/
// Constructors
// The constructors resemble the CDialog constructors.
/**************************************************************/

ZPinupDialog::ZPinupDialog (LPCSTR lpszTemplateName, 
                            CWnd *pParent)
    :   CDialog (lpszTemplateName, pParent), m_bPinup (FALSE)
{
}

ZPinupDialog::ZPinupDialog (UINT nID, CWnd *pParent)
    :   CDialog (nID, pParent), m_bPinup (FALSE)
{
}

/**************************************************************/
// Raja 11/01/1995.
// Message map.
/**************************************************************/

BEGIN_MESSAGE_MAP(ZPinupDialog, CDialog)
    //{{AFX_MSG_MAP(ZPinupDialog)
    ON_WM_DESTROY()
    ON_WM_SYSCOMMAND()
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/**************************************************************/
// Create
// Calls CDialog::Create. If this is the first window to be 
// created, hooks the mouse through a call to SetWindowsHookEx.
// For Win32, use GetCurrentThread instead of GetCurrentTask.
/**************************************************************/

BOOL ZPinupDialog::Create (LPCSTR lpszTemplate, CWnd *pWndParent)
{
    BOOL bRet = CDialog::Create (lpszTemplate, pWndParent);
    if (!bRet) {
        ASSERT (FALSE);
        return FALSE;
    }
    if (!m_nNofDialogs) {
#ifdef WIN32
        m_hOldHook = SetWindowsHookEx (WH_MOUSE, 
                    (HOOKPROC)ZPinupDialog::MouseFunc, 
                    GetInstanceModule (AfxGetInstanceHandle ()), 
                    (DWORD)GetCurrentThread ());
#else
        m_hOldHook = SetWindowsHookEx (WH_MOUSE, 
                    (HOOKPROC)ZPinupDialog::MouseFunc, 
                    GetInstanceModule (AfxGetInstanceHandle ()), 
                    GetCurrentTask ());
#endif      
        if (m_hOldHook == NULL) {
            TRACE0 ("ZPinupDialog::Create - Mouse Hook failed.\n");
            ASSERT (FALSE);
            return FALSE;
        }
    }
    m_nNofDialogs ++;
    return bRet;
}

/**************************************************************/
// OnDestroy - Message handler for WM_DESTROY.
// If the last dialog of this class is destroyed, then unhook 
// the mouse hook.
/**************************************************************/

void ZPinupDialog::OnDestroy()
{
    if (--m_nNofDialogs == 0) {
        if (m_hOldHook)
            UnhookWindowsHookEx (m_hOldHook);
    }
    CDialog::OnDestroy();
}

/**************************************************************/
// MouseFunc - Mouse Hook function.
// Refer to the SDK topic MouseProc for details.
//
// If the action is WM_LBUTTONDOWN and WM_NCLBUTTONDOWN, 
// do the following:
//   If the active window is a Pinup dialog, then 
//    If the mouse click is within the dialog, then do nothing.
//   If the click is outside, call a virtual function to 
//    handle the processing.
//   If the virtual function returns TRUE, 
//         Pass the message onto the next hook.
//    Else, discard the message by returning 1.
// NOTE: _export is not required for Win32.
/**************************************************************/

#ifndef WIN32
LRESULT CALLBACK _export ZPinupDialog::MouseFunc (int nCode,
          WPARAM wParam, LPARAM lParam)
#else         // _export not required for Win32
LRESULT CALLBACK ZPinupDialog::MouseFunc (int nCode,
          WPARAM wParam, LPARAM lParam)
#endif // WIN32
{
    MOUSEHOOKSTRUCT *lpStruct;
    BOOL bRet = TRUE;

    if (nCode >= 0) {
        if ((wParam == WM_NCLBUTTONDOWN) 
                        || (wParam == WM_LBUTTONDOWN)) {
            CWnd *pWnd = GetActiveWindow ();
            if (pWnd) {
                if (pWnd->IsKindOf (RUNTIME_CLASS (ZPinupDialog))) {
                    ZPinupDialog *pThis = (ZPinupDialog *)pWnd;
                    ASSERT (pThis);
                    lpStruct = (LPMOUSEHOOKSTRUCT) lParam;
                    if (!pThis->IsPtInDialog (lpStruct->pt))
                        bRet = pThis->DoClickProcessing ();
                }           // ZPinupDialog type
            }               // pWnd != NULL
        }                   // wParam == WM_xxLBUTTONDOWN
    }                       // nCode >= 0.
// If bRet == 1, then pass the message on, else discard 
// by returning 1.
    if (bRet)
        return CallNextHookEx (m_hOldHook, nCode, 
                            wParam, lParam);
    else
        return 1;           // Discard message.
}

/**************************************************************/
// DoClickProcessing - virtual function to handle mouse clicks 
// outside dialog. Called from MouseFunc (mouse hook function)
// Returns TRUE if the dialog is pinned up, FALSE otherwise.
// If the dialog is pinned up, returning TRUE will cause the 
// mouse function to pass the message to the next hook function. 
// If not, the function just beeps and discards the message.
// Note: Made virtual so that you can change the functionality.
/**************************************************************/

BOOL ZPinupDialog::DoClickProcessing () 
{
    if (!GetPinupState ()) {
        MessageBeep (1);
        return FALSE;   // Causes MouseFunc to discard the message.
    }
    else
        return TRUE;
}

/**************************************************************/
// OnSysCommand - MFC Handler for WM_SYSCOMMAND
// nID - specifies the type of command.
// lParam - contains mouse coordinates if command is
// generated using mouse. If mouse is generated using 
// keyboard, contains the key that was used.
// Required to trap the Alt+F6 key. The message is discarded
// if the dialog state is modal.
/**************************************************************/

void ZPinupDialog::OnSysCommand (UINT nID, LPARAM lParam)
{
    nID &= 0xFFF0;      // The last 4 bits are used internally.
    if (((nID == SC_NEXTWINDOW) || (nID == SC_PREVWINDOW)) 
                    && (lParam == VK_F6)) {
        if (!GetPinupState ())
            return;         // Discard message by not passing to base.
    }
    CDialog::OnSysCommand (nID, lParam);
}


/**************************************************************/
// OnOK - virtual function handler for IDOK
// Calls DestroyWindow, since this dialog is actually modeless.
/**************************************************************/

void ZPinupDialog::OnOK ()
{
    UpdateData ();
    DestroyWindow ();
}

/**************************************************************/
// OnCancel - virtual function handler for IDCANCEL
// Calls DestroyWindow, since this dialog is actually modeless.
/**************************************************************/

void ZPinupDialog::OnCancel ()
{
    DestroyWindow ();
}

/**************************************************************/
// DisplayDialog - Public Interface to display dialog box.
// Parameter: Flag to indicate the initial pinup state.
// Calls Create internally. Calls ShowWindow since it is a 
// modeless dialog.
// Uses CDialog internal variables: m_lpDialogTemplate, 
// m_pParentWnd.
/**************************************************************/

BOOL ZPinupDialog::DisplayDialog (BOOL bPinup)
{
    BOOL bRet = Create (m_lpDialogTemplate, m_pParentWnd);
    if (bRet) {
        ShowWindow (SW_SHOW);
        SetPinupState (bPinup);
    }
    return bRet;
}

/**************************************************************/
// IsPtInDialog
// Parameters: Screen coordinates of point.
// Returns TRUE if the point is inside the dialog, 
// FALSE otherwise.
/**************************************************************/

BOOL ZPinupDialog::IsPtInDialog (POINT pt)
{
    RECT rect;
    GetWindowRect (&rect);
    return PtInRect (&rect, pt);
}

/**************************************************************/
// SetPinupState.
// Parameter: Flag indicating whether to pinup or not.
/**************************************************************/

void ZPinupDialog::SetPinupState (BOOL bPinup)
{
    m_bPinup = bPinup;
}

// End of file.
