r/*
	File: WinExec.c
	Author: Raja
	Date: 07/06/1995.

	Supplies a WinExec like function which returns only when the launched
	application exits. Similar to WinExec, but waits till launched
	application to exit.

	UINT WinExecEx (LPCSTR lpszCmdLine, int nCmdShow);

	Remember: This program will not work correctly when run under the debug
	mode. That is because, NotifyRegister handlers cannot be chained, and
	since most debuggers trap this notification, your application handler 
	will never get called.

	Only one application can call WinExecEx, since hWndGlobal is a global
	variable and is shared. Therefore WinExecEx will return an error code of
	31 if another application has already launched an application with
	WinExecEx and waiting for it to terminate.
*/

#include <windows.h>
#include <toolhelp.h>
#include "winexec.h"

// Private declarations.

LRESULT FAR PASCAL _export WinExecWndProc (HWND hwnd, UINT message, 
								WPARAM wParam, LPARAM lParam);
BOOL FAR PASCAL _export NotifyCallbackProc (WORD wID, DWORD dwData);
BOOL RegisterWinExecWindow (HANDLE hInstance);
int CALLBACK LibMain (HINSTANCE hInstance, WORD wDataSeg, WORD cbHeap,
					LPSTR lpszCmdLine);

#define WM_USER_TRACK_INSTANCE 	(WM_USER+1)
#define WM_USER_TASK_ENDED 		(WM_USER+2)

#define GET_TASK_ENDED(hWnd)		(BOOL)GetWindowWord (hWnd, 2)
#define SET_TASK_ENDED(hWnd, bFlag)		SetWindowWord (hWnd, 2, bFlag)
#define GET_HTASK(hWnd)				(HTASK)GetWindowWord (hWnd, 0)
#define SET_HTASK(hWnd, hTask)	SetWindowWord (hWnd, 0, hTask)

#define WINEXEC_ERROR 31		// Less than 32.

HWND hWndGlobal = NULL;
HINSTANCE hInstLib = NULL;

/*************************************************************************/
// Raja 07/06/1995.
// LibMain - DLL Entry point.
// If called for the first time, register the window class.
// Returns TRUE on success, FALSE otherwise.
/*************************************************************************/

int CALLBACK LibMain (HINSTANCE hInstance, WORD wDataSeg, WORD cbHeap,
					LPSTR lpszCmdLine)
{
	if (hInstLib == NULL) {
		hInstLib = hInstance;
		if (!RegisterWinExecWindow (hInstance))
			return 0;
	}
	return 1;
}

/*************************************************************************/
// Raja 07/06/1995.
// WinExecWndProc - Window procedure for the hidden window created.
// Paramterers and return values like other Window procedures.
// WM_CREATE: Register the callback.
// WM_USER_TRACK_INSTANCE: wParam contains the instance of the app to track.
// WM_USER_TASK_ENDED: Message posted by NotifyRegister callback.
// WM_DESTROY: Free the proc, and the callback function.
/*************************************************************************/

LRESULT FAR PASCAL _export WinExecWndProc (HWND hWnd, UINT message, 
								WPARAM wParam, LPARAM lParam)
{
	static LPFNNOTIFYCALLBACK lpfnNotifyProc;
	HANDLE hInstance, hInst;
	HTASK hTask;
	switch (message) {
		case WM_CREATE:
			hInstance = ((LPCREATESTRUCT) lParam)->hInstance;
			lpfnNotifyProc = (LPFNNOTIFYCALLBACK) MakeProcInstance
							(NotifyCallbackProc, hInstance);
			if (!NotifyRegister (NULL, lpfnNotifyProc, NF_NORMAL))
				return -1;
			return 0;

		case WM_USER_TRACK_INSTANCE:		// wParam contains the instance
			hInst = (HINSTANCE)wParam;
			hWndGlobal = hWnd;
			hTask = GetTaskFromInstance (hInst);
			if (hTask) {
				SET_HTASK (hWnd, hTask);
				return TRUE;
			}
			return 0;

		case WM_USER_TASK_ENDED:
			if (wParam == GET_HTASK (hWnd)) {
				SET_TASK_ENDED (hWnd, TRUE);
			}
			return 0;

		case WM_DESTROY:
			FreeProcInstance (lpfnNotifyProc);
			NotifyUnRegister (NULL);
			return 0;
	}
	return DefWindowProc (hWnd, message, wParam, lParam);
}

/*************************************************************************/
// Raja 07/06/1995.
// NotifyCallbackProc
// Returns: FALSE to indicate that it has not handled it.
// Each time an NFY_EXITTASK notification is received, this function
// posts a message to the WinExec window procedure, with wParam
// equal to the current task handle. 
/*************************************************************************/

BOOL FAR PASCAL _export NotifyCallbackProc (WORD wID, DWORD dwData)
{
	if (wID == NFY_EXITTASK)
		PostMessage (hWndGlobal, WM_USER_TASK_ENDED,
					(WPARAM) GetCurrentTask (), 0);
	return FALSE;
}

/*************************************************************************/
// Raja 07/06/1995.
// RegisterWinExecWindow
// Parameter: Instance handle
// Returns: TRUE if RegisterClass succeeds, FALSE otherwise.
// Registers the WinExecWndClass.
/*************************************************************************/

BOOL RegisterWinExecWindow (HANDLE hInstance)
{
    WNDCLASS  wc;

    wc.style = NULL;
    wc.lpfnWndProc = WinExecWndProc;
    wc.cbClsExtra = 0; 
    wc.cbWndExtra = 4;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH); 
    wc.lpszMenuName =  NULL;
    wc.lpszClassName = "WinExecWndClass";
    return (RegisterClass(&wc));
}

/*************************************************************************/
// Raja 07/06/1995.
// WinExecEx - exposed interface
// Parameter and return values similar to WinExec.
// First determine the current instance. This can be done by getting the
// current task, and walking the chain to get the instance handle. 
// Create a window of class ExecWinWndProc. Send it a WM_USER_TRACK_INSTANCE
// message passing the instance in the wParam. Call the original WinExec
// function. If returns without error, go into a message loop after
// disabling the currently active window. Break only when we get a task end
// notification through the NotifyRegister callback function. Enable the
// window and destroy the window.
// If hWndGlobal is not NULL, it means that this function has been executed
// and is waiting for some application to terminate. If so, return error.
/*************************************************************************/

UINT FAR PASCAL __export WinExecEx (LPCSTR szCmdLine, UINT nCmdShow)
{
	HINSTANCE hInst = hInstLib;
	MSG msg;
	HWND hWndTemp, hWndActive;

	if (hWndGlobal != NULL) {
#ifdef _DEBUG
		OutputDebugString ("WinExecEx: Another application already waiting"
			" to terminate.\n");
#endif // _DEBUG
		return WINEXEC_ERROR;
	}

	hWndActive = GetActiveWindow ();
	hWndTemp = CreateWindow("WinExecWndClass", "", WS_OVERLAPPEDWINDOW, 
			0, 0,	0, 0, NULL, NULL, hInst, NULL);
	if (!hWndTemp) {
#ifdef _DEBUG
		OutputDebugString ("WinExecEx: Could not create internal window.\n");
#endif // _DEBUG
		return WINEXEC_ERROR;
	}

// Launch app using SDK WinExec.

	hInst = WinExec (szCmdLine, nCmdShow);
	if (hInst < 32) {
#ifdef _DEBUG
		OutputDebugString ("WinExecEx: WinExec returned error.\n");
#endif // _DEBUG
		return hInst; 
	}

	if (!SendMessage (hWndTemp, WM_USER_TRACK_INSTANCE, hInst, 0)) {
#ifdef _DEBUG
		OutputDebugString ("WinExecEx: TrackInstance returned error.\n");
#endif // _DEBUG
		return WINEXEC_ERROR;
	}

// Disable calling window.
	if (hWndActive)
		EnableWindow (hWndActive, FALSE);

// Go into message loop, checking for termination glag.
	while (GetMessage(&msg,	NULL, NULL,	NULL)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		if (GET_TASK_ENDED (hWndTemp) == TRUE) {
			break;
		}
	}
	DestroyWindow (hWndTemp);
	hWndGlobal = NULL;				// Reinitialize.
	if (hWndActive) {
		EnableWindow (hWndActive, TRUE);
		SetActiveWindow (hWndActive);
	}
	return hInst;
}

/*************************************************************************/
// Raja 07/06/1995.
// GetInstanceFromTask
// Parameter: hTask of the application.
// Returns: The instance handle of the given task.
// Walks the Task chain using the ToolHelp functions TaskNext.
/*************************************************************************/

HINSTANCE FAR PASCAL __export GetInstanceFromTask (HTASK hTask)
{
	TASKENTRY te;
	BOOL bResult;

	te.dwSize = sizeof (te);
	bResult = TaskFirst (&te);
	while (bResult) {
		if (te.hTask == hTask)
			return te.hInst;
		bResult = TaskNext (&te);
	}
	return NULL;
}

/*************************************************************************/
// Raja 07/06/1995.
// GetTaskFromInstance
// Parameter: Instance handle of an application
// Returns: Its task handle.
/*************************************************************************/

HTASK FAR PASCAL __export GetTaskFromInstance (HINSTANCE hInst)
{
	TASKENTRY te;
	BOOL bResult;

	te.dwSize = sizeof (te);
	bResult = TaskFirst (&te);
	while (bResult) {
		if (te.hInst == hInst)
			return te.hTask;
		bResult = TaskNext (&te);
	}
	return NULL;
}

/*************************************************************************/
// Raja 07/06/1995.
// GetCurrentInstance
// Returns the instance handle of the current task.
/*************************************************************************/

HINSTANCE FAR PASCAL __export GetCurrentInstance ()
{
	HINSTANCE hInst;
	HTASK hTask = GetCurrentTask ();
	if (hTask == NULL)
		return NULL;
	hInst = GetInstanceFromTask (hTask);
	return hInst;
}

// End-of-text.
