/*
    File: ListView.cpp
    Author: V.Ramachandran
    Date: 23/11/1995.

    ZListView Class with drag-drop support.
*/

#include "stdafx.h" // For pre-compiled header support.
#include "listview.h"

IMPLEMENT_DYNAMIC (ZListCtrl, CListCtrl)
BEGIN_MESSAGE_MAP(ZListCtrl, CListCtrl)
    //{{AFX_MSG_MAP(ZListCtrl)
    ON_WM_LBUTTONDOWN()
    ON_WM_LBUTTONUP ()      // For drag-drop.
    ON_WM_MOUSEMOVE ()      // For drag-drop.
    ON_MESSAGE (WM_GETDLGCODE, OnCustomGetDlgCode)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

#define new DEBUG_NEW   // For debugging purposes only.

// Default Constructor.
ZListCtrl::ZListCtrl () : CListCtrl (), 
        m_bDragging (FALSE), m_bDragDrop (FALSE)
{
}

/***********************************************************/
// OnLButtonDown - MFC handler for WM_LBUTTONDOWN
// Just set focus to yourself. The listview control does 
// not set focus to itself if the user clicks and drags the 
// left mouse button.
/***********************************************************/

void ZListCtrl::OnLButtonDown (UINT nFlags, CPoint pt)
{
    SetFocus ();    
    CListCtrl::OnLButtonDown (nFlags, pt);
}

/***********************************************************/
// OnLButtonUp - Handler for WM_LBUTTONUP
// Call DoEndDrag to end drag. This the only way to end drag.
/***********************************************************/

void ZListCtrl::OnLButtonUp (UINT nFlags, CPoint point)
{
    CListCtrl::OnLButtonUp (nFlags, point);
    if (m_bDragging)
        DoEndDrag (point, TRUE);
}

/***********************************************************/
//  OnMouseMove - MFC handler for WM_MOUSEMOVE. 
//  Handle visual effects while user is dragging.
/***********************************************************/

void ZListCtrl::OnMouseMove (UINT nFlags, CPoint point)
{
    POINT ptScreen;
    if (m_bDragging) {
        ASSERT (m_hDragImageList);
        CImageList *pDragImage = 
            CImageList::FromHandle (m_hDragImageList);
        ASSERT (pDragImage != NULL);
// Call out to virtual function to handle move.
        HandleDragMove (pDragImage, point);
// Convert to screen coordinates, since DragMove uses the 
// desktop window.
        ptScreen = point;
        ClientToScreen (&ptScreen);
        pDragImage->DragMove (ptScreen);
    }
    CListCtrl::OnMouseMove (nFlags, point);
}

/***********************************************************/
//  OnChildNotify - protected MFC override.
//  Return TRUE if you have handled message, FALSE to pass 
// onto parent. Handle the LVN_BEGINDRAG notification to 
// begin dragging.
/***********************************************************/

BOOL ZListCtrl::OnChildNotify (UINT message, WPARAM wParam, 
                LPARAM lParam, LRESULT *pLResult)
{
    NMHDR *pNotify;
    NM_LISTVIEW *pLV;
    int nView;
    if (message == WM_NOTIFY) {
        pNotify = (NMHDR *)lParam;
        if (pNotify->code == LVN_BEGINDRAG) {
            nView = GetView ();
            if (m_bDragDrop) { 
                pLV = (NM_LISTVIEW *)lParam;
// Call out to virtual function.                    
                if (AllowDragOperation (nView, pLV->iItem))
                    DoBeginDrag (pLV);
            }
        }
    }
    return CListCtrl::OnChildNotify (message, wParam, 
                    lParam, pLResult);
}

/***********************************************************/
//  DoBeginDrag - private helper function
//  Parameter: The listview pointer that comes with the 
// LVN_BEGINDRAG notification.
//  CreateDragImage, get the initial location, the cursor 
// position and then calculate the hotspot position. Hide 
// cursor and capture mouse.
/***********************************************************/

void ZListCtrl::DoBeginDrag (NM_LISTVIEW *pLV)
{
    ASSERT (pLV);
    POINT ptCurrent;
    m_nDragItem = pLV->iItem;

    m_hDragImageList = (HIMAGELIST)SendMessage (
            LVM_CREATEDRAGIMAGE, m_nDragItem, (LPARAM)&ptCurrent);
    if (m_hDragImageList == NULL)   // Error! Should never happen.
        return;
    CImageList *pDragImage = 
                CImageList::FromHandle (m_hDragImageList);
    ASSERT (pDragImage);

// Get the click position. We use GetMessagePos since 
// LVN_BEGINDRAG notification does not give us this info.
    DWORD dwPos = GetMessagePos ();
    POINT ptClickClient, ptClickScreen;
    ptClickScreen.x = LOWORD (dwPos);
    ptClickScreen.y = HIWORD (dwPos);
    ptClickClient = ptClickScreen;
    ScreenToClient (&ptClickClient);

// Get the cursor hotspot position with respect to 
// the top-left of the drag image.
    POINT ptHotSpot;
    ptHotSpot.x = ptClickClient.x - ptCurrent.x;
    ptHotSpot.y = ptClickClient.y - ptCurrent.y;

    if (pDragImage->BeginDrag (0, ptHotSpot)) {
// DragEnter uses the desktop window, so that the cursor 
// gets displayed even if the user moves the cursor out 
// of the listview.
        ImageList_DragEnter (NULL, ptClickScreen.x,
                        ptClickScreen.y);
        ShowCursor (FALSE);
        SetCapture ();
        m_bDragging = TRUE;
    }
}

/***********************************************************/
// DoEndDrag - private helper function
// Ends the drag operation. Called from WM_LBUTTONUP handler.
// Parameter: Point to move to, flag indicating whether to move 
// the item or not. Flag is false if the user presses 
// Escape to abort dragging. If the item is to be moved, then 
// call out to virtual function to handle the actual move.
/***********************************************************/

void ZListCtrl::DoEndDrag (CPoint point, BOOL bMoveItem)
{
    if (m_bDragging) {
        ASSERT (m_hDragImageList);
        CImageList *pDragImage = 
                CImageList::FromHandle (m_hDragImageList);
        ASSERT (pDragImage);
        ImageList_DragLeave (NULL);
        pDragImage->EndDrag ();
        ShowCursor (TRUE);
        ReleaseCapture ();
        ImageList_Destroy (m_hDragImageList);
        m_hDragImageList = NULL;
        m_bDragging = FALSE;
        if (bMoveItem)      
// Call out to virtual function which will handle 
// the actual moving.
            HandleItemMove (m_nDragItem, point);
    }
}

/***********************************************************/
// OnCustomGetDlgCode - handler for WM_GETDLGCODE. We are 
// not using MFC's OnGetDlgCode because we need the w and 
// lParams. For more information, read the note on the MSDN
// CD-ROM PSS No. Q83302.
// This just handles the escape key while you are dragging.
/***********************************************************/

LRESULT ZListCtrl::OnCustomGetDlgCode (UINT wParam, 
                    LPARAM lParam)
{
    LRESULT lResult = Default ();
    if ((m_bDragging) && (lParam)) {
        LPMSG lpMsg = (LPMSG)lParam;
        if ((lpMsg->message == WM_KEYDOWN) && 
                    (lpMsg->wParam == VK_ESCAPE)) {
            POINT pt;       // Not used logically.
            DoEndDrag (pt, FALSE);
            lResult |= DLGC_WANTMESSAGE;
        }
    }
    return lResult;
}

// Default overrideables.
/***********************************************************/
// AllowDragOperation - virtual function. Override to 
// control whether an item can be moved or not.
// Current implementation allows dragging for icon and 
// small icon views.
// Parameters: Current view and item to move.
// Return TRUE to allow drag, FALSE to disallow.
/***********************************************************/

BOOL ZListCtrl::AllowDragOperation (WORD wView, int nItem)
{
    if ((wView == LVS_ICON) || (wView == LVS_SMALLICON))
        return TRUE;
    else
        return FALSE;   
}

/***********************************************************/
// HandleDragMove - overrideable virtual function.
// Parameters: CImageList pointer. This is initially set 
// to the default listview image. pt indicates the client 
// position of the mouse.
// The default version does nothing. 
// Set the imagelist pointer to what you want displayed and 
// the point at which it is to be displayed.
/***********************************************************/

void ZListCtrl::HandleDragMove (CImageList *pDragImage, 
                POINT& pt)
{
    return;
}

/***********************************************************/
// HandleItemMove - virtual function. Override to do the 
// actual move of the item.
// Current implementation moves the item for icon and 
// small icon views.
// Parameters: Dragged item, point at which item was 
// dropped (client coordinates).
/***********************************************************/

void ZListCtrl::HandleItemMove (int nDragItem, CPoint point)
{
    WORD wView = GetView ();
    if ((wView == LVS_ICON) || (wView == LVS_SMALLICON))
        SetItemPosition (m_nDragItem, point);
}

