// DIB Section.cpp : implementation file
//

#include "stdafx.h"
#include "DIB Section.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CDIBSection::CDIBSection()
{
	m_sizeImage.cx			= 0;
	m_sizeImage.cy			= 0;
	m_sizeBitmap.cx		= 0;
	m_sizeBitmap.cy		= 0;
	m_nSize					= 0;
	m_wColorDepth			= 0;
	m_pbyBits				= NULL;
	m_pbyOriginalBits		= NULL;
	m_hBitmap				= NULL;
	m_pPalette				= NULL;
	m_pOriginalPalette	= NULL;
}

CDIBSection::~CDIBSection()
{
	UnloadDIBSection();
}

// Loads the bitmap from a file
BOOL CDIBSection::Load(LPCTSTR pszFileName)
{
	m_hBitmap = LoadDIBSection(NULL, pszFileName);

	return m_hBitmap != NULL;
}

// Loads the bitmap from a resource
BOOL CDIBSection::Load(HINSTANCE hInstance, LPCTSTR pszResourceName)
{
	m_hBitmap = LoadDIBSection(hInstance, pszResourceName);
	
	return m_hBitmap != NULL;
}

// Unloads the bitmap from memory
void CDIBSection::Unload()
{
	UnloadDIBSection();
}

// Converts the bitmap to an HBITMAP handle
CDIBSection::operator HBITMAP() const
{
	return m_hBitmap;
}
	
// Returns the width of the image data
CSize CDIBSection::GetImageSize() const
{
	return m_sizeImage;
}

// Returns the height of the image data
CSize CDIBSection::GetBitmapSize() const
{
	return m_sizeBitmap;
}

// Returns the color depth of the image
// (1 = 2-color, 4 = 16-color,
// 8 = 256-color, 24 = no palette)
WORD CDIBSection::GetColorDepth() const
{
	return m_wColorDepth;
}

// Returns a pointer to the bit data
PBYTE CDIBSection::GetBits() const
{
	return m_pbyBits;
}

// Returns a pointer to the original bit data
PBYTE CDIBSection::GetOriginalBits() const
{
	return m_pbyOriginalBits;
}

// Returns a pointer to the bitmap's palette.
// Returns NULL if bitmap has no palette.
CPalette* CDIBSection::GetPalette() const
{
	return m_pPalette;
}

// Copies the given palette to the bitmap palette.
// Does not affect the original palette.
void CDIBSection::SetPalette(CPalette& palNew)
{
	if (m_hBitmap == NULL || m_pPalette == NULL)
		return;
	
	int nNumColors = m_pPalette->GetEntryCount();

	ASSERT(nNumColors == palNew.GetEntryCount());
	
	PPALETTEENTRY pPalEntries = new PALETTEENTRY[nNumColors];

	if (pPalEntries == NULL)
		return;

	palNew.GetPaletteEntries(0, nNumColors, pPalEntries);
	m_pPalette->SetPaletteEntries(0, nNumColors, pPalEntries);

	LPRGBQUAD pRGBColors;

	pRGBColors = new RGBQUAD[nNumColors];

	if (pRGBColors == NULL)
	{
		delete [] pPalEntries;
		return;
	}

	for (int i = 0; i < nNumColors; i++)
	{
		pRGBColors[i].rgbRed			= pPalEntries[i].peRed;
		pRGBColors[i].rgbGreen		= pPalEntries[i].peGreen;
		pRGBColors[i].rgbBlue		= pPalEntries[i].peBlue;
		pRGBColors[i].rgbReserved	= 0;
	}

	CDC BitmapDC;
	CBitmap* pBitmap;
	CBitmap* pOldBitmap;

	BitmapDC.CreateCompatibleDC(NULL);
	pBitmap = CBitmap::FromHandle(m_hBitmap);
	pOldBitmap = BitmapDC.SelectObject(pBitmap);
	
	SetDIBColorTable(BitmapDC, 0, nNumColors, pRGBColors);

	BitmapDC.SelectObject(pOldBitmap);

	delete [] pPalEntries;
	delete [] pRGBColors;
}

// Renders the image on a device context
void CDIBSection::Draw(CDC* pdc,	const CRect& rcBounds) const
{
	if (m_hBitmap == NULL)
		return;

	if (m_pPalette != NULL)
	{
		pdc->SelectPalette(m_pPalette, FALSE);
		pdc->RealizePalette();
	}

	CBitmap* pBitmap;
	CBitmap* pOldBitmap;
	CDC BitmapDC;

	pBitmap = CBitmap::FromHandle(m_hBitmap);
		
	BitmapDC.CreateCompatibleDC(pdc);

	pOldBitmap = BitmapDC.SelectObject(pBitmap);
										
	pdc->BitBlt(rcBounds.left,
					rcBounds.top,
					rcBounds.Width(),
					rcBounds.Height(),
					&BitmapDC,
					0,
					0,
					SRCCOPY);

	BitmapDC.SelectObject(pOldBitmap);
}

// Restores the bitmap bits to their original value
void CDIBSection::ResetBits()
{
	if (m_hBitmap != NULL)
	{
		ASSERT(m_pbyBits != NULL);
		ASSERT(m_pbyOriginalBits != NULL);
		ASSERT(m_nSize > 0);

		CopyMemory(m_pbyBits, m_pbyOriginalBits, m_nSize);
	}
}

void CDIBSection::ResetPalette()
{
	if (m_pOriginalPalette != NULL)
		SetPalette(*m_pOriginalPalette);
}

//////////////////////////////////////////////////////////
// CDIBSection protected functions

// To load from a file, set hInstance to NULL and pszName
// to the file name. To load from a resource, set
// hInstance to the resource's instance handle and pszName
// to the resource name (or the MAKEINTRESOURCE
// equivalent). Because bitmaps are aligned on a DWORD
// boundary, their bit width might not equal their
// viewable image width. Therefore, this function obtains
// two types of dimensions: one for the viewable image
// (sizeImage) and one for the actual bits (sizeBitmap).
HBITMAP CDIBSection::LoadDIBSection(HINSTANCE hInstance, LPCTSTR pszName)
{
	UnloadDIBSection();
	
	HBITMAP hBitmap;
	DIBSECTION dib;
	UINT fuLoad = LR_CREATEDIBSECTION;

	if (hInstance == NULL)
		fuLoad |= LR_LOADFROMFILE;
	
	hBitmap = (HBITMAP) LoadImage(hInstance,
											pszName,
											IMAGE_BITMAP,
											0,
											0,
											fuLoad);
	if (hBitmap == NULL)
		return NULL;

	GetObject(hBitmap, sizeof(DIBSECTION), &dib);

	CreatePalette(hBitmap, dib.dsBm.bmBitsPixel);

	// Take into account DWORD alignment of scan lines
	m_sizeBitmap.cx = (dib.dsBm.bmWidth + 3) / 4 * 4;
	m_sizeBitmap.cy = dib.dsBm.bmHeight;

	m_nSize = dib.dsBmih.biSizeImage;

	m_pbyOriginalBits = new BYTE[m_nSize];

	if (m_pbyOriginalBits == NULL)
	{
		UnloadDIBSection();
		return NULL;
	}
		
	m_pbyBits = (PBYTE) dib.dsBm.bmBits;
	CopyMemory(m_pbyOriginalBits, m_pbyBits, m_nSize);

	m_sizeImage.cx = dib.dsBm.bmWidth;
	m_sizeImage.cy = dib.dsBm.bmHeight;

	m_wColorDepth = dib.dsBm.bmBitsPixel;
	
	return hBitmap;
}

void CDIBSection::UnloadDIBSection()
{
	if (m_hBitmap != NULL)
		DeleteObject(m_hBitmap);

	if (m_pPalette != NULL)
		delete m_pPalette;

	if (m_pOriginalPalette != NULL)
		delete m_pOriginalPalette;

	if (m_pbyOriginalBits != NULL)
		delete [] m_pbyOriginalBits;

	m_sizeImage.cx			= 0;
	m_sizeImage.cy			= 0;
	m_sizeBitmap.cx		= 0;
	m_sizeBitmap.cy		= 0;
	m_nSize					= 0;
	m_wColorDepth			= 0;
	m_pbyBits				= NULL;
	m_pbyOriginalBits		= NULL;
	m_hBitmap				= NULL;
	m_pPalette				= NULL;
	m_pOriginalPalette	= NULL;
}

void CDIBSection::CreatePalette(HBITMAP hBitmap, WORD wColorDepth)
{
	WORD wNumColors;
	LPRGBQUAD pDIBColors;
	PBYTE pbyTemp;
	PLOGPALETTE pLogPalette;

	// Determine the number of entries in the palette...
	
	switch (wColorDepth)
	{
		case 1:
			wNumColors = 2;
			break;

		case 4:
			wNumColors = 16;
			break;

		case 8:
			wNumColors = 256;
			break;

		default:
			return;
	}
	
	// Obtain an array of RGBQUADs from the DIB section...
	
	pDIBColors = new RGBQUAD[wNumColors];

	if (pDIBColors == NULL)
		return;

	CDC dc;
	CBitmap* pOldBitmap;
	
	dc.CreateCompatibleDC(NULL);
	pOldBitmap = dc.SelectObject( CBitmap::FromHandle(hBitmap) );

	GetDIBColorTable(dc, 0, wNumColors, pDIBColors);
	
	dc.SelectObject(pOldBitmap);
	
	// Create a LOGPALETTE structure and copy the RGBQUADS
   // into it...
	
	pbyTemp = new BYTE[sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * wNumColors];

	if (pbyTemp == NULL)
	{
		delete [] pDIBColors;
		return;
	}
	
	pLogPalette = (PLOGPALETTE) pbyTemp;
      
   pLogPalette->palVersion    = 0x300;
   pLogPalette->palNumEntries = wNumColors;
	
	for (int i = 0; i < wNumColors; i++)
	{
		pLogPalette->palPalEntry[i].peRed	= pDIBColors[i].rgbRed;
      pLogPalette->palPalEntry[i].peGreen = pDIBColors[i].rgbGreen;
      pLogPalette->palPalEntry[i].peBlue  = pDIBColors[i].rgbBlue;
      pLogPalette->palPalEntry[i].peFlags = 0;
	}

	delete [] pDIBColors;
	
	// Finally, create the CPalette from the LOGPALETTE...
	
	m_pPalette = new CPalette;

	if (m_pPalette != NULL)
		m_pPalette->CreatePalette(pLogPalette);
	
	m_pOriginalPalette = new CPalette;
	
	if (m_pOriginalPalette != NULL)
		m_pOriginalPalette->CreatePalette(pLogPalette);
	
	delete [] (PBYTE)pLogPalette;
}

