#include <windows.h>
#include "getver.h"

#define HINSTANCE32 DWORD
#define HWND32      DWORD
#define HFILE32     DWORD

// MS win32 generic thunk functions
HINSTANCE32 (WINAPI* LoadLibraryEx32W) (LPSTR, HFILE32, DWORD);
BOOL        (WINAPI* FreeLibrary32W)   (HINSTANCE32);
FARPROC     (WINAPI* GetProcAddress32W)(HINSTANCE32, LPCSTR);
DWORD       (WINAPI* CallProc32W)      (FARPROC, DWORD, DWORD);

const char os_is_win16[] = "Windows (16 bit)";
const char os_is_win95[] = "Windows 95";
const char os_is_winnt_orig_gui[] = "Windows NT (Orig GUI)";
const char os_is_winnt_new_gui[] = "Windows NT (New GUI)";
const char os_is_undefined[] = "Unknown";


static int  InitThunking()
    {
    int         Result  = FALSE;
    /* no known penalty for failing to unload kernel */
    HINSTANCE   Kernel = LoadLibrary("kernel.dll");

    if(Kernel >= HINSTANCE_ERROR)
        {
        LoadLibraryEx32W    = GetProcAddress(Kernel, "LoadLibraryEx32W");
        FreeLibrary32W      = GetProcAddress(Kernel, "FreeLibrary32W");
        GetProcAddress32W   = GetProcAddress(Kernel, "GetProcAddress32W");
        CallProc32W         = GetProcAddress(Kernel, "CallProc32W");
        if(LoadLibraryEx32W && FreeLibrary32W
        && GetProcAddress32W && CallProc32W)
            Result  = TRUE;
        }
    return Result;
    }


//****************************************************************
// function determines what operating system we are operating   
// Only needs to call kernel32.dll if is Win NT
//
//                              |-just like combined/reversed 
//                              | retcode format of GetVersion()
//                              |
//                              |  Major Minor 
// returns:    O.S.           dwVer  Ver Ver *sLongOSName
//          -----------      ------ ---- --- ------------------
// 16bit app under Win-3.1    3.10   3   10 "Windows (16 bit)"
// 16bit app under Win95      3.95   3   95 "Windows 95"
// 16bit app under WinNT-3.51 3.51   3   51 "Windows NT (Orig GUI)"
// 16bit app under WinNT-4.0  4.0    4    0 "Windows NT (New GUI)"
// undefined OS                 0    0    0 "Unknown"
//
// return code:  OS_IS_WIN16           for Windows 3.0 - 3.11
//               OS_IS_WIN95           for Windows 95: ver >= 3.95
//               OS_IS_WINNT_ORIG_GUI  for Windows NT: ver <= 3.51
//               OS_IS_WINNT_NEW_GUI   for Windows NT: ver >= 4.0
//               OS_IS_UNDEFINED       can't load win32 api -can't
//                                     get version
//
//
//****************************************************************
int TrueGetVersion(DWORD *dwVersion, int *MajorVersion,
           int *MinorVersion, const char **sLongOSName)
{

DWORD MyKernel32Lib;  // our kernel32.dll ptr
// GetVersion() address in kernel32.dll
static FARPROC MyTrueGetVersion32 = NULL;
DWORD dwVersion16BitAPI, dwVersion32BitAPI;

    // check out actual OS with win16 api call
    // note: the win16 api call is not valid for WinNT
    // in that it cant distinguish between different version
    // numbers. it returns 3.1 for both. So will need win32 
    // thunk
    dwVersion16BitAPI = GetVersion();

    if ( (LOBYTE(LOWORD(dwVersion16BitAPI)) >= 3) &&
             (HIBYTE(LOWORD(dwVersion16BitAPI)) >= 95) ) {
        // is win 95 >= 3.95
        *dwVersion = dwVersion16BitAPI;
        *MajorVersion = LOBYTE(LOWORD(dwVersion16BitAPI));
        *MinorVersion = HIBYTE(LOWORD(dwVersion16BitAPI));
        *sLongOSName = os_is_win95;
        return(OS_IS_WIN95);
    } else {
        // not win 95. must be NT or WIN 3.1 or unknown
        #define WF_WINNT 0x4000    // defined in win32.h

        if ( (GetWinFlags() & WF_WINNT) ) {
            // WOW under NT. we need to call GetVersion
            // from win32 api to determine correct 
            // versions via thunking...

            if(!LoadLibraryEx32W && !InitThunking())
                {
                *dwVersion = 0;
                *MajorVersion = 0;
                *MinorVersion = 0;
                return(OS_IS_UNDEFINED);
                }
            // need to load our dll to access our 32bit function
            MyKernel32Lib = LoadLibraryEx32W( "kernel32.dll",
                                               NULL, 0);
            if (!MyKernel32Lib) {
                *dwVersion = 0;
                *MajorVersion = 0;
                *MinorVersion = 0;
                return(OS_IS_UNDEFINED);
            } else {
                // loaded dll ok
                // Attempt to get the address of the win32 
                // GetVersion
                MyTrueGetVersion32 = (FARPROC) GetProcAddress32W(
                                      MyKernel32Lib, "GetVersion");
                if (!MyTrueGetVersion32) {
                    *dwVersion = 0;
                    *MajorVersion = 0;
                    *MinorVersion = 0;
                    *sLongOSName = os_is_undefined;
                    FreeLibrary32W(MyKernel32Lib);
                    return(OS_IS_UNDEFINED);
                } else {
                    // got it so can call function
                    ;
                }
            }
            // call our win32 function via generic thunking 
            // call by ref=1 or value=0, arg count
            // GetVersion() has no parameters....
            dwVersion32BitAPI = (DWORD) CallProc32W(
                                 MyTrueGetVersion32, 0, 0);
            *dwVersion = dwVersion32BitAPI;
            *MajorVersion = LOBYTE(LOWORD(dwVersion32BitAPI));
            *MinorVersion = HIBYTE(LOWORD(dwVersion32BitAPI));

            if ((*MajorVersion <= 3) && (*MinorVersion <= 51)) {
                // is win NT old gui (ver <=3.51)
                *sLongOSName = os_is_winnt_orig_gui;
                // clean up
                FreeLibrary32W(MyKernel32Lib);
                return(OS_IS_WINNT_ORIG_GUI);
            } else {
                // see if new gui
                if ((*MajorVersion >= 4) && (*MinorVersion >= 0)) {
                    // is win NT new gui (ver >= 4.0)
                    *sLongOSName = os_is_winnt_new_gui;

                    // clean up
                    FreeLibrary32W(MyKernel32Lib);
                    return(OS_IS_WINNT_NEW_GUI);
                } else {
                    // bad NT O.S.
                    *dwVersion = 0;
                    *MajorVersion = 0;
                    *MinorVersion = 0;
                    *sLongOSName = os_is_undefined;

                    // clean up
                    FreeLibrary32W(MyKernel32Lib);
                    return(OS_IS_UNDEFINED);
                }
            }
            // no need to unload our kernel32.dll since it is 
            // core part of NT O.S and OS will keep it loaded.
        } else {
            // we are win 3.0 - 3.1. No need to load win32 dll since
            // 16 bit GetVersion(0 works fine in this OS
            *dwVersion = dwVersion16BitAPI;
            *MajorVersion = LOBYTE(LOWORD(dwVersion16BitAPI));
            *MinorVersion = HIBYTE(LOWORD(dwVersion16BitAPI));
            *sLongOSName = os_is_win16;
            return(OS_IS_WIN16);
        }
    }
}
