// gameport.c

#include "ntddk.h"
#include "gameport.h"

typedef struct _DEVICE_EXTENSION {
   PDEVICE_OBJECT    DeviceObject;
   ULONG             TranslatedPort;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS GamePortCreate(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS GamePortClose(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS GamePortRead(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS GamePortWrite(IN PDEVICE_OBJECT, IN PIRP);
VOID GamePortUnload(IN PDRIVER_OBJECT);

//-----------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    NTSTATUS           status = STATUS_SUCCESS;
    UNICODE_STRING     symName, deviceName;
    PDEVICE_OBJECT     deviceObject = NULL;
    PDEVICE_EXTENSION  deviceExt;
    CM_RESOURCE_LIST   resource;
    PHYSICAL_ADDRESS   address, translatedAddress;
    ULONG              addressSpace, busNumber;
    INTERFACE_TYPE     busType;
    BOOLEAN            conflict;

    UNREFERENCED_PARAMETER(RegistryPath);

    KdPrint(("GamePort!DriverEntry: Driver loading\n"));

    //
    // Assume our I/O port is on the first (0) ISA bus.
    //

    busType = Isa;
    busNumber = 0;
    addressSpace = 1;               // I/O address space
    address.LowPart  = GAME_PORT;
    address.HighPart = 0;
 
    //
    // Claim resource usage and check for conflicts.
    //

    RtlZeroMemory((PVOID)&resource, sizeof(resource));
    resource.Count = 1;
    resource.List[0].InterfaceType = busType;
    resource.List[0].BusNumber = busNumber;
    resource.List[0].PartialResourceList.Version = 0;
    resource.List[0].PartialResourceList.Revision = 0;
    resource.List[0].PartialResourceList.Count = 1;
    resource.List[0].PartialResourceList.PartialDescriptors[0].
             Type = CmResourceTypePort;
    resource.List[0].PartialResourceList.PartialDescriptors[0].
             ShareDisposition = CmResourceShareDriverExclusive;
    resource.List[0].PartialResourceList.PartialDescriptors[0].
             Flags = CM_RESOURCE_PORT_IO;
    resource.List[0].PartialResourceList.PartialDescriptors[0].
             u.Port.Start = address;
    resource.List[0].PartialResourceList.PartialDescriptors[0].
             u.Port.Length = 1;

    status = IoReportResourceUsage(NULL, DriverObject, &resource, 
                                   sizeof(resource), NULL, NULL, 0, 
                                   FALSE, &conflict);

    if (conflict) {
        status = STATUS_DEVICE_CONFIGURATION_ERROR;
    }

    if (!NT_SUCCESS(status)) {
         KdPrint(("GamePort!DriverEntry: Can't claim resource\n"));
         return status;
    }

    //
    // Translate bus address into a logical address that NT can use.
    //

    HalTranslateBusAddress(busType, busNumber, address, 
                           &addressSpace, &translatedAddress);

    //
    // Create device object and symbolic link to represent gameport
    //

    RtlInitUnicodeString(&deviceName, GAMEPORT_DEV_NAME);
    RtlInitUnicodeString(&symName, GAMEPORT_LINK_NAME);
                     
    status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_GAMEPORT, 0,
                            FALSE, &deviceObject);

    if (!NT_SUCCESS(status)) {
        RtlZeroMemory((PVOID)&resource, sizeof(resource));
        IoReportResourceUsage(NULL, DriverObject, &resource,
                              sizeof(resource), NULL, NULL, 0, 
                              FALSE, &conflict);
        return status;
    }

    deviceObject->Flags |= DO_BUFFERED_IO;

    status = IoCreateSymbolicLink(&symName, &deviceName);

    if (!NT_SUCCESS(status)) {
        RtlZeroMemory((PVOID)&resource, sizeof(resource));
        IoReportResourceUsage(NULL, DriverObject, &resource,
                              sizeof(resource), NULL, NULL, 0, 
                              FALSE, &conflict);
        IoDeleteDevice(deviceObject);
        return status;
    }

    //
    // Save info in the device extension that we'll need later.
    //

    deviceExt = (PDEVICE_EXTENSION)deviceObject->DeviceExtension;
    deviceExt->DeviceObject = deviceObject;
    deviceExt->TranslatedPort = translatedAddress.LowPart;

    //
    // Fill in the dispatch table.
    //

    DriverObject->DriverUnload                 = GamePortUnload;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = GamePortCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]  = GamePortClose;
    DriverObject->MajorFunction[IRP_MJ_READ]   = GamePortRead;
    DriverObject->MajorFunction[IRP_MJ_WRITE]  = GamePortWrite;
    return status;

} // DriverEntry

//-----------------------------------------------------------------
NTSTATUS GamePortCreate(IN PDEVICE_OBJECT DevObject, IN PIRP Irp)
{
    NTSTATUS status = STATUS_SUCCESS;
    PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);

    KdPrint(("GamePortCreate\n"));

    if (irpStack->FileObject->FileName.Length != 0) {
        status = STATUS_INVALID_PARAMETER;
    }

    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

} // GamePortCreate

//-----------------------------------------------------------------
NTSTATUS GamePortClose(IN PDEVICE_OBJECT DevObject, IN PIRP Irp)
{
    KdPrint(("GamePortClose\n"));

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;

} // GamePortClose

//-----------------------------------------------------------------
NTSTATUS GamePortRead(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    NTSTATUS status = STATUS_SUCCESS;
    PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION deviceExt = DeviceObject->DeviceExtension;

    KdPrint(("GamePortRead\n"));

    //
    // Validate parameters and read a byte from the gameport.
    //   

    if (irpStack->Parameters.Read.Length == 0) {
        status = STATUS_INVALID_PARAMETER;
        Irp->IoStatus.Information = 0;
    } else {
        *(PUCHAR)(Irp->AssociatedIrp.SystemBuffer) = 
            READ_PORT_UCHAR((PUCHAR)(deviceExt->TranslatedPort));
        Irp->IoStatus.Information = 1;   // 1 byte transfered
    }

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

} // GamePortRead

//-----------------------------------------------------------------
NTSTATUS GamePortWrite(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    NTSTATUS status = STATUS_SUCCESS;
    PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION deviceExt = DeviceObject->DeviceExtension;

    KdPrint(("GamePortWrite\n"));

    //
    // Validate parameters and write a byte to the gameport.
    //   

    if (irpStack->Parameters.Write.Length == 0) {
        status = STATUS_INVALID_PARAMETER;
        Irp->IoStatus.Information = 0;
    } else {
        WRITE_PORT_UCHAR((PUCHAR)(deviceExt->TranslatedPort),
            *(PUCHAR)(Irp->AssociatedIrp.SystemBuffer));
        Irp->IoStatus.Information = 1;   // 1 byte transfered
    }

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

} // GamePortWrite

//-----------------------------------------------------------------
VOID GamePortUnload(IN PDRIVER_OBJECT DriverObject)
{
    UNICODE_STRING   symName;
    CM_RESOURCE_LIST resource;  
    BOOLEAN          conflict;

    KdPrint(("GamePortUnload\n"));
        
    //
    // Release claimed hardware resources, symbolic link and device
    // object.
    //
   
    RtlZeroMemory((PVOID)&resource, sizeof(resource));
    IoReportResourceUsage(NULL, DriverObject, &resource,
                          sizeof(resource), NULL, NULL, 0, 
                          FALSE, &conflict);

    RtlInitUnicodeString(&symName, GAMEPORT_LINK_NAME);
    IoDeleteSymbolicLink(&symName);
    IoDeleteDevice(DriverObject->DeviceObject);

} // GamePortUnload
